from Tkinter             import *                  # get widget classes
from tkMessageBox        import *                  # get standard dialogs
from tkFileDialog        import *                  # get askopenfiles dialog
from lib.DEVS            import *
from lib.Simulator       import Simulator
from DevsPlotter         import DevsPlotter
from DataStructure       import *
import os
import sys
import string
import pickle


class DEVSEnv(Frame):                                # an extended frame
  def __init__(self, parent=None, modelDict=None):   # attach to top-level?
    Frame.__init__(self, parent)                     # do superclass init
    self.parent = parent
    self.pack(expand=YES, fill=BOTH)
    self.master.title("DEVS Simulation Environment")        # set window-manager info
    self.master.iconname("DEVS Simulation Environment")     # label when iconified
    self.modelDict    = modelDict
    self.simModel     = None                                # current simulation model
    #Variables for parameters
    self.parameters   = []
    self.paramEntries = {}
    self.paramValues  = {}
    #Variables for plot options
    self.simTime      = 1000
    self.scale        = 1.0
    self.realTime     = False
    self.instances    = []
    self.options      = {}
    #self.optionGrid   = {}
    self.optionPlotter= {}
    self.variables    = {}
    self.mainPlotter  = None
    #self.plotter   = None    
    #self.paramDialog  = None
    self.modelListbox = None
    self.project      = None
    self.createWidgets()                       # attach frames/widgets
        
  def createWidgets(self):
    self.makeMenuBar()
    self.makeToolBar()
    self.makeStatusBar()
      
    self.plotterFrame = Frame(self, cursor='hand2', relief=RIDGE, bd=2)
    self.plotterFrame.pack(expand=YES, fill=BOTH) #grid(row=1, column=1, sticky=NSEW)


  def makeToolBar(self):
    toolbar = Frame(self, cursor='hand2', relief=SUNKEN, bd=2)
    toolbar.pack(side=TOP, fill=X)

    #Build the list area
    listArea = Frame(toolbar, cursor='hand2', relief=RAISED, bd=3)
    listArea.pack(side=LEFT, fill=Y)
    L = Label(listArea, text='Select a Model:')
    L.config(relief=FLAT, anchor=W)
    L.pack(side=TOP, fill=X)

    sbar = Scrollbar(listArea)
    listbox = Listbox(listArea, relief=SUNKEN)
    sbar.config(command=listbox.yview)
    listbox.config(yscrollcommand=sbar.set)
    listbox.config(selectmode=SINGLE, setgrid=1, width=20, height=5)
    listbox.bind('<Double-1>', self.listHandler)
    sbar.pack(side=RIGHT, fill=Y)
    listbox.pack(side=LEFT, fill=Y)
    self.modelListbox = listbox
        
    #Build the parameter area
    paraArea = Frame(toolbar, cursor='hand2', relief=RIDGE, bd=2)
    paraArea.pack(side=TOP, fill=X)
    L = Label(paraArea, text='Parameters:')
    L.config(relief=FLAT, anchor=W)
    L.grid(row=0, column=0, sticky=NSEW)
    self.paramFrame = Frame(paraArea, cursor='hand2', relief=RIDGE, bd=2)
    self.paramFrame.grid(row=1, column=0, sticky=NSEW)
    
    L = Label(paraArea, text='Plot Options:')
    L.config(relief=FLAT, anchor=W)
    L.grid(row=0, column=1, sticky=NSEW)    
    self.plotOptionFrame = Frame(paraArea, cursor='hand2', relief=RIDGE, bd=2)
    self.plotOptionFrame.grid(row=1, column=1, sticky=NSEW)
    
    top=self.winfo_toplevel() 
    top.rowconfigure(0, weight=1)
    top.columnconfigure(1, weight=1) 
    paraArea.rowconfigure(0, weight=1)
    #paraArea.columnconfigure(0, weight=1)    
    paraArea.columnconfigure(1, weight=1)    
    self.paraArea=paraArea
    
    #Build the button areas
    btnArea = Frame(toolbar, cursor='hand2', relief=RAISED, bd=2)
    btnArea.pack(side=BOTTOM, expand=YES, fill=BOTH)	
    self.btnStart = Button( btnArea
                          , text='Start'
                          , state=DISABLED
                          , command=self.startSimulation)
    self.btnStart.grid(row=0, column=0, sticky=W)
    #Button(btnArea, text='Pause',  command=self.quit).grid(row=0, column=1, sticky=W)
    #Button(btnArea, text='Reset',  command=self.quit).grid(row=0, column=2, sticky=W)
    self.btnStop = Button( btnArea
                         , text='Stop'
                         , state=DISABLED
                         , command=self.stop)
    self.btnStop.grid(row=0, column=1, sticky=W)
        
  def makeStatusBar(self):
    #build status bar
    statusbar = Frame(self, cursor='hand2', relief=FLAT, bd=2)
    statusbar.pack(side=BOTTOM, fill=X)
    L = Label(statusbar, text='DEVS Simulation Environment')
    L.config(relief=SUNKEN)
    L.pack(expand=YES, fill=BOTH)

  def makeMenuBar(self):
    self.menubar = Menu(self.master, relief=FLAT)
    self.master.config(menu=self.menubar)    # master=top-level window
    self.fileMenu()
    #self.editMenu()
    #self.imageMenu()

  def fileMenu(self):
    pulldown = Menu(self.menubar)
    pulldown.add_command(label='Open...', command=self.openFile)
    pulldown.add_command(label='Quit',    command=self.quit)
    self.menubar.add_cascade(label='File', underline=0, menu=pulldown)

  def editMenu(self):
    pulldown = Menu(self.menubar)
    pulldown.add_command(label='Paste',   command=self.notdone)
    pulldown.add_command(label='Spam',    command=self.greeting)
    pulldown.add_separator()
    pulldown.add_command(label='Delete',  command=self.greeting)
    pulldown.entryconfig(4, state=DISABLED)
    self.menubar.add_cascade(label='Edit', underline=0, menu=pulldown)
    
  def createParamEntries(self):
  
    if(self.paramFrame):
      self.paramFrame.destroy()
      
    self.paramFrame = Frame(self.paraArea, cursor='hand2', relief=RIDGE, bd=2)
    self.paramFrame.grid(row=1, column=0, sticky=NSEW)

    i = 0
    for param in self.parameters:
      name = Label(self.paramFrame, text=param.name, anchor=W, relief=FLAT)
      name.grid(row=i/2, column=0+i%2*3, sticky=W)
      type = Label(self.paramFrame, text='('+param.type+')', anchor=W, relief=FLAT)
      type.grid(row=i/2, column=1+i%2*3, sticky=W)
      value = Entry(self.paramFrame, relief=SUNKEN) #RIDGE)
      value.grid(row=i/2, column=2+i%2*3, sticky=W)
      value.insert(END, param.value)
      self.paramEntries[param.name]=value
      i = i + 1      

    self.paramFrame.columnconfigure(2, weight=1)    
                
  def getParamString(self):
    for param in self.parameters:
      param.value = self.paramEntries[param.name].get()  #type conversion possiblly needed

    paramStr = ''
    first = True
    for param in self.parameters:
      if(first):
        paramStr += param.name + '=' + param.value  #self.paramValues[param.name]
        first = False
      else:
        paramStr += ', ' + param.name + '=' + param.value #self.paramValues[param.name]
        
    return paramStr
  
  def createAtomicPlotOption(self): 
    if(self.plotOptionFrame):
      self.plotOptionFrame.destroy()
      
    self.plotOptionFrame = Frame(self.paraArea, cursor='hand2', relief=RIDGE, bd=2)
    self.plotOptionFrame.grid(row=1, column=1, sticky=NSEW)
    
    l = Label(self.plotOptionFrame, text='Simulation Time:')
    l.grid(row=0, column=0, sticky=E) #pack(side=LEFT)  
    self.timeEntry = Entry(self.plotOptionFrame, width=12)
    self.timeEntry.insert(END, str(self.simTime))
    self.timeEntry.grid(row=0, column=1, sticky=W) #pack(side=LEFT)
    #self.scaleEntry.bind('<KeyPress-Return>', self.setScale)

    l = Label(self.plotOptionFrame, text='Plotter Scale:')
    l.grid(row=0, column=2, sticky=E) #pack(side=LEFT)
    self.scaleEntry = Entry(self.plotOptionFrame, width=12)
    self.scaleEntry.insert(END, str(self.scale))
    self.scaleEntry.grid(row=0, column=3, sticky=W) #pack(side=LEFT)
    
    self.simModeVar = IntVar()
    ch=Checkbutton( self.plotOptionFrame
                  , text='Real Time'
                  , variable=self.simModeVar)
    ch.grid(row=0, column=4, columnspan=2) #pack(side=LEFT)

    self.plotOptionFrame.rowconfigure(0, weight=1)
    self.plotOptionFrame.columnconfigure(0, weight=1)    
    self.plotOptionFrame.columnconfigure(1, weight=1)
    self.plotOptionFrame.columnconfigure(2, weight=1)    
    self.plotOptionFrame.columnconfigure(3, weight=1)        
    self.plotOptionFrame.columnconfigure(4, weight=2) 
    
  def createCoupledPlotOption(self):
    if(self.plotOptionFrame):
      self.plotOptionFrame.destroy()
      
    self.plotOptionFrame = Frame(self.paraArea, cursor='hand2', relief=RIDGE, bd=2)
    self.plotOptionFrame.grid(row=1, column=1, sticky=NSEW)
    
  
    optionbar = Frame(self.plotOptionFrame, cursor='hand2', relief=RIDGE, bd=2)
    optionbar.grid(row=0, column=0, sticky=NSEW) #pack(side=LEFT) #, fill=X)        

    i = 0
    for inst in self.instances:
      var = IntVar()
      ch=Checkbutton( optionbar
                    , text=inst.type+'.'+inst.name
                    , variable=var
                    , command=self.optionChanged)
      ch.config(anchor=W, width=20)
      ch.select()
      ch.grid(row=i/2, column=i%2) #pack(side=LEFT)
      self.variables[inst]=var
      i = i+1
      
    timebar = Frame(self.plotOptionFrame, cursor='hand2', relief=RIDGE, bd=2)
    timebar.grid(row=1, column=0, sticky=NSEW) #pack(side=LEFT) #, fill=X)        
      
    l = Label(timebar, text='Simulation Time:')
    l.grid(row=0, column=0, sticky=E) #pack(side=LEFT)
    self.timeEntry = Entry(timebar, width=12)
    self.timeEntry.insert(END, str(self.simTime))
    self.timeEntry.grid(row=0, column=1, sticky=W) #pack(side=LEFT)
    #self.scaleEntry.bind('<KeyPress-Return>', self.setScale)

    l = Label(timebar, text='Plotter Scale:')
    l.grid(row=0, column=2, sticky=E) #pack(side=LEFT)
    self.scaleEntry = Entry(timebar, width=12)
    self.scaleEntry.insert(END, str(self.scale))
    self.scaleEntry.grid(row=0, column=3, sticky=W) #pack(side=LEFT)
     
    self.simModeVar = IntVar()
    ch=Checkbutton( timebar
                  , text='Real Time'
                  , variable=self.simModeVar)
    ch.grid(row=0, column=4, columnspan=2) #pack(side=LEFT)
   
    self.plotOptionFrame.rowconfigure(0, weight=1)
    self.plotOptionFrame.rowconfigure(1, weight=1)    
    self.plotOptionFrame.columnconfigure(0, weight=1)    

    timebar.rowconfigure(0, weight=1)
    timebar.columnconfigure(0, weight=1)    
    timebar.columnconfigure(1, weight=1)    
    timebar.columnconfigure(2, weight=1)    
    timebar.columnconfigure(3, weight=1)    
    timebar.columnconfigure(4, weight=2) 
                          
  def getTime(self):
    self.simTime = int(self.timeEntry.get())
    return self.simTime

  def getScale(self):
    self.scale = float(self.scaleEntry.get())
    return self.scale
    
  def isRealTime(self):
    self.realTime = self.simModeVar.get()
    return self.realTime
    
   
  def getOptions(self):
    if(self.simModel.isAtomic()):
      return {}
    else:
      for inst in self.instances:
        self.options[inst]=self.variables[inst].get()
      return self.options
          
  def listHandler(self, event):
    #print dir(event)
    #print event.num
    index = self.modelListbox.curselection()
    label = self.modelListbox.get(index)
    self.simModel = self.modelDict[label]
    self.parameters = self.simModel.parameters #self.modelDict[label].parameters

    self.createParamEntries()
    
    if(self.simModel.isAtomic()): #self.modelDict[label].isAtomic()):
      self.createAtomicPlotOption()
    else:
      self.options      = {}
#      self.optionGrid   = {}
      self.optionPlotter= {}
      self.variables    = {}
      self.mainPlotter  = None

#      self.optionPlotter={}
      self.instances = self.simModel.sources #self.modelDict[label].sources
      #print self.instances
      self.createCoupledPlotOption()

    if(self.plotterFrame):
      self.plotterFrame.destroy()
    
    self.btnStart.config(state=NORMAL)
    self.btnStop.config(state=DISABLED)
   
  def setModelList(self, modelDict):
    self.modelDict = modelDict
    if(self.modelListbox.size()>0):
      self.modelListbox.delete(0, self.modelListbox.size()-1)
    pos = 0
    keys = modelDict.keys()
    for key in keys:
      self.modelListbox.insert(pos, key)
      pos = pos + 1

  def openFile(self):
    pathFile = askopenfilename( title       = "Select a project file to open"
                              , filetypes   = [("Select an empty directory", "*.prj")]
			      , initialfile = ''
			      , initialdir   = os.getcwd()
			      , parent = self.parent
			      )
    if(pathFile):
      projectFile = open(pathFile, 'r')
      self.project = pickle.load(projectFile)
      self.setModelList(self.project.components)
      self.project.addToSysPath()

  def startSimulation(self):
    self.btnStart.config(state=DISABLED)
    self.btnStop.config(state=NORMAL)
    self.simModelName = self.simModel.modelName  
    self.getOptions()  
    self.plotterFrame = Frame(self, cursor='hand2', relief=RIDGE, bd=2)
    self.plotterFrame.pack(expand=YES, fill=BOTH) #grid(row=1, column=1, sticky=NSEW)

    simStat = ''
    simStat += '\nimport ' + self.simModelName
    simStat += '\nself.simInst = ' + self.simModelName+'.'+ self.simModelName \
               + '(' + self.getParamString() + ')'
    simStat += '\nself.simInst.instName = \'simInst\''
    simStat += '\nself.simInst.instType = \'' + self.simModelName + '\''
    stat = compile( simStat, '<string>', 'exec')
    exec(stat)

    #if(self.plotter!=None): del self.plotter  #.destroy()
    mList = []
    #print mList
    i = 0
    if(self.simModel.isAtomic()):
      rTitle='State'
      mList = self.simModel.sources
    else:
      self.optionGrid={}
      self.optionPlotter={}
      rTitle='Instance'
      for inst in self.simModel.sources:
        mList.append(inst.type+'.'+inst.name)
        model = self.modelDict[inst.type]
        if(model.isAtomic()):
          subList = model.sources
          subTitle = 'state'    
        self.plotter = DevsPlotter( parent=self.plotterFrame
                                  , rowList=subList
                                  , eventList=[]
                                  , title=inst.type+'.'+inst.name
                                  , rowTitle=subTitle
                                  , plotterScale=self.getScale())
        #if(self.options[inst]):
        #  self.plotter.grid(row=i/2, column=i%2, sticky=NSEW)  #pack(expand=YES, fill=BOTH, ipady=10)
        #  i = i + 1
        #print inst
        self.optionPlotter[inst]=self.plotter
        #self.optionGrid[inst]=i/2, i%2
        simStat=''
        simStat+='\nself.plotter.setEventSource(self.simInst.'+inst.name+')'
        simStat+='\nprint self.simInst.'+inst.name+'.handler.listeners'
        #print simStat
        stat = compile(simStat, '<string>', 'exec')
        exec(stat)

      mList.append(self.simInst.instType+'.'+self.simInst.instName)

    if(self.simModel.isAtomic()):
      self.mainPlotter = DevsPlotter( parent=self.plotterFrame
                           , rowList=mList
                           , eventList=[]
                           , title=self.simInst.instType+'.'+self.simInst.instName
                           , rowTitle=rTitle
                           , plotterScale=self.getScale())
      self.mainPlotter.grid(row=0, column=0, sticky=NSEW)  #pack(expand=YES, fill=BOTH, ipady=10)
      self.mainPlotter.setEventSource(self.simInst)

      self.plotterFrame.rowconfigure(0, weight=1)
      self.plotterFrame.columnconfigure(0, weight=1)

    else:
      self.mainPlotter = DevsPlotter( parent=self.plotterFrame
                           , rowList=mList
                           , eventList=[]
                           , title=self.simInst.instType+'.'+self.simInst.instName
                           , rowTitle=rTitle
                           , plotterScale=self.getScale())
      #self.mainPlotter.grid(row=i/2, column=i%2, sticky=NSEW)  #pack(expand=YES, fill=BOTH, ipady=10)
      self.mainPlotter.setEventSource(self.simInst)
      self.showCoupledPlotters()
      #for j in range((i+1)/2):
      #  self.plotterFrame.rowconfigure(j, weight=1)

      #self.plotterFrame.columnconfigure(0, weight=1)
      #self.plotterFrame.columnconfigure(1, weight=1)
      #print 'self.realTime:', self.realTime
    self.simulator = Simulator( self.simInst, time=self.getTime(), isRealTime=self.isRealTime())
    self.simulator.start()

  def optionChanged(self):
    self.getOptions()
    self.showCoupledPlotters()

    #self.plotterFrame.destroy()
  def showCoupledPlotters(self):
    if(len(self.optionPlotter.keys())>0):
      sumValue = 0
      for inst in self.instances:
        sumValue += self.options[inst]     
        self.optionPlotter[inst].grid_remove()

      #print sumValue
      if(self.mainPlotter):
        self.mainPlotter.grid_remove()
      #self.mainPlotter.grid(row=0, column=0, sticky=NSEW)

      if(sumValue>2):
        i=0
        for inst in self.instances:
          if(self.options[inst]):
            self.optionPlotter[inst].grid(row=i/2,column=i%2, sticky=NSEW)
            i=i+1

        self.mainPlotter.grid(row=i/2, column=i%2, sticky=NSEW)

        for j in range((i+1)/2):
          self.plotterFrame.rowconfigure(j, weight=1)

        self.plotterFrame.columnconfigure(0, weight=1)
        self.plotterFrame.columnconfigure(1, weight=1)
      else:
        i=0
        for inst in self.instances:
          if(self.options[inst]):
            self.optionPlotter[inst].grid(row=i,column=0, sticky=NSEW)
            i=i+1

        self.mainPlotter.grid(row=i, column=0, sticky=NSEW)
        for j in range(i+1):
          self.plotterFrame.rowconfigure(i, weight=1)

        self.plotterFrame.columnconfigure(0, weight=1)

  def stop(self):
    self.btnStop.config(state=DISABLED)    
    self.simulator.setDone()
  
  def quit(self):
    if(self.project):
      self.project.save()   
      self.project.removeFromSysPath()
    self.parent.destroy()
    
if __name__ == '__main__':
  root = Tk()
  devs = DEVSEnv(root)
  root.mainloop()  # if I'm run as a script
