# AtomicDEVS model: Processor 


from lib.DEVS import *
from lib.Simulator import *
from whrandom import *
from Job import *

class Processor(AtomicDEVS):
  Idle='Idle'
  Busy='Busy'
  Discarding='Discarding'
  INFINITY = 1000000

  def __init__(self, n):
    AtomicDEVS.__init__(self)
    self.n = n
    self.queue = []
    self.currentJob = None
    self.timeElapsed = 0.0
    self.state = Processor.Idle
    self.IN = self.addInPort()
    self.IN.instName = 'IN'
    self.IN.instType = 'InPort'
    self.OUT = self.addOutPort()
    self.OUT.instName = 'OUT'
    self.OUT.instType = 'OutPort'
    self.DISCARD = self.addOutPort()
    self.DISCARD.instName = 'DISCARD'
    self.DISCARD.instType = 'OutPort'

  def intTransition(self):
    if(self.state==Processor.Busy):
      self.timeElapsed = 0.0
      if(len(self.queue)<=0):
        self.currentJob = None
        return Processor.Idle
      elif(len(self.queue)>0):
        self.currentJob = self.queue[0]
        self.queue = self.queue[1:]
        return Processor.Busy
    elif(self.state==Processor.Discarding):
      self.queue=self.queue[:-1]
      return Processor.Busy

  def extTransition(self):
    p = self.peek(self.IN)
    self.queue.append(p)
    if(self.state==Processor.Idle):
      self.currentJob = self.queue[0]
      self.queue = self.queue[1:]
      return Processor.Busy
    elif(self.state==Processor.Busy):
      self.timeElapsed = self.timeElapsed + self.elapsed
      if(len(self.queue)>self.n):
        return Processor.Discarding
      elif(len(self.queue)<=self.n):
        return Processor.Busy

  def timeAdvance(self):
    if(self.state==Processor.Idle):
      return Processor.INFINITY
    elif(self.state==Processor.Busy):
      return self.currentJob.size - self.timeElapsed
    elif(self.state==Processor.Discarding):
      return 0

  def outputFnc(self):
    if(self.state==Processor.Idle):
      print 'Somethin Wrong'
    elif(self.state==Processor.Busy):
      self.poke(self.OUT, self.currentJob)
    elif(self.state==Processor.Discarding):
      self.poke(self.DISCARD, self.queue[-1])