from graph.GraphRepr.Graph import Graph
from graph.GraphRepr.Node import Node

from graph.GraphManager.AGLLoader import AGLLoader
from graph.GraphManager.AGLGenerator.AGLGenerator import AGLGenerator

from GraphRepr.Types.SimpleType.SimpleType import SimpleType
from GraphRepr.Types.SimpleType.Int import Int
from GraphRepr.Types.SimpleType.Bool import Bool
from GraphRepr.Types.SimpleType.String import String
from GraphRepr.Types.SimpleType.Float import Float

from GraphRepr.Types.CollectionType.CollectionType import CollectionType
from GraphRepr.Types.CollectionType.List import List

from GraphRepr.Types.MappingType.MappingType import MappingType
from GraphRepr.Types.MappingType.SymbolTable import SymbolTable

from PrettyVisitor import PrettyVisitor

class Tree:

    # Grammar for this simple language:

    #    program -> statements
    #    statements -> statement | statements statement
    #    statement -> exp ';'
    #    exp -> expINT | expPLUS
    #    expINT -> intConstToken
    #    expPLUS -> expINT '+' expINT

    def __init__(self, name="AST"):
        self.tree=Graph(name)
        self.tree.addElement(Node("+"))
        self.tree.addElement(Node("EXPSEQUENCE"))
        self.tree.addElement(Node("STATEMENT"))
        self.tree.addElement(Node("STATEMENTSEQUENCE"))
        self.root=None
        self.prettyV=PrettyVisitor()
        self.id=0
        self.name=name


    def makeEXPInt(self, value):
        self.id+=1
        ID="INTEXP"+str(self.id)
        elem=Int(value, ID)
        self.tree.addElement(elem)
        return elem

    def makeEXPPlus(self, EXP1, EXP2):
        self.id+=1
        ID="PLUSEXP"+str(self.id)
        elem=Node(ID, type="+")
        self.tree.addElement(elem)
        self.tree.connect(ID, EXP1.getLocalId())
        self.tree.connect(ID, EXP2.getLocalId())
        self.root=elem
        return elem

    def makeEXPSequence(self, EXP1, EXP2):
        self.id+=1
        ID="EXPSEQUENCE"+str(self.id)
        elem=Node(ID, type="EXPSEQUENCE")
        self.tree.addElement(elem)
        self.tree.connect(ID, EXP1.getLocalId())
        self.tree.connect(ID, EXP2.getLocalId())
        self.root=elem
        return elem

    def makeEXPStatement(self, EXP):
        self.id+=1
        ID="STATEMENT"+str(self.id)
        elem=Node(ID, type="STATEMENT")
        self.tree.addElement(elem)
        self.tree.connect(ID, EXP.getLocalId())
        self.root=elem
        return elem

    def makeStatementSequence(self, STM1, STM2):
        self.id+=1
        ID="STATEMENTSEQUENCE"+str(self.id)
        elem=Node(ID, type="STATEMENTSEQUENCE")
        self.tree.addElement(elem)
        self.tree.connect(ID, STM1.getLocalId())
        self.tree.connect(ID, STM2.getLocalId())
        self.root=elem
        return elem

    def toXML(self, location):
        mapping={}
        mapping[self.name]=location
        gen=AGLGenerator(self.tree, mapping)
        gen.genCode()

    def prettyPrint(self,elem=None):
        if elem==None:
            elem=self.root

        elem.enter(self.prettyV)

        outC=elem.outC().iterate()
        while not outC.isDone():
            curElem=outC.currentElement()
            self.prettyPrint(curElem)
            elem.enter(self.prettyV)
            outC.next()

        elem.exit(self.prettyV)


T=Tree()

# Simulating a parse for :

# 5 + 10;
# 1 + 5;
# 3 + 4;
# 999;


a=T.makeEXPInt(5)
b=T.makeEXPInt(10)
c=T.makeEXPPlus(a,b)

Sc = T.makeEXPStatement(c)

d=T.makeEXPInt(1)
e=T.makeEXPInt(5)
f=T.makeEXPPlus(d,e)

Sf = T.makeEXPStatement(f)

s=T.makeStatementSequence(Sc,Sf)

g=T.makeEXPInt(3)
h=T.makeEXPInt(4)
i=T.makeEXPPlus(g,h)

Si = T.makeEXPStatement(i)

r=T.makeStatementSequence(s,Si)

j=T.makeEXPInt(999)
Sj = T.makeEXPStatement(j)

z=T.makeStatementSequence(r,Sj)


T.prettyPrint()

T.toXML("C:/mcgill/simpleAST/Tree.agl")


