# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the AST module

from ASTExp import *

indent='  '
#---------------EquationPart--------------------

# abstracrt class 
class EquationStm:

    lineNum=-1
    colNum=-1

    def __init__(self, lineNumP, colNumP):
	raise "Abstract class EquationStm can not be instantiated!"

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

#---------------Atomic Equation--------------------
# abstracrt class
class AtomicEqStm(EquationStm):
    def __init__(self):
        raise "Abstract class AtomicEqStm can not be instantiated!"
    
class RegularEqStm(AtomicEqStm):
    def __init__(self, lineNumP, colNumP, LHSP, RHSP, commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.LHS=LHSP # Exp
	self.RHS=RHSP
	self.comment=commentP # CommentExp

    def getLHS(self):
	return self.LHS

    def getRHS(self):
	return self.RHS

    def setLHS(self, lhs):
	self.LHS=lhs

    def setRHS(self, rhs):
	self.RHS=rhs

    def setOuterScope(self, st):
        pass

    def accept(self, v):
        v.visitRegularEqStm(self)

    # types: RealType, IntegerType, BooleanType, StringType, Void,
    #        TypeError
    def getType1(self):
        if self.LHS.getType()==self.RHS.getType(): #or \
	   #(self.LHS.getType()=='RealType' and \
	   # self.RHS.getType()=='IntegerType') or \
	   #(self.LHS.getType()=='IntegerType' and \
	   # self.RHS.getType()=='RealType'):
            return 'Void'
	elif self.LHS.getType()=='RealType' and \
	   self.RHS.getType()=='IntegerType':
	    #type coercion
	    #newExp=ConversionExp([self.RHS])
	    #self.RHS=newExp
	    return 'Void'
	elif self.LHS.getType()=='IntegerType' and \
	   self.RHS.getType()=='RealType':
	    #type coercion
	    #newExp=ConversionExp([self.LHS])
	    #self.LHS=newExp
	    return 'Void'
        return 'TypeError'

    def __str__(self):
	result=str(self.LHS) + ' = ' + str(self.RHS)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	#result=result+';'
	return result

    def dump(self, indentLevel):
	result=indent*indentLevel
	result=result+str(self.LHS) + ' = ' + str(self.RHS)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result


class ConnectEqStm(AtomicEqStm):
    def __init__(self, lineNumP, colNumP, firstP, secondP, commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.first=firstP #ConnectRef
	self.second=secondP #ConnectRef
	self.comment=commentP # CommentExp

    def setOuterScope(self, st):
        pass

    def getFirstConnector(self):
	return self.first

    def getSecondConnector(self):
	return self.second

    def accept(self, v):
        v.visitConnectEqStm(self)

    def getType(self):
        pass

    def __str__(self):
        result='connect(' + str(self.first)+', '+str(self.second)+')'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	#result=result+';'
	return result

    def dump(self, indentLevel):
        result=indent*indentLevel
	result=result+'connect(' + str(self.first)+', '+str(self.second)+')'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result


class FunCallEqStm(AtomicEqStm):
    def __init__(self, lineNumP, colNumP, funCallP, commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.funCall=funCallP # Exp::FunCallExp
	self.comment=commentP # CommentExp

    def setOuterScope(self, st):
        pass

    def getFunCall(self):
	return self.funCall

    def accept(self, v):
        v.visitFunCallEqStm(self)

    def getType(self):
        pass

    def __str__(self):
        result=str(self.funCall) 
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	#result=result+';'
	return result

    def dump(self, indentLevel):
        result=indent*indentLevel
	result=result+str(self.funCall)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result


#---------------Composite Equation--------------------
# abstracrt class
class CompositeEqStm(EquationStm):
    def __init__(self):
        raise "Abstract class CompositeEqStm can not be instantiated!"
    
class ConditionalEqStm(CompositeEqStm):
    def __init__(self, lineNumP, colNumP, ifTestP, ifBodyP,\
		elseifListP=[], elseBodyP=[], commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.ifTest=ifTestP
	self.ifBody=ifBodyP # list[] of EquationStm
	self.elseifList=elseifListP
	# list of ConditionalEqStm with elseifList=None and elseBody=None
	self.elseBody=elseBodyP # list[] of EquationStm
	self.comment=commentP # CommentExp

    def getIfTest(self):
	return self.ifTest

    def getIfBody(self):
	return self.ifBody

    def getElseifList(self):
	return self.elseifList

    def getElseBody(self):
	return self.elseBody

    def setOuterScope(self, st):
        pass

    def accept(self, v):
        v.visitConditionalEqStm(self)

    def getType(self):
	if not self.ifTest.getType()=='BooleanType':
	    return 'TypeError'
	if not self.ifBody==None:
	    for i in self.ifBody:
	    	if not i.getType()=='Void':
		    return 'TypeError'
	if not self.elseifList==None:
	    for i in self.elseifList:
		if not i.getType()=='Void':
		    return 'TypeError'
	if not self.elseBody==None:
	    for i in self.elseBody:
		if not i.getType()=='Void':
		    return 'TypeError'
	return 'Void'

    def __str__(self):
	result='if '+str(self.ifTest)+' then\n'
	for i in self.ifBody:
	    result=result+str(i)+';\n'
	if len(self.elseifList)>0:
	    for j in self.elseifList:
		result=result+'else'+str(j)[0:len(str(j))-6]
	if len(self.elseBody)>0:
	    result=result+'else\n'
	    for k in self.elseBody:
		result=result+str(k)+';\n'
	result=result+'end if'
        if not self.comment==None:
            result=result+' '+str(self.comment)	
	#result=result+';'
	
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
	result=result+'if '+str(self.ifTest)+' then\n'
	for i in self.ifBody:
	    result=result+i.dump(indentLevel+1)+';\n'
	if len(self.elseifList)>0:
	    for j in self.elseifList:
		temp=j.dump(indentLevel)
		result=result+indent*indentLevel+'else'+ \
			temp[len(indent)*(indentLevel):\
			len(temp)-len(indent)*(indentLevel)-6]
			# -------might need modify the index here
	if len(self.elseBody)>0:
	    result=result+indent*indentLevel+'else\n'
	    for k in self.elseBody:
		result=result+indent*indentLevel+k.dump(indentLevel)+ \
			';\n'
	result=result+indent*indentLevel+'end if'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result


class ForEqStm(CompositeEqStm):
    def __init__(self, lineNumP, colNumP, forIndicesP, bodyP, \
		commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.forIndices=forIndicesP # Exp::ForIndicesExp
        self.body=bodyP # list of EquationStm
        self.comment=commentP # CommentExp
	self.scope=SymbolTable()

    def getForIndices(self):
	return self.forIndices

    def getBody(self):
	return self.body

    def getScope(self):
        return self.scope

    def setOuterScope(self, st):
        self.scope.setOuter(st)

    def accept(self, v):
        v.visitForEqStm(self)

    def getType(self):
	# what is type of RangeExp?
	pass

    def __str__(self):
	result=str(self.forIndices)+'\n'
	for i in self.body:
	    result=result+str(i)+';\n'
	result=result+'end for'
        if not self.comment==None:
            result=result+' '+str(self.comment)
	#result=result+';'
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
	result=result+str(self.forIndices)+'\n'
	for i in self.body:
	    result=result+i.dump(indentLevel+1)+';\n'
	result=result+indent*indentLevel+'end for'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result


class WhenEqStm(CompositeEqStm):
    def __init__(self, lineNumP, colNumP, whenTestP, bodyP, \
		elsewhenListP=[], commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.whenTest=whenTestP
	self.body=bodyP # list of EquationStm
	self.elsewhenList=elsewhenListP
	# list of whenEqStm with elsewhenList=[]
	self.comment=commentP # CommentExp

    def getWhenTest(self):
	return self.whenTest

    def getBody(self):
	return self.body

    def getElsewhenList(self):
	return self.elsewhenList

    def setOuterScope(self, st):
        pass

    def accept(self, v):
        v.visitWhenEqStm(self)

    def getType(self):
	# TBH
	pass

    def __str__(self):
        result='when '+str(self.whenTest)+' then\n'
        for i in self.body:
            result=result+str(i)+'\n'
        if len(self.elsewhenList)>0:
            for j in self.elsewhenList:
                result=result+'else'+str(j)[0:len(str(j))-8]
        result=result+'end when'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        #result=result+';'
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
        result=result+'when '+str(self.whenTest)+' then\n'
	for i in self.body:
	    result=result+i.dump(indentLevel+1)+';\n'
	if len(self.elsewhenList)>0:
	    for j in self.elsewhenList:
		#result=result+indent*indentLevel+'else'+ \
		#	j.dump(indentLevel+1)[0:len(str(j))-9]
		temp=j.dump(indentLevel)
                result=result+indent*indentLevel+'else'+ \
                        temp[len(indent)*(indentLevel):\
                        len(temp)-len(indent)*(indentLevel)-8]
	result=result+indent*indentLevel+'end when'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result

    def accept(self, v):
        v.visitWhenEqStm(self)


#-----------------AlgorithmPart---------------------------------

# abstracrt class
class AlgorithmStm:
    def __init__(self, lineNumP, colNumP):
        raise "Abstract class AlgorithmStm can not be instantiated!"

#---------------Atomic AlgStm--------------------
# abstracrt class
class AtomicAlgStm(AlgorithmStm):
    def __init__(self):
        raise "Abstract class AtomicAlgStm can not be instantiated!"

class AtomicAlgStm1(AtomicAlgStm):
    def __init__(self, lineNumP, colNumP, compRefP, expP, \
		funCallP=None, commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.compRef=compRefP # CompRefExp
	self.exp=expP # Exp
        self.funCall=funCallP # FunArgsExp
        self.comment=commentP # CommentExp


    def accept(self, v):       #Added by Hongyan Song, Mar 27, 2006
        v.visitAtomicAlgStm1(self)


    def __str__(self):
	result=str(self.compRef)
	if not self.exp==None:
	    result=result+' := '+str(self.exp)
	elif not self.funCall==None:
	    result=result+'('+str(self.funCall)+')'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        return result

    def dump(self, indentLevel):
        #result=indent*indentLevel
        #result=result+str(self.LHS) + ' = ' + str(self.RHS)
        #if not self.comment==None:
        #    result=result+' '+str(self.comment)
        #return result
	return indent*indentLevel+str(self)

class AtomicAlgStm2(AtomicAlgStm):
    def __init__(self, lineNumP, colNumP, expListP, compRefP, \
                funCallP, commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
	self.expList=expListP # ExpListExp(kind='PAR') '('
        self.compRef=compRefP # CompRefExp
        self.funCall=funCallP # FunArgsExp
        self.comment=commentP # CommentExp

    def __str__(self):
        result='('+str(self.expList)+')'+' := '+str(self.compRef)+ \
		'('+str(self.funCall)+')'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        return result

    def dump(self, indentLevel):
        #result=indent*indentLevel
        #result=result+str(self.LHS) + ' = ' + str(self.RHS)
        #if not self.comment==None:
        #    result=result+' '+str(self.comment)
        #return result
        return indent*indentLevel+str(self)


#---------------Composite AlgStm--------------------
# abstracrt class
class CompositeAlgStm(AlgorithmStm):
    def __init__(self):
        raise "Abstract class CompositeAlgStm can not be instantiated!"

class ConditionalAlgStm(CompositeAlgStm):
    def __init__(self, lineNumP, colNumP, ifTestP, ifBodyP,\
                elseifListP=[], elseBodyP=[], commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.ifTest=ifTestP
        self.ifBody=ifBodyP
        self.elseifList=elseifListP
        # list of ConditionalAlgStm with elseifList=[] and elseBody=[]
        self.elseBody=elseBodyP
        self.comment=commentP # CommentExp

    def __str__(self):
        result='if '+str(self.ifTest)+' then\n'
        for i in self.ifBody:
            result=result+str(i)+'\n'
        if len(self.elseifList)>0:
            for j in self.elseifList:
                result=result+'else'+str(j)[0:len(str(j))-7]
        if len(self.elseBody)>0:
            result=result+'else\n'
            for k in self.elseBody:
                result=result+str(k)+'\n'
        result=result+'end if'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        result=result+';'
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
        result=result+'if '+str(self.ifTest)+' then\n'
        for i in self.ifBody:
            result=result+i.dump(indentLevel+1)+';\n'
        if len(self.elseifList)>0:
            for j in self.elseifList:
                temp=j.dump(indentLevel)
                result=result+indent*indentLevel+'else'+ \
                        temp[len(indent)*(indentLevel):\
                        len(temp)-len(indent)*(indentLevel)-6]
                        # -------might need modify the index here
        if len(self.elseBody)>0:
            result=result+indent*indentLevel+'else\n'
            for k in self.elseBody:
                result=result+indent*indentLevel+k.dump(indentLevel)+ \
                        ';\n'
        result=result+indent*indentLevel+'end if'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        return result

class ForAlgStm(CompositeAlgStm):
    def __init__(self, lineNumP, colNumP, forIndicesP, bodyP, \
                commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.forIndices=forIndicesP # Exp::ForIndicesExp
        self.body=bodyP # list of AlgorithmStm
        self.comment=commentP # CommentExp
	self.scope=SymbolTable()

    def getScope(self):
        return self.scope

    def setOuterScope(self, st):
        self.scope.setOuter(st)

    def getForIndices(self):
        return self.forIndices

    def getBody(self):
        return self.body

    def __str__(self):
        result=str(self.forIndices)+'\n'
        for i in self.body:
            result=result+str(i)+';\n'
        result=result+'end for'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        #result=result+';'
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
        result=result+str(self.forIndices)+'\n'
        for i in self.body:
            result=result+i.dump(indentLevel+1)+';\n'
        result=result+indent*indentLevel+'end for'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        return result

class WhileAlgStm(CompositeAlgStm):
    def __init__(self, lineNumP, colNumP, whileTestP, bodyP, \
		commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.whileTest=whileTestP # Exp
	self.body=bodyP # list of AlgorithmStm
	self.comment=commentP # CommentExp

    def __str__(self):
	result='while '+str(self.whileTest)+' loop\n'
        #result=str(self.forIndices)+'\n'
        for i in self.body:
            result=result+str(i)+';\n'
        result=result+'end while'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        #result=result+';'
        return result

    def accept(self, v):
        return v.visitWhileAlgStm(self)
        #TODO huh?
        #return ""

    def dump(self, indentLevel):
        result=indent*indentLevel
	result=result+'while '+str(self.whileTest)+' loop\n'
        #result=result+str(self.forIndices)+'\n'
        for i in self.body:
            result=result+i.dump(indentLevel+1)+';\n'
        result=result+indent*indentLevel+'end while'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        return result

class WhenAlgStm(CompositeAlgStm):
    def __init__(self, lineNumP, colNumP, whenTestP, bodyP, \
                elsewhenListP=[], commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.whenTest=whenTestP
        self.body=bodyP # list of AlgorithmStm
        self.elsewhenList=elsewhenListP
        # list of whenAlgStm with elsewhenList=[]
        self.comment=commentP # CommentExp

    def __str__(self):
        result='when '+str(self.whenTest)+' then\n'
        for i in self.body:
            result=result+str(i)+';\n'
        if len(self.elsewhenList)>0:
            for j in self.elsewhenList:
                result=result+'else'+str(j)[0:len(str(j))-8]
        result=result+'end when'
        if not self.comment==None:
            result=result+' '+str(self.comment)
        #result=result+';'
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel
        result=result+'when '+str(self.whenTest)+' then\n'
        for i in self.body:
            result=result+i.dump(indentLevel+1)+';\n'
        if len(self.elsewhenList)>0:
            for j in self.elsewhenList:
                temp=j.dump(indentLevel)
                result=result+indent*indentLevel+'else'+ \
                        temp[len(indent)*(indentLevel):\
                        len(temp)-len(indent)*(indentLevel)-8]
        result=result+indent*indentLevel+'end when'
        if not self.comment==None:
            result=result+' '+str(self.comment)
	return result

