# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the AST module

import copy
from AST.ASTComponent import EqualsModification, ClassModification

class ComponentPrefix:
    def __init__(self, isFinalP=False, innerOuterP=None, \
		isReplaceableP=False):
	self.isFinal=isFinalP # Boolean
	self.innerOuter=innerOuterP # Enum{None, 'INNER', 'OUTER'}
        self.isReplaceable=isReplaceableP # Boolean

    def getFinal(self):
	return self.isFinal

    def getInnerOuter(self):
	return self.innerOuter

    def getReplaceable(self):
	return self.isReplaceable

class TableEntry:
    'abstract class'
    lineNum=-1
    colNum=-1
    name=None
    definition=None

    def __init__(self):
	raise "Abstract class TableEntry can not be instantiated!"

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        self.colNum

    def getName(self):
        return self.name

    def getDefinition(self):
        return self.definition

class DefinitionEntry(TableEntry):
    def __init__(self, lineNumP, colNumP, nameP, definitionP, \
		visibilityP): #, originalScopeP):#, modifP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.name=nameP # name: ASTExp::IdentExp
	self.definition=definitionP # ASTDefinition::ClassDefinition
	self.visibility=visibilityP
	#self.originalScope=originalScopeP # ASTScope::SymbolTable
	self.modif=[] # modification in ShortClassDefinition

    def getName(self):
        return self.name

    def getDefinition(self):
        return self.definition

    #def getOriginalScope(self):
	#return self.originalScope

    def setDefinition(self, d):
        self.definition=d

    def setVisibility(self, value):
	self.visibility=value

    def getModif(self):
        return self.modif

    def setModif(self, m):
        self.modif=m

    def hasModif(self):
        return len(self.modif)>=1

    def addModif(self, m):
        self.modif.append(m)

    def __str__(self):
        return str(self.name)

    def dump(self):
	return str(self.name)+'(:'+str(self.definition.getKind())+')'

    def accept(self, v):
        v.visitDefinitionEntry(self)

class DeclarationEntry(TableEntry):
    def __init__(self, lineNumP, colNumP, nameP, definitionP, \
		componentPrefixP, typePrefixP, typeP, typeArrSubP, \
		nameArrSubP): #, modifP): #, valueP, ):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.name=nameP # name: ASTExp::IdentExp
	self.definition=definitionP # ASTDefinition::ClassDefinition
	#self.originalScope=orgScopeP # ASTScope::SymbolTable
	self.componentPrefix=componentPrefixP # ComponentPrefix
	self.typePrefix=typePrefixP # TypePrefix
	self.type=typeP # type: ASTExp::NameExp
	self.typeArrSub=typeArrSubP
	self.nameArrSub=nameArrSubP
	self.modif=[] # list []
	#self.value=valueP
	self.env={}
	self.instance=None # ASTScope::ClassInstance

    def getComponentPrefix(self):
	return self.componentPrefix

    def getTypePrefix(self):
	return self.typePrefix

    def getType(self):
	return self.type

    def getName(self):
	return self.name

    def getDefinition(self):
	return self.definition

    #def getOriginalScope(self):
    #    return self.originalScope

    def setDefinition(self, d):
	self.definition=d

    def setVisibility(self, value):
	self.typePrefix.setVisibility(value)

    def isParameter(self):
	if self.typePrefix.getVariability()=='PARAMETER':
	    return True
	else:
	    return False

    def isConstant(self):
	if self.typePrefix.getVariability()=='CONSTANT':
	    return True
        else:
            return False

    def isFlow(self):
	return self.typePrefix.getFlow()
	    
    def getModif(self):
	return self.modif

    def setModif(self, m):
	self.modif=m

    def hasModif(self):
	return len(self.modif)>=1

    def addModif(self, m):
	self.modif.append(m)

    #def getValue(self):
	#return self.value

    #def setValue(self, v):
	#self.value=v

    def getEnv(self):
	return self.env

    def setEnv(self, e):
	self.env=e

    def createEnv(self):
	if self.getDefinition()==None:
	    return
	for i in self.getDefinition().getScope().getDeclarations():
	    self.env[str(i)]=None

    def addToEnv(self, id, value):
	self.env[id]=value

    def getInstance(self):
	return self.instance
 
    def hasInstance(self):
	return not self.instance==None

    def setInstance(self, i):
	self.instance=i

    # deep copy
    def deepCopy(self):
	lineNumP=self.lineNum
	colNumP=self.colNum
	nameP=copy.deepcopy(self.name)
	#nameP=self.name
	definitionP=self.definition
	#originalScopeP=self.originalScope
	componentPrefixP=copy.deepcopy(self.componentPrefix)
	typePrefixP=copy.deepcopy(self.typePrefix)
	typeP=self.type
	typeArrSubP=self.typeArrSub
	nameArrSubP=self.nameArrSub
	modifP=copy.deepcopy(self.modif)
	#modifP=self.modif
	envP=copy.deepcopy(self.env)
	result=DeclarationEntry(lineNumP, colNumP, nameP, definitionP, \
		#originalScopeP, 
		componentPrefixP, typePrefixP, typeP, \
		typeArrSubP, nameArrSubP)#, modifP)#, valueP)
	result.setModif(modifP)
	result.setEnv(envP)
	return result	

    def __str__(self):
	return str(self.name)

    def dump(self):
	result=''
	if self.typePrefix.getVisibility()=='PROTECTED':
	    result='protected '
	result=result+str(self.typePrefix)+str(self.type)+\
		' '+str(self.name)
	if self.hasModif():
	    result=result+'('
	    for i in self.modif:
	    	if isinstance(i, EqualsModification):
	    	    result=result+'self'+str(i)+', '
	    	elif isinstance(i, ClassModification):
		    if not i.getEqualsModification()==None:
			for j in i.getArgumentList():
		    	    #result=result+str(i)[1:len(str(i))-1]+', '
			    result=result+str(j)+', '
			result=result+'self='+\
				str(i.getEqualsModification())+', '
		    else:
			result=result+str(i)[1:len(str(i))-1]+', '
	#if len(self.modif)>0:
	    result=result[0:len(result)-2]+')'
	return result+';'

    def dumpEnv(self):
	result=''
        if self.typePrefix.getVisibility()=='PROTECTED':
            result='protected '
        result=result+str(self.typePrefix)+str(self.type)+\
                ' '+str(self.name)
	#if not self.env==None:
	result=result+';'
	for i in self.env.keys():
	    if not self.env[i]==None:
	    	result=result+'\n    //'+i+'='+str(self.env[i])
        return result

    def accept(self, v):
	v.visitDeclarationEntry(self)

class SymbolTable:
    def __init__(self, outerP=None):
	self.table={}
	self.outer=outerP

    def setOuter(self, outerP):
        self.outer=outerP # outer is a SymbolTable

    def getOuter(self):
        return self.outer

    def getTable(self):
        return self.table

    def __str__(self):
	return str(self.table.keys())

    # ident: String
    # entry: ASTScope::TableEntry
    # return: Flase if there already exists an element with the same name
    def add(self, ident, entry):
	if ident in self.table.keys():
	    return False
	else:
	    self.table[ident]=entry
	    return True

    def remove(self, ident):
	del self.table[ident]

    def getPublicDeclarations(self):
	result=[]
	tableKeys=self.table.keys()
	for i in tableKeys:
	    tableEntry=self.table[i]
	    if isinstance(tableEntry, DeclarationEntry):
		if (tableEntry.typePrefix.getVisibility()=='DEFAULT') or \
		   (tableEntry.typePrefix.getVisibility()=='PUBLIC'):
		    result.append(tableEntry)
	return result

    def getDeclarations(self):
	result=[]
	tableKeys=self.table.keys()
	for i in tableKeys:
	    tableEntry=self.table[i]
	    if isinstance(tableEntry, DeclarationEntry):
		result.append(tableEntry)
	return result

    def getDefinitions(self):
	result=[]
        tableKeys=self.table.keys()
        for i in tableKeys:
            tableEntry=self.table[i]
            if isinstance(tableEntry, DefinitionEntry):
                result.append(tableEntry)
        return result

    def getParameterDecl(self):
	result=[]
	tableKeys=self.table.keys()
	for i in tableKeys:
	    tableEntry=self.table[i]
	    if isinstance(tableEntry, DeclarationEntry):
		if tableEntry.isParameter():
		    result.append(tableEntry)
	return result

    def getConstantDecl(self):
	result=[]
        tableKeys=self.table.keys()
        for i in tableKeys:
	    tableEntry=self.table[i]
	    if isinstance(tableEntry, DeclarationEntry):
		if tableEntry.isConstant():
		    result.append(tableEntry)
	return result

    def getAllElems(self):
	result=[]
	tableKeys=self.table.keys()
	for i in tableKeys:
	    tableEntry=self.table[i]
	    result.append(tableEntry)
	return result

    # ident: String
    # look up in current and all outer scopes
    def lookup(self, ident):
	found=self.table.has_key(ident)
	if found:
	    return self.table[ident]
	outerScope=self.outer
        while not outerScope==None:
	    found=outerScope.getTable().has_key(ident)
	    if found:
		return outerScope.getTable()[ident]
	    else:
		outerScope=outerScope.getOuter()
	return None

    # ident: String
    # look up in current scope only
    def localLookup(self, ident):
	found=self.table.has_key(ident)
        if found:
            return self.table[ident]
	return None

    def accept(self, v):
	v.visitSymbolTable(self)


class ClassInstance:
    def __init__(self): #, symbolTableP, initEqListP, eqListP):
	# symbolTable in ClassInstance only contains DeclarationEntrys
	self.declTable={} 
	self.initEqList=[]
	self.eqList=[]

    def getDeclTable(self):
	return self.declTable

    def getInitEqs(self):
	return self.initEqList

    def getEqs(self):
	return self.eqList

    def getDeclarations(self):
        result=[]
        tableKeys=self.declTable.keys()
        for i in tableKeys:
            result.append(self.declTable[i])
        return result

    def getPublicDeclarations(self):
        result=[]
        tableKeys=self.declTable.keys()
        for i in tableKeys:
            tableEntry=self.declTable[i]
            if isinstance(tableEntry, DeclarationEntry):
                if (tableEntry.typePrefix.getVisibility()=='DEFAULT') or \
                   (tableEntry.typePrefix.getVisibility()=='PUBLIC'):
                    result.append(tableEntry)
        return result

    def addDeclarationEntry(self, name, entry):
	self.declTable[name]=entry

    def addInitEquation(self, eq):
	self.initEqList.append(eq)

    def addEquation(self, eq):
	self.eqList.append(eq)
   
    def __str__(self):
	result=''
	for i in self.getDeclarations():
	    result=result+'  '+i.dump()+'\n'
	if len(self.getInitEqs())>0:
	    result=result+'initial equations\n'
	    for i in self.getInitEqs():
		result=result+'  '+str(i)+';\n'
	if len(self.getEqs())>0:
	    result=result+'equations\n'
	    for i in self.getEqs():
		result=result+'  '+str(i)+';\n'
	return result
