# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

class Node:

    id=None
    outgoingEdgesInG=None
    outgoingEdgesInL=None
    isIsolated=True
    visited=False
    level=-10000000

    def __init__(self, idP):
	self.id=idP
	self.outgoingEdgesInG={}
	self.outgoingEdgesInL={}

    def getId(self):
	return self.id

    def isVisited(self):
	return self.visited

    def setVisited(self, value):
	self.visited=value

    def setIsolated(self, value):
	self.isIsolated=value

    def hasKeyInG(self, key):
	if key in self.outgoingEdgesInG.keys():
	    return True
	return False

    def setLevel(self, l):
	self.level=l

    def addOutgoingEdgeInG(self, idP, edgeP):
	self.outgoingEdgesInG[idP]=edgeP

    def getOutgoingEdgeInG(self, idP):
	return self.outgoingEdgesInG[idP]

    def deleteOutgoingEdgeInG(self, idP):
        del self.outgoingEdgesInG[idP]

    def getAllOutgoingEdgesInG(self):
	result=[]
        keys=self.outgoingEdgesInG.keys()
        for i in keys:
            e=self.outgoingEdgesInG[i]
            result.append(e)
        return result

    def getAllSaturatedEdgesInG(self):
        result=[]
        keys=self.outgoingEdgesInG.keys()
        for i in keys:
            e=self.outgoingEdgesInG[i]
            if e.getCap()-e.getFlow()==0 and e.getFlow()!=0:
                result.append(e)
        return result

    def getAllUnsaturatedEdgesInG(self):
	result=[]
	keys=self.outgoingEdgesInG.keys()
	for i in keys:
	    e=self.outgoingEdgesInG[i]
	    if e.getCap()-e.getFlow()>0:
		result.append(e)
	return result

    def addOutgoingEdgeInL(self, idP, edgeP):
        self.outgoingEdgesInL[idP]=edgeP

    def getOutgoingEdgeInL(self, idP):
        return self.outgoingEdgesInL[idP]

    def deleteOutgoingEdgeInL(self, idP):
	del self.outgoingEdgesInL[idP]

    def deleteAllOutgoingEdgesInL(self):
	self.outgoingEdgesInL={}

    def getAllOutgoingEdgesInL(self):
        result=[]
        keys=self.outgoingEdgesInL.keys()
        for i in keys:
            e=self.outgoingEdgesInL[i]
            result.append(e)
        return result

    def getAllUnsaturatedEdgesInL(self):
        result=[]
        keys=self.outgoingEdgesInL.keys()
        for i in keys:
            e=self.outgoingEdgesInL[i]
            if e.getCap()-e.getFlow()>0:
                result.append(e)
        return result

    def __str__(self):
	return self.id

    def dumpL(self):
        return self.id+'('+str(self.level)+'): '+\
		str(self.outgoingEdgesInL.keys())

class EqNode(Node):

    eq=None

    def __init__(self, idP, eqP):
	self.id=idP
	self.eq=eqP # ASTEquations::RegularEqStm
	self.outgoingEdgesInG={}
	self.outgoingEdgesInL={}

    def getEq(self):
	return self.eq

    def deleteAllOutgoingEdgesInG(self):
	self.outgoingEdgesInG={}

    def __str__(self):
	return self.id +': '+str(self.eq)


class VarNode(Node):

    # DeclarationEntry in symbol table
    definition=None
    hasDerivativeForm=False
    integralCausality=False

    def __init__(self, idP, defP):
	self.id=idP
	self.definition=defP # ASTScope::DeclarationEntry
	# container for outgoing edges in graph
	self.outgoingEdgesInG={}
	# container for outgoing edges in level graph
	self.outgoingEdgesInL={}

    def getDefinition(self):
	return self.definition

    def dumpL(self):
	if self.integralCausality==False:
            return self.id+'('+str(self.level)+'): '+\
                str(self.outgoingEdgesInL.keys())
	else:
	    return 'der('+self.id+')'+'('+str(self.level)+'): '+\
                str(self.outgoingEdgesInL.keys())

    def __str__(self):
	if self.hasDerivativeForm and self.integralCausality:
	    return 'der('+self.id+')'
	else:
	    return self.id

class DNode:
    "Node in dependency graph for loop detection"
    # A DNode uses the same id as its varNode
    id=None # string
    varNode=None # VarNode
    eqNode=None # EqNode
    # container for adjacent nodes(DNode)
    dependencies=None # {}
    # container for adjacent nodes in reversed dependency graph
    reverseDependencies=None # {}
    visited=False
    orderNumber=0

    def __init__(self, idP, varNodeP, eqNodeP):
        self.id=idP
	self.varNode=varNodeP
	self.eqNode=eqNodeP
	self.dependencies={}
        self.reverseDependencies={}

    def getId(self):
	return self.id

    def getVarNode(self):
	return self.varNode

    def getEqNode(self):
	return self.eqNode

    def addDependency(self, idP, dNodeP):
        self.dependencies[idP]=dNodeP

    def getDependency(self, idP):
	if idP in self.dependencies.keys():
            return self.dependencies[idP]
	return None

    # return all dependent dNodes
    def getAllDependencies(self):
        result=[]
        keys=self.dependencies.keys()
        for i in keys:
            dNode=self.dependencies[i]
            result.append(dNode)
        return result

    def addReverseDependency(self, idP, dNodeP):
        self.reverseDependencies[idP]=dNodeP

    # return all adjacent dNodes in reverse graph
    def getAllReverseDependencies(self):
        result=[]
        keys=self.reverseDependencies.keys()
        for i in keys:
            dNode=self.reverseDependencies[i]
            result.append(dNode)
        return result

    def isVisited(self):
        return self.visited

    def setVisited(self, value):
        self.visited=value

    def getOrderNumber(self):
        return self.orderNumber

    def setOrderNumber(self, n):
        self.orderNumber=n

    def __str__(self):
	return str(self.varNode)+'---'+str(self.eqNode)
