from AST.ASTDefinition import *
from AST.ASTComponent import *
from AST.ASTEquations import *
from AST.ASTExp import *
from AST.ASTScope import *

EOF_ = 1
ALGORITHM = 4
AND = 5
ANNOTATION = 6
BLOCK = 7
BOUNDARY = 8
CLASS = 9
CONNECT = 10
CONNECTOR = 11
CONSTANT = 12
DISCRETE = 13
EACH = 14
ELSE = 15
ELSEIF = 16
ELSEWHEN = 17
END = 18
ENUMERATION = 19
EQUATION = 20
ENCAPSULATED = 21
EXTENDS = 22
EXTERNAL = 23
FALSE = 24
FINAL = 25
FLOW = 26
FOR = 27
FUNCTION = 28
IF = 29
IMPORT = 30
IN = 31
INITIAL = 32
INNER = 33
INPUT = 34
LOOP = 35
MODEL = 36
NOT = 37
OUTER = 38
OR = 39
OUTPUT = 40
PACKAGE = 41
PARAMETER = 42
PARTIAL = 43
PROTECTED = 44
PUBLIC = 45
RECORD = 46
REDECLARE = 47
REPLACEABLE = 48
RESULTS = 49
THEN = 50
TRUE = 51
TYPE = 52
UNSIGNED_REAL = 53
WHEN = 54
WHILE = 55
WITHIN = 56
LPAR = 57
RPAR = 58
LBRACK = 59
RBRACK = 60
LBRACE = 61
RBRACE = 62
EQUALS = 63
ASSIGN = 64
PLUS = 65
MINUS = 66
STAR = 67
SLASH = 68
DOT = 69
COMMA = 70
LESS = 71
LESSEQ = 72
GREATER = 73
GREATEREQ = 74
EQEQ = 75
LESSGT = 76
COLON = 77
SEMICOLON = 78
POWER = 79
WS = 80
ML_COMMENT = 81
ML_COMMENT_CHAR = 82
SL_COMMENT = 83
IDENT = 84
NONDIGIT = 85
DIGIT = 86
EXPONENT = 87
UNSIGNED_INTEGER = 88
STRING = 89
SCHAR = 90
SESCAPE = 91
ESC = 92

ALGORITHM_STATEMENT = 93
ARGUMENT_LIST = 94
CLASS_DEFINITION = 95
CLASS_MODIFICATION = 96
COMMENT = 97
DECLARATION = 98
DEFINITION = 99
ENUMERATION_LITERAL = 100
ELEMENT = 101
ELEMENT_MODIFICATION = 102
ELEMENT_REDECLARATION = 103
EQUATION_STATEMENT = 104
INITIAL_EQUATION = 105
INITIAL_ALGORITHM = 106
EXPRESSION_LIST = 107
EXTERNAL_FUNCTION_CALL = 108
FOR_INDICES = 109
FUNCTION_CALL = 110
INITIAL_FUNCTION_CALL = 111
FUNCTION_ARGUMENTS = 112
NAMED_ARGUMENTS = 113
QUALIFIED = 114
RANGE2 = 115
RANGE3 = 116
STORED_DEFINITION = 117
STRING_COMMENT = 118
UNARY_MINUS = 119
UNARY_PLUS = 120
UNQUALIFIED = 121

class ASTbuilder:
    def __init__(self, parseTree, filenameP, dir):
	self.data=parseTree
	self.filename=filenameP
	self.dir=dir #String
	self.ast=None


    def go(self):
	self.ast=self.makeClassFile()
	return self.ast

    def makeClassFile(self):
	#print 'makeClassFile(): ', self.filename
	firstElem=self.data.getElement("0")
	withinClause=None
	classDefinitions=[]
	localRef=self.getChild(firstElem)
	if self.getNodeTypeValue(localRef)==WITHIN: # within
            withinClause=self.makeWithinClause(localRef)
            localRef=self.getSibling(localRef)
	while not localRef==None:
	    # 'FINAL' and 'CLASS_DEFINITION' are siblings
	    if self.getNodeTypeValue(localRef)==FINAL:
		localRef=self.getSibling(localRef)
		classDefinition=self.makeClassDefinition(localRef)
		classDefinition.setFinal(True)
		classDefinitions.append(classDefinition)
	    elif self.getNodeTypeValue(localRef)==CLASS_DEFINITION:
		classDefinition=self.makeClassDefinition(localRef)
		classDefinitions.append(classDefinition)
	    localRef=self.getSibling(localRef)
	return ClassFile(self.filename, self.dir, withinClause, \
			classDefinitions)

    def makeWithinClause(self, elem):
	#print 'makeWithinClause()'
	child=self.getChild(elem)
	name=self.makeNameExp(child)
	return WithinClause(-1, -1, name)


    # ------------------------ ClassDefinition -----------
    # input node: 'CLASS_DEFINITION'
    def makeClassDefinition(self, elem):
        if elem==None:
	    return None
	#localRef=elem
	localRef=self.getChild(elem)
	#if self.getNodeTypeValue(localRef)==REPLACEABLE:
	#    return self.makeReplaceableClause(elem)
	isEncapsulated=False
	isPartial=False
	if self.getNodeTypeValue(localRef)==ENCAPSULATED:
	    isEncapsulated=True
	    localRef=self.getSibling(localRef)
	if self.getNodeTypeValue(localRef)==PARTIAL:
	    isPartial=True
	    localRef=self.getSibling(localRef)
	classNameNode=self.getSibling(localRef)
	next=self.getSibling(classNameNode)
	if self.getNodeTypeValue(next)==EQUALS:
            # Enumeration or ClassModifier
            child=self.getChild(next)
            if self.getNodeTypeValue(child)==ENUMERATION:
                tempClsDef=self.makeEnumClassDefinition(localRef)
	    else: # ShortClassDefinition
                tempClsDef=self.makeShortClassDefinition(localRef)
        else:
            # RegularClassDefinition
            tempClsDef=self.makeRegularClassDefinition(localRef)
	tempClsDef.setEncapsulated(isEncapsulated)
	tempClsDef.setPartial(isPartial)
	return tempClsDef

    # input Node: keyword 'class', 'model' ...
    def makeRegularClassDefinition(self, elem):
	#print 'makeClass()'
	prefix=ClassPrefix()
	stringComment=None
	#publicElementList=[]
	#protectedElementList=[]
	elementList=[]
	eqList=[]
	initEqList=[]
	algList=[]
	initAlgList=[]
	kind=self.getNodeTextValue(elem)
	nextSib=self.getSibling(elem)
	name=IdentExp(-1, -1, self.getNodeTextValue(nextSib))
	allSiblings=self.getAllSiblings(nextSib) # a list []
	#for i in allSiblings:
	#    print self.getNodeTextValue(i)
	#empty class
	if len(allSiblings)==0:
	    return RegularClassDefinition(-1, -1, prefix, kind, name, \
				stringComment)
	firstNodeType=self.getNodeTypeValue(allSiblings[0])
	# regular class definition
	for node in allSiblings:
	    typeValue=self.getNodeTypeValue(node)
	    if typeValue==PUBLIC:
		tempList=self.makePublicElements(node)
		for i in tempList:
		    #publicElementList.append(i)
		    i.setVisibility('PUBLIC')
		    elementList.append(i)
	    elif typeValue==PROTECTED:
		tempList=self.makeProtectedElements(node)
		for i in tempList:
		    #protectedElementList.append(i)
		    i.setVisibility('PROTECTED')
		    elementList.append(i)
	    elif typeValue==IMPORT:
		#publicElementList.append(self.makeImportClause(node))
		elementList.append(self.makeImportClause(node))
	    elif typeValue==EXTENDS:
		#publicElementList.append(self.makeExtendsClause(node))
		elementList.append(self.makeExtendsClause(node))
	    elif typeValue==ANNOTATION:
		#raise 'not yet supported in AST: to be implemented...'
		#publicElementList.append(self.makeAnnotationExp(node))
		elementList.append(self.makeAnnotationExp(node))
	    elif typeValue==DECLARATION:
		#print 'call makeDeclaration()...'
		#publicElementList.append(self.makeDeclarationElement(node))
		elementList.append(self.makeDeclarationElement(node))
	    elif typeValue==DEFINITION:
		isFinal=False
		innerOuter=None
		replaceable=False
		localRef=self.getChild(node)
		while not self.getNodeTypeValue(localRef)==CLASS_DEFINITION:
		    if self.getNodeTypeValue(localRef)==FINAL:
			isFinal=True
		    if self.getNodeTypeValue(localRef)==INNER:
			innerOuter='INNER'
		    elif self.getNodeTypeValue(localRef)==OUTER:
			innerOuter='OUTER'
		    if self.getNodeTypeValue(localRef)==REPLACEABLE:
			replaceable=True
		    localRef=self.getSibling(localRef)
		if replaceable:
		    temp=self.makeReplaceableClause(node)
		else:
		    temp=self.makeClassDefinition(localRef)
		    #temp.setReplaceable(True)
		temp.setFinal(isFinal)
		temp.setInnerOuter(innerOuter)
		#publicElementList.append(temp)
		elementList.append(temp)
	    elif typeValue==EQUATION: 
                tempList1=self.makeEqStmList(node) #return a list[]
		if  len(tempList1)>0:
                    for i in tempList1:
                    	eqList.append(i)
	    elif typeValue==ALGORITHM:
                tempList2=self.makeAlgStmList(node) #return a list[]
		if  len(tempList2)>0:
                    for i in tempList2:
                    	algList.append(i)
	    elif typeValue==INITIAL_EQUATION:
                tempList3=self.makeEqStmList(self.getChild(node)) # list[]
		if  len(tempList3)>0:
                    for i in tempList3:
                    	initEqList.append(i)
	    elif typeValue==INITIAL_ALGORITHM:
                tempList4=self.makeAlgStmList(self.getChild(node)) # list[]
		if  len(tempList4)>0:
                    for i in tempList4:
                    	initAlgList.append(i)
	    elif typeValue==STRING_COMMENT:
		stringComment=StringExp(-1,-1, \
			self.getNodeTextValue(self.getChild(node)))
	    else:
		print self.getNodeTextValue(node) 
		raise 'not supported Modelia construct'
	#return RegularClassDefinition(-1, -1, prefix, kind, name, \
        #            publicElementListP=publicElementList, \
	#	    protectedElementListP=protectedElementList, \
	#	    initEquationPartP=initEqList, equationPartP=eqList, \
	#	    initAlgorithmPartP=initAlgList, algorithmPartP=algList)
	return RegularClassDefinition(-1, -1, prefix, kind, name, \
		stringComment, elementListP=elementList, \
		initEquationPartP=initEqList, equationPartP=eqList, \
		initAlgorithmPartP=initAlgList, algorithmPartP=algList)	

    # input Node: keyword 'class', 'model' 'type' ...
    def makeShortClassDefinition(self, elem):
	#print 'makeShortClassDefinition()'
	classPrefix=ClassPrefix()
	kind=self.getNodeTextValue(elem)
	nameNode=self.getSibling(elem)
	className=IdentExp(-1, -1, self.getNodeTextValue(nameNode))
	equalsNode=self.getSibling(nameNode)
	#child=self.getChild(equalsNode)
	#classModifier=self.makeClassModifier(child)
        basePrefix=TypePrefix()
        arrSub=None
        classModification=None
        comment=None
        localRef=self.getChild(equalsNode)
        typeValue=self.getNodeTypeValue(localRef)
        while not (typeValue==IDENT or typeValue==DOT):
            if typeValue==FLOW:
                basePrefix.setFlow(True)
            if typeValue==CONSTANT:
                basePrefix.setVariability('CONSTANT')
            if typeValue==DISCRETE:
                basePrefix.setVariability('DISCRETE')
            if typeValue==PARAMETER:
                basePrefix.setVariability('PARAMETER')
            if typeValue==INPUT:
                basePrefix.setCausality('INPUT')
            if typeValue==OUTPUT:
                basePrefix.setCausality('OUTPUT')
            localRef=self.getSibling(localRef)
            typeValue=self.getNodeTypeValue(localRef)
        name=self.makeNameExp(localRef)
        localRef=self.getSibling(localRef)
        if self.getNodeTypeValue(localRef)==LBRACK:
            arrSub=self.makeArrSubExp(localRef)
            localRef=self.getSibling(localRef)
        if self.getNodeTypeValue(localRef)==CLASS_MODIFICATION:
            classModification=self.makeClassModification(localRef)
            localRef=self.getSibling(localRef) # 'COMMENT' node
        if not localRef==None:
            if not self.getChild(localRef)==None:
                comment=self.makeCommentExp(localRef)
        return ShortClassDefinition(-1, -1, classPrefix, kind, \
		className, basePrefix, name, arrSub, classModification, \
		comment)

    # input node: 'ENUMERATION' (original, to be modified)
    # input Node: keyword 'class', 'model' ...
    def makeEnumClassDefinition(self, elem):
        #print 'makeEnumClassDefinition()'
	prefix=ClassPrefix()
	kind=self.getNodeTextValue(elem)
	nameNode=self.getSibling(elem)
	name=IdentExp(-1, -1, self.getNodeTextValue(nameNode))        
        enumList=[]
        comment=None
        equalsNode=self.getSibling(nameNode)
        enumerationNode=self.getChild(equalsNode)
        allChildren=self.getAllChildren(enumerationNode)
        for node in allChildren:
            if self.getNodeTypeValue(node)==ENUMERATION_LITERAL:
                enumList.append(self.makeEnumLiteral(node))
            elif self.getNodeTypeValue(node)==COMMENT:
		if not self.getChild(node)==None:
        	    comment=self.makeCommentExp(node)
        #return EnumClassDefinition(-1, -1, prefix, kind, name, \
	#	enumList, comment)
	result=EnumClassDefinition(-1, -1, prefix, kind, name, \
		enumList, comment)
	return self.transformEnumClassDefinition(result)

    # input node: 'ENUMERATION_LITERAL'
    def makeEnumLiteral(self, elem):
        #print 'makeEnumLiteral()'
        literal=None
        comment=None
        allChildren=self.getAllChildren(elem)
        for node in allChildren:
            if self.getNodeTypeValue(node)==IDENT:
                literal=IdentExp(-1, -1, self.getNodeTextValue(node))
            elif self.getNodeTypeValue(node)==COMMENT:
		if not self.getChild(node)==None:
                    comment=self.makeCommentExp(node)
        return EnumLiteral(-1, -1, literal, comment)

    # input node: 'IDENT', 'DOT', TypePrefix
    #def makeClassModifier(self, elem):
    #    #print 'makeClassModifier()'
    #    prefix=TypePrefix()
    #    arrSub=None
	#classModification=None
        #comment=None
        #localRef=elem
        #typeValue=self.getNodeTypeValue(localRef)
        #while not (typeValue==IDENT or typeValue==DOT):
        #    if typeValue==FLOW:
        #        prefix.setFlow(True)
        #    if typeValue==CONSTANT:
        #        prefix.setVariability('CONSTANT')
        #    if typeValue==DISCRETE:
        #        prefix.setVariability('DISCRETE')
        #    if typeValue==PARAMETER:
        #        prefix.setVariability('PARAMETER')
        #    if typeValue==INPUT:
        #        prefix.setCausality('INPUT')
        #    if typeValue==OUTPUT:
        #        prefix.setCausality('OUTPUT')
        #    localRef=self.getSibling(localRef)
        #    typeValue=self.getNodeTypeValue(localRef)
        #name=self.makeNameExp(localRef)
        #localRef=self.getSibling(localRef)
        #if self.getNodeTypeValue(localRef)==LBRACK:
        #    arrSub=self.makeArrSubExp(localRef)
        #    localRef=self.getSibling(localRef)
	#if self.getNodeTypeValue(localRef)==CLASS_MODIFICATION:
        #    classModification=self.makeClassModification(localRef)
        #    localRef=self.getSibling(localRef) # 'COMMENT' node
        #if not localRef==None:
	#    if not self.getChild(localRef)==None:
        #        comment=self.makeCommentExp(localRef)
        #return ClassModifier(-1, -1, prefix, name, classModification, \
        #                     arrSub, comment)


    # return a list of declaration/definition
    # input node: 'PUBLIC'
    def makePublicElements(self, elem):
        #print 'makePublicElements(): to be implemented ...'
	result=[]
	allChildren=self.getAllChildren(elem)
	if len(allChildren)==0:
	    return []
	for node in allChildren:    
	    if self.getNodeTypeValue(node)==DECLARATION:
	    	result.append(self.makeDeclarationElement(node))
	    elif self.getNodeTypeValue(node)==DEFINITION:
	    	result.append(self.makeClassDefinition(self.getChild(node)))
            elif self.getNodeTypeValue(node)==IMPORT:
                result.append(self.makeImportClause(node))
            elif self.getNodeTypeValue(node)==EXTENDS:
                result.append(self.makeExtendsClause(node))
	    else:
		raise 'not supported Modelia construct'
        return result

    # input node: 'PROTECTED'
    def makeProtectedElements(self, elem):
        #print 'makeProtectedElements(): to be implemented ...'
        result=[]
        allChildren=self.getAllChildren(elem)
        if len(allChildren)==0:
            return []
        for node in allChildren:
            if self.getNodeTypeValue(node)==DECLARATION:
                result.append(self.makeDeclarationElement(node))
	    elif self.getNodeTypeValue(node)==DEFINITION:
                result.append(self.makeClassDefinition(self.getChild(node)))
	    elif self.getNodeTypeValue(node)==IMPORT:
		result.append(self.makeImportClause(node))
	    elif self.getNodeTypeValue(node)==EXTENDS:
		result.append(self.makeExtendsClause(node))
            else:
                raise 'not supported Modelia construct'
        return result

    # input node: 'DECLARATION'
    def makeDeclarationElement(self, elem):
	isFinal=False
	innerOuter=None
	localRef=self.getChild(elem)
	typeValue=self.getNodeTypeValue(localRef)
	while not (typeValue==IDENT or typeValue==DOT):
            if typeValue==FINAL:
                isFinal=True
            if typeValue==INNER:
                innerOuter='INNER'
            if typeValue==OUTER:
                innerOuter='OUTER'
	    if typeValue==REPLACEABLE:
		return self.makeReplaceableClause(elem)
		#redeclareClause=self.makeReplaceableClause(localRef)
		#redeclareClause.setFinal(isFinal)
		#redeclareClause.setInnerOuter(innerOuter)
		#return redeclareClause
	    localRef=self.getSibling(localRef)
	    typeValue=self.getNodeTypeValue(localRef)
	return self.makeComponentClause(elem)

    # input node: 'DECLARATION'
    def makeComponentClause(self, elem):
	isFinal=False
	innerOuter=None
	prefix=TypePrefix()
	type=None
	arrSub=None
	componentList=[]
	localRef=self.getChild(elem)
        typeValue=self.getNodeTypeValue(localRef)
        while not (typeValue==IDENT or typeValue==DOT):
	    if typeValue==FINAL:
		isFinal=True
	    if typeValue==INNER:
		innerOuter='INNER'
	    if typeValue==OUTER:
		innerOuter='OUTER'
            if typeValue==FLOW:
                prefix.setFlow(True)
            if typeValue==CONSTANT:
                prefix.setVariability('CONSTANT')
            elif typeValue==DISCRETE:
                prefix.setVariability('DISCRETE')
            elif typeValue==PARAMETER:
                prefix.setVariability('PARAMETER')
            if typeValue==INPUT:
                prefix.setCausality('INPUT')
            elif typeValue==OUTPUT:
                prefix.setCausality('OUTPUT')
            localRef=self.getSibling(localRef)
            typeValue=self.getNodeTypeValue(localRef)
	type=self.makeNameExp(localRef)
	localRef=self.getSibling(localRef)
	#allSiblings=self.getAllSiblings(localRef)
	if self.getNodeTypeValue(localRef)==LBRACK:
	    arrSub=self.makeArrSubExp(localRef)
	    localRef=self.getSibling(localRef)
	while not localRef==None:
	    if self.getNodeTypeValue(localRef)==IDENT:
		componentList.append(self.makeDeclaration(localRef))
	    #else:
		#pass 
		# the next sibling of IDENT--the COMMENT part 
		# is handled by the makeDeclaration() method
	    localRef=self.getSibling(localRef)
	return ComponentClause(-1, -1, prefix, type, componentList, \
			arrSub, isFinal, innerOuter)

    # input node: 'DEFINITION', 'DECLARATION'
    def makeReplaceableClause(self, elem):
	#print 'call makeReplaceableClause()...'
        isFinal=False
        innerOuter=None
	classDefinition=None
	componentClause=None
	constrainingClause=None
	comment=None
        localRef=self.getChild(elem)
	typeValue=self.getNodeTypeValue(localRef)
        while not (typeValue==CLASS_DEFINITION or typeValue==IDENT \
		   or typeValue==DOT):
            if self.getNodeTypeValue(localRef)==FINAL:
                isFinal=True
            if self.getNodeTypeValue(localRef)==INNER:
                innerOuter='INNER'
            elif self.getNodeTypeValue(localRef)==OUTER:
                innerOuter='OUTER'
            localRef=self.getSibling(localRef)
	    typeValue=self.getNodeTypeValue(localRef)
	if self.getNodeTypeValue(elem)==DEFINITION:
	    classDefinition=self.makeClassDefinition(localRef)
	    classDefinition.setFinal(isFinal)
	    classDefinition.setInnerOuter(innerOuter)
	    classDefinition.setReplaceable(True)
	elif self.getNodeTypeValue(elem)==DECLARATION:
	    componentClause=self.makeComponentClause(elem)
	    componentClause.setFinal(isFinal)
	    componentClause.setInnerOuter(innerOuter)
	    componentClause.setReplaceable(True)
	    while not typeValue==COMMENT: # COMMENT of declaration(handled)
	    	localRef=self.getSibling(localRef)
	    	typeValue=self.getNodeTypeValue(localRef)
	localRef=self.getSibling(localRef)
        typeValue=self.getNodeTypeValue(localRef)
	if typeValue==EXTENDS:
	    constrainingClause=self.makeExtendsClause(localRef)
	    localRef=self.getSibling(localRef)
	    typeValue=self.getNodeTypeValue(localRef)
	if typeValue==COMMENT:
	    comment=self.makeCommentExp(localRef)
	return ReplaceableClause(-1, -1, classDefinition, \
		componentClause, constrainingClause, comment, \
		isFinal, innerOuter)

    # input node: 'IMPORT'
    def makeImportClause(self, elem):
        #print 'makeImportClause()'
        ident=None
        name=None
        qualified=False
        comment=None
        localRef=self.getChild(elem)
        if self.getNodeTypeValue(localRef)==EQUALS:
            child=self.getChild(localRef)
            ident=IdentExp(-1,-1, self.getNodeTextValue(child))
            nextSibling=self.getSibling(child)
            name=self.makeNameExp(nextSibling)
            qualified=True
	elif self.getNodeTypeValue(localRef)==QUALIFIED:
	    name=self.makeNameExp(self.getChild(localRef))
	    qualified=True
        elif self.getNodeTypeValue(localRef)==UNQUALIFIED:
            name=self.makeNameExp(self.getChild(localRef))
        localRef=self.getSibling(localRef)
        comment=self.makeCommentExp(localRef)
        return ImportClause(-1, -1, name, ident, qualified, comment)

    # input node: 'EXTENDS'
    def makeExtendsClause(self, elem):
        #print 'makeExtendsClause()'
        child=self.getChild(elem)
        name=self.makeNameExp(child)
        modification=None
        nextSibling=self.getSibling(child)
        if self.getNodeTypeValue(nextSibling)==CLASS_MODIFICATION:
            modification=self.makeClassModification(nextSibling)
        return ExtendsClause(-1, -1, name, modification)

 
    # -------------------End of ClassDefinition -----------


    # ------------------- Declaration part -------------------

    # input node: 'IDENT'
    def makeDeclaration(self, elem):
	#print 'makeDeclaration()'
	#print self.getNodeTextValue(elem)
        ident=IdentExp(-1, -1, self.getNodeTextValue(elem))
        arrSub=None
	modification=None
	comment=None
        #modifications=[]
        allChildren=self.getAllChildren(elem)
        for node in allChildren:
            if self.getNodeTypeValue(node)==LBRACK: # arrSub
		#print str(node)
		#print self.getNodeTypeValue(node)
                arrSub=self.makeArrSubExp(node)
            elif self.getNodeTypeValue(node)==EQUALS:
                modification=self.makeEqualsModification(node)
		#modifications.append(\
		#	self.makeEqualsModification(node))
            elif self.getNodeTypeValue(node)==ASSIGN:
                modification=self.makeAssignModification(node)
		#modifications.append(\
		#	self.makeAssignModification(node))
            elif self.getNodeTypeValue(node)==CLASS_MODIFICATION:
                modification=self.makeClassModification(node)
		#modifications.append(\
		#	self.makeClassModification(node))
            else:
                pass
        next=self.getSibling(elem)
	if not self.getChild(next)==None:
            comment=self.makeCommentExp(next)
        return Declaration(-1, -1, ident, arrSub, modification, comment)
    

    # -----------------End of Declaration part ---------------



    # --------------------- Modification -------------------
    # input node: 'EQUALS', 'ASSIGN', 'CLASS_MODIFICATION'
    def makeModification(self, elem):
        if self.getNodeTypeValue(elem)==EQUALS:
            modification=self.makeEqualsModification(elem)
        elif self.getNodeTypeValue(elem)==ASSIGN:
            modification=self.makeAssignModification(elem)
        elif self.getNodeTypeValue(elem)==CLASS_MODIFICATION:
            modification=self.makeClassModification(elem)
        else:
            return None
        return modification
    
    # input is the 'CLASS_MODIFICATION' node
    def makeClassModification(self, elem):
	#print 'makeClassModification()'
	argList=[]
	modificationExp=None
	localRef=self.getChild(elem)
	if self.getNodeTypeValue(localRef)==ARGUMENT_LIST:
            allChildren=self.getAllChildren(localRef)
            for node in allChildren:
                argList.append(self.makeArgument(node))
	localRef=self.getSibling(elem)
	if not (localRef==None or \
		self.getNodeTypeValue(localRef)==COMMENT):
	    #modificationExp=self.makeExp(localRef)
	    modificationExp=EqualsModification(-1, -1, self.makeExp(localRef))
	return ClassModification(-1, -1, argList, modificationExp)

    # input is the 'EQUALS' node
    def makeEqualsModification(self, elem):
	#print 'makeEqualsModification()'
	child=self.getChild(elem)
	equals=self.makeExp(child)
	return EqualsModification(-1, -1, equals)

    # input is the 'ASSIGN' node
    def makeAssignModification(self, elem):
	#print 'makeAssignModification()'
	child=self.getChild(elem)
	equals=self.makeExp(child)
	return AssignModification(-1, -1, equals)

    # input is either 'ELEMENT_MODIFICATION' or 'ELEMENT_REDECLARATION'
    def makeArgument(self, elem):
	if self.getNodeTypeValue(elem)==ELEMENT_MODIFICATION:
	    return self.makeElementModification(elem)
	elif self.getNodeTypeValue(elem)==ELEMENT_REDECLARATION:
	    return self.makeElementRedeclaration(elem)
	else:
	    raise 'unsupported Modelica features ...'

    # input is the 'ELEMENT_MODIFICATION' node
    def makeElementModification(self, elem):
        #print 'makeElementModification( )'
	hasEach=False
	isFinal=False
        compRef=None
        modification=None
        stringComment=None
        localRef=self.getChild(elem)
        if self.getNodeTypeValue(localRef)==EACH:
	    hasEach=True
	    #raise 'EACH is not supported in this compiler ...'
	    localRef=self.getSibling(localRef)
	if self.getNodeTypeValue(localRef)==FINAL:
	    isFinal=True
            #raise 'FINAL is not supported in this compiler ...'
            localRef=self.getSibling(localRef)
        if self.getNodeTypeValue(localRef)==IDENT or \
           self.getNodeTypeValue(localRef)==DOT:
            compRef=self.makeCompRefExp(localRef)
            localRef=self.getSibling(localRef)
            modification=self.makeModification(localRef)
            localRef=self.getSibling(localRef)
            if self.getNodeTypeValue(localRef)==STRING_COMMENT:
                stringComment=self.makeCommentExp(localRef)
	else: # 'REDECLARE'
	    raise 'unsupported Modelica features (ElementModification)...'
        return ElementModification(-1, -1, compRef, modification, \
			stringComment, hasEach, isFinal)
    
    # input is the 'ELEMENT_REDECLARATION' node
    def makeElementRedeclaration(self, elem):
	classDefinition=None
	componentClause1=None
	replaceableClause1=None
	hasEach=False
	isFinal=False
	child=self.getChild(elem) # redeclare' node
	localRef=self.getChild(child)
	if self.getNodeTypeValue(localRef)==EACH:
	    hasEach=True
	    localRef=self.getSibling(localRef)
	if self.getNodeTypeValue(localRef)==FINAL:
	    isFinal=True
	    localRef=self.getSibling(localRef)
	if self.getNodeTypeValue(localRef)==CLASS_DEFINITION:
	    classDefinition=self.makeClassDefinition(localRef)
	    classDefinition.setFinal(isFinal)
	    #localRef=self.getSibling(localRef)
	elif self.getNodeTypeValue(localRef)==REPLACEABLE:
	    replaceableClause1=self.makeReplaceableClause1(localRef)
	    replaceableClause1.setFinal(isFinal)
	    # need setEach()??????????
	else: # 'componentClause!
	    componentClause1=self.makeComponentClause1(localRef)
	    componentClause1.setFinal(isFinal)
	return ElementRedeclaration(-1, -1, classDefinition, \
		componentClause1, replaceableClause1, hasEach, isFinal)

    # input node: 'replaceable'
    def makeReplaceableClause1(self, elem):
        isFinal=False
        innerOuter=None
        classDefinition=None
        componentClause=None
        constrainingClause=None
        comment=None
        localRef=self.getSibling(elem)
        typeValue=self.getNodeTypeValue(localRef)
	if typeValue==CLASS_DEFINITION:
	    classDefinition=self.makeClassDefinition(localRef)
	    classDefinition.setReplaceable(True)
	else: # ComponentClause1
	    componentClause=self.makeComponentClause1(localRef)
	    componentClause.setReplaceable(True)
	localRef=self.getSibling(localRef)	    
	typeValue=self.getNodeTypeValue(localRef)
        while not (typeValue==EXTENDS or localRef==None):
	    #print self.getNodeTextValue(localRef)
	    localRef=self.getSibling(localRef)
	    typeValue=self.getNodeTypeValue(localRef)
	if typeValue==EXTENDS:
	    constrainingClause=self.makeExtendsClause(localRef)
	return ReplaceableClause(-1, -1, classDefinition, \
                componentClause, constrainingClause, comment, \
                isFinal, innerOuter)

	

    # input node: could be typePrefix, 'IDENT' 'DOT'
    def makeComponentClause1(self, elem):
	prefix=TypePrefix()
        type=None
        localRef=elem
        typeValue=self.getNodeTypeValue(localRef)
        while not (typeValue==IDENT or typeValue==DOT):
            if typeValue==FLOW:
                prefix.setFlow(True)
            if typeValue==CONSTANT:
                prefix.setVariability('CONSTANT')
            elif typeValue==DISCRETE:
                prefix.setVariability('DISCRETE')
            elif typeValue==PARAMETER:
                prefix.setVariability('PARAMETER')
            if typeValue==INPUT:
                prefix.setCausality('INPUT')
            elif typeValue==OUTPUT:
                prefix.setCausality('OUTPUT')
            localRef=self.getSibling(localRef)
            typeValue=self.getNodeTypeValue(localRef)
	type=self.makeNameExp(localRef)
        localRef=self.getSibling(localRef)
        componentDeclaration=self.makeDeclaration(localRef)
	return ComponentClause1(-1, -1, prefix, type, \
		componentDeclaration)


    # ------------------End of Modification ------------



    # ----------------------- Equation Part -----------------
    # input node: 'EQUATION'.
    # return: []
    def makeEqStmList(self, elem):
        #print 'makeEqStmList()'
        result=[]
        allChildren=self.getAllChildren(elem)
        for node in allChildren: #all nodes are 'EQUATION_STATEMENT'
            result.append(self.makeEqStm(node))
        return result

    # input node: 'EQUATION_STATEMENT'
    def makeEqStm(self, elem):
        #print 'makeEqStm()'
        result=None
        child=self.getChild(elem)
        nodeType=self.getNodeTypeValue(child)
        if nodeType==EQUALS:
            result=self.makeRegularEqStm(elem)
        elif nodeType==CONNECT:
            result=self.makeConnectEqStm(elem)
        elif nodeType==IDENT: # Function call
            result=self.makeFunCallEqStm(elem)
        elif nodeType==IF:
            result=self.makeConditionalEqStm(child)
        elif nodeType==FOR:
            result=self.makeForEqStm(elem)
        elif nodeType==WHEN:
            result=self.makeWhenEqStm(child)         
        return result

    # input node: 'EQUATION_STATEMENT'
    def makeRegularEqStm(self, elem):
        #print 'makeRegularEqStm()'
	comment=None
        child=self.getChild(elem)
        next=self.getChild(child)
        LHS=self.makeExp(next)
        next=self.getSibling(next)
        RHS=self.makeExp(next)
	secondChild=self.getSibling(child)
	if not secondChild==None:
	    if not self.getChild(secondChild)==None:
            	comment=self.makeCommentExp(secondChild)
        return RegularEqStm(-1, -1, LHS, RHS, comment)

    # input node: 'EQUATION_STATEMENT'
    def makeConnectEqStm(self, elem):
        #print 'makeConnectEqStm()'
	comment=None
        child=self.getChild(elem) # 'CONNECT'
        next=self.getChild(child)
        firstElem=self.makeConnectRefExp(next)
        next=self.getSibling(next)
        secondElem=self.makeConnectRefExp(next)
	secondChild=self.getSibling(child)
	if not secondChild==None:
	    if not self.getChild(secondChild)==None:
            	comment=self.makeCommentExp(secondChild)
        return ConnectEqStm(-1, -1, firstElem, secondElem, comment)

    # input node: 'EQUATION_STATEMENT'
    def makeFunCallEqStm(self, elem):
        #print 'makeFunCallEqStm()'
        funCallOperandList=[]
        child=self.getChild(elem) #'IDENT'
        ident=IdentExp(-1, -1, self.getNodeTextValue(child))
        funCallOperandList.append(ident)
        next=self.getSibling(child) # 'FUNCTION_ARGUMENTS'
        funArgs=self.makeFunArgsExp(next)
        funCallOperandList.append(funArgs)
        funCall=FunCallExp(-1, -1, funCallOperandList)
        next=self.getSibling(next)
        comment=self.makeCommentExp(next) # 'COMMENT'
        return FunCallEqStm(-1, -1, funCall, comment)

    # input node: 'if' or 'elseif'
    def makeConditionalEqStm(self, elem):
        #print 'makeConditionalEqStm()'
	#firstChild=self.getChild(elem) # 'IF' node
	ifBody=[]
	elseifList=[]
	elseBody=[]
	comment=None
	localRef=self.getChild(elem) # ifTest
	ifTest=self.makeExp(localRef)
	localRef=self.getSibling(localRef)
	while self.getNodeTypeValue(localRef)==EQUATION_STATEMENT:
	    ifBody.append(self.makeEqStm(localRef))
	    localRef=self.getSibling(localRef)
	while self.getNodeTypeValue(localRef)==ELSEIF:
	    elseifList.append(self.makeConditionalEqStm(localRef))
	    localRef=self.getSibling(localRef)
	if self.getNodeTypeValue(localRef)==ELSE:
	    localRef=self.getSibling(localRef)
	    while self.getNodeTypeValue(localRef)==EQUATION_STATEMENT:
		elseBody.append(self.makeEqStm(localRef))
		localRef=self.getSibling(localRef)
	commentNode=self.getSibling(elem)
	if not commentNode==None:
	    if not self.getChild(commentNode)==None:
		comment=self.makeCommentExp(commentNode)
        return ConditionalEqStm(-1, -1, ifTest, ifBody, elseifList, \
				elseBody, comment)

    # input node: 'EQUATION_STATEMENT'
    def makeForEqStm(self, elem):
        #print 'makeForEqStm()'
        identInExpList=[]
        bodyEqList=[]
	comment=None
        child=self.getChild(elem) # 'FOR' 27
        next=self.getChild(child) # 'IN' 31
        while self.getNodeTypeValue(next)==IN:
            identInExpList.append(self.makeIdentInExpExp(next))
            next=self.getSibling(next)
        while self.getNodeTypeValue(next)==EQUATION_STATEMENT:
            bodyEqList.append(self.makeEqStm(next))
            next=self.getSibling(next)
        secondChild=self.getSibling(child)
        if not secondChild==None:
            if not self.getChild(secondChild)==None:
                comment=self.makeCommentExp(secondChild)
        forIndices=ForIndicesExp(-1,-1,identInExpList)
        return ForEqStm(-1, -1, forIndices, bodyEqList, comment)

    # input node: 'when' or 'elsewhen'
    def makeWhenEqStm(self, elem):
        #print 'makeWhenEqStm(): to be implemented ...'
	body=[]
	elsewhenList=[]
	comment=None
        localRef=self.getChild(elem) # whenTest
	whenTest=self.makeExp(localRef)
        localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==EQUATION_STATEMENT:
            body.append(self.makeEqStm(localRef))
            localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==ELSEWHEN:
            elsewhenList.append(self.makeWhenEqStm(localRef))
            localRef=self.getSibling(localRef)
        commentNode=self.getSibling(elem)
        if not commentNode==None:
            if not self.getChild(commentNode)==None:
                comment=self.makeCommentExp(commentNode)
        return WhenEqStm(-1, -1, whenTest, body, elsewhenList, comment)


    # inout node: 'ALGORITHM'.
    # Return: []    
    def makeAlgStmList(self, elem):
        #print 'makeAlgStmList(): to be implemented'
        result=[]
        allChildren=self.getAllChildren(elem)
        for node in allChildren: #all nodes are 'ALGORITHM_STATEMENT'
            result.append(self.makeAlgStm(node))
        return result

    # input node: 'ALGORITHM_STATEMENT'
    def makeAlgStm(self, elem):
        #print 'makeAlgStm()'
        result=None
        child=self.getChild(elem)
        nodeType=self.getNodeTypeValue(child)
        if nodeType==ASSIGN:
	    nextChild=self.getChild(child)
	    if self.getNodeTypeValue(nextChild)==DOT or \
			self.getNodeTypeValue(nextChild)==IDENT:
		result=self.makeAtomicAlgStm1(elem)
	    elif self.getNodeTypeValue(nextChild)==EXPRESSION_LIST:
		result=self.makeAtomicAlgStm2(elem)
	elif nodeType==DOT or nodeType==IDENT:
	    result=self.makeAtomicAlgStm1(elem)
        elif nodeType==IF:
            result=self.makeConditionalAlgStm(child)
        elif nodeType==FOR:
            result=self.makeForAlgStm(elem)
	elif nodeType==WHILE:
	    result=self.makeWhileAlgStm(elem)
        elif nodeType==WHEN:
            result=self.makeWhenAlgStm(child)
        return result

    # input node: 'EQUATION_STATEMENT'
    def makeAtomicAlgStm1(self, elem):
        #print 'makeAtomicAlgStm()'
	exp=None
	funCall=None 
        comment=None
        child=self.getChild(elem)
	if self.getNodeTypeValue(child)==ASSIGN:
	    nextChild=self.getChild(child)
	    compRef=self.makeCompRefExp(nextChild)
	    exp=self.makeExp(self.getSibling(nextChild))
	    secondChild=self.getSibling(child)
	    if not secondChild==None:
		if not self.getChild(secondChild)==None:
		    comment=self.makeCommentExp(secondChild)
	else:
	    compRef=self.makeCompRefExp(child)
	    nextSibling=self.getSibling(child)
	    funCall=self.makeFunArgsExp(nextSibling)
	    nextSibling=self.getSibling(nextSibling)
	    if not nextSibling==None:
		if not self.getChild(nextSibling)==None:
		    comment=self.makeCommentExp(nextSibling)
	return AtomicAlgStm1(-1, -1, compRef, exp, funCall, comment)

    # input node: 'EQUATION_STATEMENT'
    def makeAtomicAlgStm2(self, elem):
	#print 'makeAtomicAlgStm2()'
	comment=None
	child=self.getChild(elem) # ASSIGN node
	nextChild=self.getChild(child) # EXPRESSION_LIST node
	expList=self.makeExp(nextChild)
	nextSibling=self.getSibling(nextChild) # DOT or IDENT node
	compRef=self.makeCompRefExp(nextSibling)
	nextSibling=self.getSibling(nextSibling) # FUNCTION_ARGUMENTS
	funCall=self.makeFunArgsExp(nextSibling)
	commentNode=self.getSibling(child) # COMMENT node
	if not commentNode==None:
	    if not self.getChild(commentNode)==None:
		comment=self.makeCommentExp(commentNode)
	return AtomicAlgStm2(-1, -1, expList, compRef, funCall, comment)

    # input node: 'if' or 'elseif'
    def makeConditionalAlgStm(self, elem):
        #print 'makeConditionalAlgStm()'
        #firstChild=self.getChild(elem) # 'IF' node
        ifBody=[]
        elseifList=[]
        elseBody=[]
        comment=None
        localRef=self.getChild(elem) # ifTest
        ifTest=self.makeExp(localRef)
        localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==ALGORITHM_STATEMENT:
            ifBody.append(self.makeAlgStm(localRef))
            localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==ELSEIF:
            elseifList.append(self.makeConditionalAlgStm(localRef))
            localRef=self.getSibling(localRef)
        if self.getNodeTypeValue(localRef)==ELSE:
            localRef=self.getSibling(localRef)
            while self.getNodeTypeValue(localRef)==ALGORITHM_STATEMENT:
                elseBody.append(self.makeAlgStm(localRef))
                localRef=self.getSibling(localRef)
        commentNode=self.getSibling(elem)
        if not commentNode==None:
            if not self.getChild(commentNode)==None:
                comment=self.makeCommentExp(commentNode)
	
        return ConditionalEqStm(-1, -1, ifTest, ifBody, elseifList, \
                                elseBody, comment)

    # input node: 'ALGORITHM_STATEMENT'
    def makeForAlgStm(self, elem):
        #print 'makeForEqStm()'
        identInExpList=[]
        body=[]
        comment=None
        child=self.getChild(elem) # 'FOR' 27
        next=self.getChild(child) # 'IN' 31
        while self.getNodeTypeValue(next)==IN:
            identInExpList.append(self.makeIdentInExpExp(next))
            next=self.getSibling(next)
        while self.getNodeTypeValue(next)==ALGORITHM_STATEMENT:
            body.append(self.makeAlgStm(next))
            next=self.getSibling(next)
        secondChild=self.getSibling(child)
        if not secondChild==None:
            if not self.getChild(secondChild)==None:
                comment=self.makeCommentExp(secondChild)
        forIndices=ForIndicesExp(-1,-1,identInExpList)
        return ForEqStm(-1, -1, forIndices, body, comment)

    # input node: 'ALGORITHM_STATEMENT'
    def makeWhileAlgStm(self, elem):
	#print 'makeWhileAlgStm()'
        body=[]
        comment=None
        child=self.getChild(elem) # 'WHILE' 27
        localRef=self.getChild(child) 
	whileTest=self.makeExp(localRef)
	localRef=self.getSibling(localRef)
        #while self.getNodeTypeValue(next)==IN:
        #    identInExpList.append(self.makeIdentInExpExp(next))
        #    next=self.getSibling(next)
        while self.getNodeTypeValue(localRef)==ALGORITHM_STATEMENT:
            body.append(self.makeAlgStm(localRef))
            localRef=self.getSibling(localRef)
        secondChild=self.getSibling(child)
        if not secondChild==None:
            if not self.getChild(secondChild)==None:
                comment=self.makeCommentExp(secondChild)
        return WhileAlgStm(-1, -1, whileTest, body, comment)

    # input node: 'when' or 'elsewhen'
    def makeWhenAlgStm(self, elem):
        #print 'makeWhenEqStm(): to be implemented ...'
        body=[]
        elsewhenList=[]
        comment=None
        localRef=self.getChild(elem) # whenTest
        whenTest=self.makeExp(localRef)
        localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==ALGORITHM_STATEMENT:
            body.append(self.makeAlgStm(localRef))
            localRef=self.getSibling(localRef)
        while self.getNodeTypeValue(localRef)==ELSEWHEN:
            elsewhenList.append(self.makeWhenAlgStm(localRef))
            localRef=self.getSibling(localRef)
        commentNode=self.getSibling(elem)
        if not commentNode==None:
            if not self.getChild(commentNode)==None:
                comment=self.makeCommentExp(commentNode)
        return WhenEqStm(-1, -1, whenTest, body, elsewhenList, comment)

    # -----------------------End of Equation Part -----------------



    # ------------------------- Exp ----------------------
    # input could be many types of node
    # including '(', 'EXPRESSION_LIST', '[', 'FUNCTION_CALL'
    # ...
    def makeExp(self, elem):
        #print 'makeExp()'
        operands=[]
        nodeType=self.getNodeTypeValue(elem)
        # --- AtomicExp ---
        if nodeType==IDENT:
            # might also be CompRef
            if self.getChild(elem)==None:
                return IdentExp(-1, -1, self.getNodeTextValue(elem))
            else:
                return self.makeCompRefExp(elem)
        elif nodeType==UNSIGNED_INTEGER:
            return IntegerExp(-1, -1, int(self.getNodeTextValue(elem)))
        elif nodeType==UNSIGNED_REAL:
            return RealExp(-1, -1, float(self.getNodeTextValue(elem)))
        elif nodeType==STRING:
            return StringExp(-1, -1, self.getNodeTextValue(elem))
        elif nodeType==TRUE:
	    return BoolExp(-1, -1, True)
	elif nodeType==FALSE:
            return BoolExp(-1, -1, False)
        # --- CompositeExp ---
        elif nodeType==OR:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return OrExp(-1, -1, operands)
        elif nodeType==AND:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return AndExp(-1, -1, operands)
        elif nodeType==NOT:
            child=self.getChild(elem)
            operands.append(self.makeExp(child))
            return NotExp(-1, -1, operands)
        elif nodeType==LESS:
	    #print 'found <: ', self.getNodeTextValue(elem)
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return LessExp(-1, -1, operands)
        elif nodeType==LESSEQ:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return LessEQExp(-1, -1, operands)
        elif nodeType==GREATER:
	    #print 'found >: ', self.getNodeTextValue(elem)
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
	    result=GreaterExp(-1, -1, operands)
            return GreaterExp(-1, -1, operands)
        elif nodeType==GREATEREQ:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return GreaterEQExp(-1, -1, operands)
        elif nodeType==EQEQ:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return EQEQExp(-1, -1, operands)
        elif nodeType==LESSGT:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return LessGtExp(-1, -1, operands)
        elif nodeType==PLUS:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return SumExp(-1, -1, operands)
        elif nodeType==MINUS:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return SubtractExp(-1, -1, operands)
        elif nodeType==STAR:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return MulExp(-1, -1, operands)
        elif nodeType==SLASH:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return DivExp(-1, -1, operands)
        elif nodeType==POWER:
            firstChild=self.getChild(elem)
            secondChild=self.getSibling(firstChild)
            operands.append(self.makeExp(firstChild))
            operands.append(self.makeExp(secondChild))
            return PowerExp(-1, -1, operands)
        elif nodeType==UNARY_MINUS:
            child=self.getChild(elem)
            operands.append(self.makeExp(child))
            return UnaryMinusExp(-1, -1, operands)
        elif nodeType==RANGE2 or nodeType==RANGE3 \
	     or nodeType==COLON:
            allChildren=self.getAllChildren(elem)
            for node in allChildren:
                operands.append(self.makeExp(node))
            return RangeExp(-1, -1, operands)
        elif nodeType==DOT:
            # makeNameExp is called directly
            return self.makeCompRefExp(elem)
        elif nodeType==FUNCTION_CALL:
            firstChild=self.getChild(elem) # 'IDENT'
            secondChild=self.getSibling(firstChild) # 'FUNCTION_ARGUMENTS'
            #ident=IdentExp(-1, -1, self.getNodeTextValue(firstChild))
	    name=self.makeCompRefExp(firstChild)
            #operands.append(ident)
	    operands.append(name)
            operands.append(self.makeFunArgsExp(secondChild))
            return FunCallExp(-1, -1, operands)
        elif nodeType==FUNCTION_ARGUMENTS:
            return self.makeFunArgsExp(elem)
        elif nodeType==EXPRESSION_LIST:
	    allChildren=self.getAllChildren(elem)
	    for node in allChildren:
		operands.append(self.makeExp(node))
            return ExpListExp(-1, -1, operands)
        elif nodeType==LPAR: #'(' followed by 'EXPRESSION_LIST'
	    child=self.getChild(elem) # 'EXPRESSION_LIST'
	    if self.getNodeTypeValue(child)==EXPRESSION_LIST:
		#allChildren=self.getAllChildren(child)
		#for node in allChildren:
		#    operands.append(self.makeExp(node))
		#return ExpListExp(-1, -1, operands, 'PAR')
		# *** the '(' & ')' should no longer appears in AST
		grandChild=self.getChild(child)
		return self.makeExp(grandChild)
        elif nodeType==LBRACK: #'[': arrSubExp or list of  EXPRESSION_LIST
	    child=self.getChild(elem)
	    if self.getNodeTypeValue(child)==EXPRESSION_LIST:
		operands.append(self.makeExp(child))
		allSiblings=self.getAllSiblings(child)
		if len(allSiblings)>0:
		    for node in allSiblings:
			operands.append(self.makeExp(node))
		#allChildren=self.getAllChildren(child)
                #for node in allChildren:
                #    operands.append(self.makeExp(node))
                return ExpListExp(-1, -1, operands, 'BRACK')
	    else: # arrSubExp
		return self.makeArrSubExp(elem)
	elif nodeType==LBRACE: # '{' EXPRESSION_LIST
	# possible for FUNCTION_ARGUMENTS????????
	    child=self.getChild(elem)
            if self.getNodeTypeValue(child)==EXPRESSION_LIST:
                allChildren=self.getAllChildren(child)
                for node in allChildren:
                    operands.append(self.makeExp(node))
                return ExpListExp(-1, -1, operands, 'BRACE')
	    else: # need check for other possibilities
		return None
        elif nodeType==IF or nodeType==ELSEIF:
	    elseifList=[]
	    elseBody=None
	    localRef=self.getChild(elem)
	    ifTest=self.makeExp(localRef)
	    localRef=self.getSibling(localRef)
	    ifBody=self.makeExp(localRef)
	    localRef=self.getSibling(localRef)
	    while self.getNodeTypeValue(localRef)==ELSEIF:
		elseifList.append(self.makeExp(localRef))
		localRef=self.getSibling(localRef)
	    if not localRef==None:
		elseBody=self.makeExp(localRef)
            return IfElseExp(-1, -1, ifTest, ifBody, elseifList, elseBody)
        elif nodeType==FOR:
            # make ForFunArgExp()
            firstChild=self.getChild(elem) # exp
            exp=self.makeExp(firstChild)
            operands.append(exp)
            allSiblings=self.getAllSiblings(firstChild)
            for node in allSiblings:
                operands.append(self.makeIdentInExpExp(node))
            return ForFunArgExp(-1,-1,operands)
        elif nodeType==EQUALS:
            # make NamedArgExp
            firstChild=self.getChild(elem) # 'IDENT'
            secondChild=self.getSibling(firstChild) # EXP
            ident=IdentExp(-1, -1, self.getNodeTextValue(firstChild))
            operands.append(ident)
            operands.append(self.makeExp(secondChild))
            return NamedArgExp(-1, -1, operands)
        elif nodeType==IN:
            return self.makeIdentInExpExp(elem)
        else:
            raise 'NOT defined in makeExp()!'
        return None
    
    # input is either the '.' or 'IDENT' node
    # Not yet expanded !!!
    def makeNameExp(self, elem):
        #print 'makeNameExp()'
        operands=[]
        if self.getNodeTypeValue(elem)==IDENT:
            operands.append(IdentExp(-1, -1, self.getNodeTextValue(elem)))
            return NameExp(-1, -1, operands)
        elif self.getNodeTypeValue(elem)==DOT:
            #firstChild=self.getChild(elem)
            #secondChild=self.getSibling(firstChild)
            #operands.append(self.makeNameExp(firstChild))
            #operands.append(self.makeNameExp(secondChild))
            allChildren=self.getAllChildren(elem)
            stop=False
            while not stop:
                numOfIdentInAllChildren=0
                for i in allChildren:
                    if self.getNodeTypeValue(i)==IDENT:
                        operands.append( \
                          IdentExp(-1, -1, self.getNodeTextValue(i)))
                        numOfIdentInAllChildren=numOfIdentInAllChildren+1
                    elif self.getNodeTypeValue(i)==DOT:
                        nextAllChildren=self.getAllChildren(i)
                if numOfIdentInAllChildren==2:
                    stop=True
		else:
		    allChildren=nextAllChildren
        return NameExp(-1, -1, operands)

    # input elem is the '[' node
    def makeArrSubExp(self, elem):
	#print 'makeArrSubExp()'
	operands=[]
	allChildren=self.getAllChildren(elem)
	for node in allChildren:
            operands.append(self.makeExp(node))
	return ArrSubExp(-1, -1, operands)
    
    # input node: 'FUNCTION_ARGUMENTS'
    # children might be 'EXPRESSION_LIST', 'NAMED_ARGUMENTS',
    def makeFunArgsExp(self, elem):
        #print 'makeFunArgsExp()'
        operands=[]
        allChhildren=self.getAllChildren(elem)
        if len(allChhildren)==0:
            return FunArgsExp(-1, -1, operands)
        for node in allChhildren:
            if self.getNodeTypeValue(node)==EXPRESSION_LIST:
                nextAllChildren1=self.getAllChildren(node)
                for node1 in nextAllChildren1:
                    operands.append(self.makeExp(node1))
            if self.getNodeTypeValue(node)==NAMED_ARGUMENTS:
                nextAllChildren2=self.getAllChildren(node)
                for node2 in nextAllChildren2:
                    operands.append(self.makeExp(node2))
        return FunArgsExp(-1, -1, operands)
        
    # input node could be 'IDENT' or 'DOT' (else?)
    def makeConnectRefExp(self, elem):
        #print 'makeConnectRefExp()'
	operands=[]
	if self.getNodeTypeValue(elem)==IDENT:
	    #operands.append(IdentExp(-1,-1,self.getNodeTextValue(elem)))
	    #child=self.getChild(elem)
	    #if not child==None:
		#operands.append(self.makeArrSubExp(child))
	    operands.append(self.makeSingleCompRefExp(elem))    
	    return ConnectRefExp(-1, -1, operands)
	elif self.getNodeTypeValue(elem)==DOT:
	    firstChild=self.getChild(elem) # must be an IDENT
	    #operands.append(\
		#IdentExp(-1,-1,self.getNodeTextValue(firstChild)))
	    #nextChild=self.getChild(firstChild)
	    #if not nextChild==None:
		#operands.append(self.makeArrSubExp(nextChild))
	    operands.append(self.makeSingleCompRefExp(firstChild))
	    secondChild=self.getSibling(firstChild) # must be an IDENT
	    #operands.append(\
            #    IdentExp(-1,-1,self.getNodeTextValue(secondChild)))
	    #nextChild=self.getChild(secondChild)
	    #if not nextChild==None:
		#operands.append(self.makeArrSubExp(nextChild))
	    operands.append(self.makeSingleCompRefExp(secondChild))
        return ConnectRefExp(-1, -1, operands)
    
    # input node: 'IDENT'
    def makeSingleCompRefExp(self, elem):
	operands=[]
	operands.append(IdentExp(-1,-1,self.getNodeTextValue(elem)))
	child=self.getChild(elem)
	if not child==None:
	    operands.append(self.makeArrSubExp(child))
	return SingleCompRefExp(-1, -1, operands)

    # input is either the '.' or 'IDENT' node
    def makeCompRefExp(self, elem):
        #print 'makeCompRefExp()'
        operands=[]
        if self.getNodeTypeValue(elem)==IDENT:
            #operands.append(IdentExp(-1,-1,self.getNodeTextValue(elem)))
            #child=self.getChild(elem)
            #if not child==None:
            #    operands.append(self.makeArrSubExp(child))
            #return CompRefExp(-1, -1, operands)
	    operands.append(self.makeSingleCompRefExp(elem))
	    return CompRefExp(-1, -1, operands)
	elif self.getNodeTypeValue(elem)==DOT:
	    allChildren=self.getAllChildren(elem)
            stop=False
            while not stop:
                numOfIdentInAllChildren=0
                for i in allChildren:
                    if self.getNodeTypeValue(i)==IDENT:
                        #operands.append( \
                        #  IdentExp(-1, -1, self.getNodeTextValue(i)))
			#child=self.getChild(i)
			#if not child==None:
			#    operands.append(self.makeArrSubExp(child))
			operands.append(self.makeSingleCompRefExp(i))
                        numOfIdentInAllChildren=numOfIdentInAllChildren+1
                    elif self.getNodeTypeValue(i)==DOT:
                        nextAllChildren=self.getAllChildren(i)
                if numOfIdentInAllChildren==2:
                    stop=True
                else:
                    allChildren=nextAllChildren
        return CompRefExp(-1, -1, operands)
    
    # input is either the 'COMMENT' or 'STRING_COMMENT' node
    def makeCommentExp(self, elem):
	operands=[]
	if self.getNodeTypeValue(elem)==STRING_COMMENT:
	    child=self.getChild(elem)
	    operands.append(StringExp(-1,-1,self.getNodeTextValue(child)))
	    return CommentExp(-1,-1,operands)
	elif self.getNodeTypeValue(elem)==COMMENT:
	    child=self.getChild(elem)
	    if self.getNodeTypeValue(child)==STRING_COMMENT:
		grandChild=self.getChild(child)
		operands.append(\
			StringExp(-1,-1,self.getNodeTextValue(grandChild)))
	        nextChild=self.getSibling(child)
		if self.getNodeTypeValue(nextChild)==ANNOTATION:
		    operands.append(\
			self.makeAnnotationExp(nextChild))
        return CommentExp(-1,-1,operands)

    # input node: 'IN'
    def makeIdentInExpExp(self, elem):
        #print 'makeIdentInExpExp()'
	operands=[]
	firstChild=self.getChild(elem)
	secondChild=self.getSibling(firstChild)
	ident=IdentExp(-1, -1, self.getNodeTextValue(firstChild))
	operands.append(ident)
	operands.append(self.makeExp(secondChild))
        return IdentInExpExp(-1,-1,operands)

    # input node: 'annotation'
    def makeAnnotationExp(self, elem):
        #print 'makeForIndicesExp(): to be implemented ...'
	annotation=None
	child=self.getChild(elem)
	if self.getNodeTypeValue(child)==CLASS_MODIFICATION:
	    annotation=self.makeClassModification(child)
	return AnnotationExp(-1,-1,annotation)

    # -------------------------End of Exp ----------------------



    # ------------------- private methods ---------------------
    def getChild(self, elem):
	try:
	    child=elem.outC().getElementsOfOrd(0)[0]
	except:
	    return None
	else:
	    return child

    # get all children of elem
    def getAllChildren(self, elem):
        list=[]
        elem=self.getChild(elem)
        while not elem==None:
            list.append(elem)
            elem=self.getSibling(elem)
        return list

    def getSibling(self, elem):
	try:
	    sibling=elem.outC().getElementsOfOrd(1)[0]
	except:
	    return None
	else:
	    return sibling

    # get all siblings of elem (not including elem)
    def getAllSiblings(self, elem):
        list=[]
        elem=self.getSibling(elem)
        while not elem==None:
            list.append(elem)
            elem=self.getSibling(elem)
        return list

    # return type: Int
    def getNodeTypeValue(self, elem):
	try:
	    result=elem["type"].getValue()
	except:
	    return -1
	else:
	    return result

    # return type: String
    def getNodeTextValue(self, elem):
	try:
	    result=elem["text"].getValue()
	except:
	    return -1
	else:
	    return result


    # ------------------- Auxiliary methods ---------------------------

    # For each enumeration
    #    type E = enumeration(e1, ...en);
    # a new simple type is conceptually defined as
    # 	type E
    #	  EnumType value;
    #	  parameter StringType quantity = "";
    #	  parameter EnumType min = e1, max = e3;
    #	  parameter EnumType start = e1;
    #	  parameter BooleanType fixed = false;
    #	  parameter BooleanType enable = true;
    #	  constant EnumType e1;
    #	  ...
    #	  constant EnumTYpe en;
    #	equation
    #	  assert(value>=min and value<=max, "Variable value out of limit");
    #	end E; 
    # enum: EnumClassDefinition
    def transformEnumClassDefinition(self, enum):
	lineNum=enum.lineNum
	colNum=enum.colNum
	prefix=enum.getClassPrefix()
	kind=enum.getKind()
	className=enum.getName()
	stringComment=None
	# declaration
	elementList=[]
	arrSub=None
	comment=None
	isFinal=False
	innerOuter=None
	isReplaceable=False
	# element[0]
	typePrefix=TypePrefix()
	nameOperands1=[]
	nameOperands1.append(IdentExp(-1,-1,'EnumType'))
	enumType=NameExp(-1,-1, nameOperands1)
	type=enumType
	componentList0=[] # list of ASTComonent::Declaration
	ident0=IdentExp(-1,-1,'value')
	modification0=None
	componentList0.append(Declaration(-1, -1, ident0, arrSub, \
				modification0,comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
			componentList0, arrSub, isFinal, innerOuter, \
			isReplaceable))	
	# element[1]
	typePrefix=TypePrefix(variabilityP='PARAMETER')
	nameOperands2=[]
	nameOperands2.append(IdentExp(-1,-1,'StringType'))
	stringType=NameExp(-1,-1, nameOperands2)
	type=stringType
	componentList1=[]
	ident1=IdentExp(-1,-1,'quantity')
	modification1=EqualsModification(-1, -1, StringExp(-1,-1,""))
	componentList1.append(Declaration(-1, -1, ident1, arrSub, \
                                modification1, comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
			componentList1, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# element[2]
	type=enumType
	componentList2=[]
	ident21=IdentExp(-1,-1,'min')
	modification21=EqualsModification(-1, -1, enum.enumList[0].ident)
	componentList2.append(Declaration(-1, -1, ident21, arrSub, \
				modification21, comment))
	ident22=IdentExp(-1,-1,'max')
	modification22=EqualsModification(-1, -1, \
			enum.enumList[len(enum.enumList)-1].ident)
	componentList2.append(Declaration(-1, -1, ident22, arrSub, \
                                modification22, comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
                        componentList2, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# element[3]
	componentList3=[]
	ident3=IdentExp(-1,-1,'start')
	modification3=EqualsModification(-1, -1, enum.enumList[0].ident)
	componentList3.append(Declaration(-1, -1, ident3, arrSub, \
				modification3, comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
			componentList3, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# element[4]
	nameOperands3=[]
        nameOperands3.append(IdentExp(-1,-1,'BooleanType'))
        booleanType=NameExp(-1,-1, nameOperands3)
	type=booleanType
	componentList4=[]
	ident4=IdentExp(-1,-1,'fixed')
	modification4=EqualsModification(-1, -1, BoolExp(-1,-1,False))
	componentList4.append(Declaration(-1, -1, ident4, arrSub, \
				modification4, comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
                        componentList4, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# element[5]
	componentList5=[]
	ident5=IdentExp(-1,-1,'enable')
	modification5=EqualsModification(-1, -1, BoolExp(-1,-1,True))
	componentList5.append(Declaration(-1, -1, ident5, arrSub, \
                                modification5, comment))
	elementList.append(ComponentClause(-1, -1, typePrefix, type, \
                        componentList5, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# for loop: e1 .... en
	typePrefix=TypePrefix(variabilityP='CONSTANT')
	type=enumType
	for i in enum.enumList: # i: EnumLiterla
	    j=i.ident 
	    modification=None
	    componentList=[]
	    componentList.append(Declaration(-1, -1, j, arrSub, \
				modification, comment))
	    elementList.append(ComponentClause(-1, -1, typePrefix, type, \
			componentList, arrSub, isFinal, innerOuter, \
			isReplaceable))
	# to be added: assert()...
	return RegularClassDefinition(lineNum, colNum, prefix, kind, \
		className, stringComment, elementListP=elementList) 
