# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the visitor module

from AST.ASTDefinition import ClassDefinition
from AST.ASTScope import DefinitionEntry
from AST.ASTScope import DeclarationEntry, SymbolTable
from FrontEnd.Exception.NameException import NameNotDeclaredException, \
                                             UnencapsulatedElementException
from FrontEnd.Exception.ScopeException import EncapsulatedInheritanceException
from FrontEnd.Visitor import AbstractVisitor

# ---------------------------------------------------------------------
class FlatModelicaNameLookuper(AbstractVisitor):
    "name lookup in flat Modelica"
    def __init__(self):
	# the debugging message
	self.msg=''
	# the current scope
        self.currentScope=None

    # ---------------------ClassDefinition------------------
    def visitAST(self, ast):
	ast.classFiles[0].accept(self)

    def visitClassFile(self, classFile):
	classFile.classDefinitions[0].accept(self)

    def visitRegularClassDefinition(self, rClassDefinition):
	self.msg=self.msg+'\n In class: '+str(rClassDefinition.getName())
	# all names in equations are looked up in the flat scope
	self.currentScope=rClassDefinition.getFlatScope()
	for i in rClassDefinition.getFlattenedInitEqs():
	    i.accept(self)
	for i in rClassDefinition.getFlattenedEqs():
	    i.accept(self)
	self.msg=self.msg+'\n end of class: '+str(rClassDefinition.getName())
	    

    def visitShortClassDefinition(self, sClassDefinition):
	self.msg=self.msg+'\n In class: '+str(sClassDefinition.getName())
        # all names are looked up in the flat scope
        self.currentScope=sClassDefinition.getFlatScope()
        for i in sClassDefinition.getFlattenedInitEqs():
            i.accept(self)
        for i in sClassDefinition.getFlattenedEqs():
            i.accept(self)
	self.msg=self.msg+'\n end of class: '+str(sClassDefinition.getName())

    def visitEnumClassDefinition(self, eClassDefinition):
	# EnumClassDefinition has been transformed to 
	# RegularClassDefinition
        pass


    # -----------------------Component----------------------------
    def visitDeclaration(self, declaration):
	pass

    def visitClassModification(self, cm):
	raise "This method needs to be implemented ..."

    def visitEqualsModification(self, em):
	raise "This method needs to be implemented ..."

    def visitAssignModification(self, am):
	raise 'AssignModification is not supported ...'

    def visitElementModification(self, elemModif):
	raise "This method needs to be implemented ..."

    def visitElementRedeclaration(self, elemRedcl):
	raise "This method needs to be implemented ..."

    # ------------------------------Equation----------------------
    def visitRegularEqStm(self, eqStm):
	self.msg=self.msg+'\n In equation: '+str(eqStm)
	eqStm.LHS.accept(self)
	eqStm.RHS.accept(self)

    def visitConnectEqStm(self, eqStm):
	self.msg=self.msg+'\n In equation: '+str(eqStm)
        eqStm.first.accept(self)
	eqStm.second.accept(self)

    def visitFunCallEqStm(self, eqStm):
        eqStm.funCall.accept(self)

    def visitConditionalEqStm(self, eqStm):
        eqStm.ifTest.accept(self)
	for i in eqStm.ifBody:
	    i.accept(self)
	for i in eqStm.elseifList:
	    i.accept(self)
	for i in eqStm.elseBody:
	    i.accept(self)

    def visitForEqStm(self, eqStm):
	raise "This method needs to be implemented ..."

    def visitWhenEqStm(self, eqStm):
	eqStm.whenTest.accept(self)
	for i in eqStm.body:
	    i.accept(self)
	for i in elsewhenList:
	    i.accept(self)

    # ------------------------------Exp-----------------------------
    # lookup of simple names (identifiers), eg, a, b
    def visitIdentExp(self, exp):
	if str(exp)=='time':
            return
	found=self.lookup(self.currentScope, str(exp))
	if found == None:
            raise NameNotDeclaredException(str(exp))
	#elif self.isVisitingFunCall:
	#    if not isinstance(found, DefinitionEntry):
	#	raise NameNotDeclaredException(str(exp))
	else:
	    if not isinstance(found, DeclarationEntry):
		raise NameNotDeclaredException(str(exp))
	exp.setValue(found)
	self.msg=self.msg+'\n '+str(exp)+' is found;'

    def visitRealExp(self, exp):
	pass

    def visitIntegerExp(self, exp):
	pass

    def visitBoolExp(self, exp):
	pass

    def visitStringExp(self, exp):
	pass

    # this method is not visited
    def visitNameExp(self, exp):
	pass

    def visitOrExp(self, exp):
	for i in exp.getOperands():
	    i.accept(self)

    def visitAndExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitNotExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitLessExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def vistLessEQExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitGreaterExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitGreaterEQExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitEQEQExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitLessGtExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitSumExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitMulExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitSubtractExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitDivExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitUnaryMinusExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitPowerExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitRangeExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitIfElseExp(self, exp):
        exp.ifTest.accept(self)
        exp.ifBody.accept(self)
        for i in exp.elseifList:
            i.accept(self)
        if not exp.elseBody==None:
            exp.elseBody.accept(self)

    def visitFunCallExp(self, exp):
	# set the flag
	#self.isVisitingFunCall=True
	# a funtion name is first looked up in predefined scope
	#found=self.predefinedScope.localLookup(str(exp.getOperands()[0]))
        #if not found ==None:
	#    self.msg=self.msg+'\n '+str(exp.getOperands()[0])+' is found;'
	#else: # normal lookup
	#    exp.getOperands()[0].accept(self)
	#self.isVisitingFunCall=False
	# lookup of parameters
	exp.getOperands()[1].accept(self)

    def visitFunArgsExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitForFunArgExp(self, exp):
	operands=exp.getOperands()
	# enter new scope
	self.currentScope=exp.getScope()
	# only the first elem is looked up in the innner scope
	operands[0].accept(self)
	# exit scope
	self.currentScope=self.currentScope.getOuter()
	for i in range(1, len(operands)):
            operands[i].accept(self)

    def visitNamedArgExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitConnectRefExp(self, exp):
	pass

    def visitSingleCompRefExp(self, exp):
	pass

    def visitCompRefExp(self, exp):
	# CompRefExps have been transformed to IdentExp
	pass

    def visitArrSubExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitCommentExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitAnnotationExp(self, exp):
	pass

    def visitIdentInExpExp(self, exp):
	exp.getOperands()[1].accept(self)

    def visitForIndicesExp(self, exp):
	for i in exp.getOperands():
	    # i: IdentInExpExp
	    i.accept(self)

    def visitExpListExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    # ----------------- lookup algorithms -----------------------

    # Look up a name locally in scope sp
    # sp: ASTScope::SymbolTable
    # exp: ASTExp::IdentExp
    def lookup(self, sp, exp):
	return sp.localLookup(str(exp))
 


