# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University



import copy
from AST.ASTDefinition import ShortClassDefinition, ExtendsClause
from AST.ASTComponent import ElementRedeclaration, ClassModification
from AST.ASTComponent import EqualsModification, AssignModification
from AST.ASTExp import IdentExp, SingleCompRefExp, CompRefExp
from AST.ASTScope import ComponentPrefix, DefinitionEntry
from AST.ASTScope import DeclarationEntry, SymbolTable
from FrontEnd.Visitor import AbstractVisitor
from FrontEnd.Exception.NameException import NameNotDeclaredException, \
                                             UnencapsulatedElementException


# ---------------------------------------------------------------------
class MyStack:
    def __init__(self):
	self.list=[]

    def pop(self):
	if len(self.list)>0:
	    del self.list[len(self.list)-1]

    def push(self, elem):
	self.list.append(elem)

    def getTopElem(self):
	if len(self.list)>0:
            return self.list[len(self.list)-1]
        else:
            return None	


class InheritanceExpander(AbstractVisitor):
    "Expand inheritances"
    def __init__(self, ast):
	self.ast=ast
        # the scope node which stores all builtin primitive types
        self.builtinScope=None
	# the scope node which stores all predefined primitive types
	self.predefinedScope=None
	self.currentScope=None
	# this stack is used to handle local class definition with 
	# extends_clauses
	self.classStack=MyStack()
	self.isEncapsulated=False
	# flag: indicates whether it is expanding a ShortClassDefinition
	# to reular form

    def setBuiltinScope(self, scope):
        self.builtinScope=scope

    def setPredefinedScope(self, scope):
	self.predefinedScope=scope

    def setCurrentScope(self, scope):
	self.currentScope=scope

    # ---------------------------ClassDefinition-----------------------
    def visitAST(self, ast):
        self.predefinedScope=self.ast.getPredefinedScope()
        self.builtinScope=self.ast.getBuiltinScope()
        for i in ast.classFiles:
            i.accept(self)

    def visitClassFile(self, classFile):
        #self.setBuiltinScope(classFile.getBuiltinScope())
	#self.setPredefinedScope(classFile.getPredefinedScope())
	self.currentScope=classFile.getScope()
	for i in classFile.classDefinitions:
	    i.accept(self)

    def visitRegularClassDefinition(self, rClassDefinition):
	self.classStack.push(rClassDefinition)
	# this affects the base-classname lookup 
	if rClassDefinition.isEncapsulated():
	    self.isEncapsulated=True
	# enter new Scope
	self.setCurrentScope(rClassDefinition.getScope())
	for i in rClassDefinition.elementList:
	    i.accept(self)
	# reset global flags
	self.setCurrentScope(self.currentScope.getOuter())
	self.isEncapsulated=False
	self.classStack.pop()

    # short class definition is transformed to RegularClassDefinition
    def visitShortClassDefinition(self, sClassDefinition):
	# code to handle array to be added
	self.classStack.push(sClassDefinition)
	if sClassDefinition.isEncapsulated():
            self.isEncapsulated=True
	tempExtendsClause=ExtendsClause(-1,-1,\
				sClassDefinition.getBaseName(),\
				sClassDefinition.getModification(),\
				sClassDefinition.getVisibility())
	tempExtendsClause.accept(self)
	# need resolve basePreix info here (merge of perfixs)
	self.isEncapsulated=False
	self.classStack.pop()

    def visitEnumClassDefinition(self, eClassDefinition):
	# EnumClassDefinition has been transformed to RegularClassDefinition
	pass

    def visitComponentClause(self, componentClause):
	pass

    def visitExtendsClause(self, extendsClause):
	#if extendsClause.isExpanded():
	#    return 
	currentClass=self.classStack.getTopElem()
	self.expandInheritance(currentClass, extendsClause)

    def visitReplaceableClause(self, replaceableClause):
	raise "\"Replaceable\" found: not supported! Abort ... "

    def visitImportClause(self, importClause):
	raise "\"Import clause\" found: not yet supported! Abort ..."


    # ------------------------------EqStm-----------------------------

    # ------------------------------Exp-----------------------------

    #---------------------- Expand Inheritance ---------------------

    # expand "extendsClause" in "classDef"
    def expandInheritance(self, classDef, extendsClause):
        if extendsClause.isExpanded():
            return
        found=self.baseClassLookup(classDef.getScope(), \
		extendsClause.getName())
        baseClass=found.getDefinition()
	#if isinstance(classDef, ShortClassDefinition):
	#    classDef.setBaseClass(baseClass)
        if not baseClass.isExpanded():
            for i in baseClass.getExtendsClauses():
                self.expandInheritance(baseClass, i)
		self.copyElements(classDef, baseClass, i)
	# resolve modifications and copy elements from base class
	self.copyElements(classDef, baseClass, extendsClause)
	# mark the extendsClause as 'expanded'
        extendsClause.setExpanded(True)

    # resolve modifications in "extendsClause", and copy elements 
    # from "baseClass" to "classDef"
    def copyElements(self, classDef, baseClass, extendsClause):
	modif=extendsClause.getModification()
	visibility=extendsClause.getVisibility()
	st=baseClass.getScope()
	# copy definition entries
	for i in st.getDefinitions():
	    # this uses the default Python shallow copy
	    j=copy.copy(i)
            if visibility=='PROTECTED':
                j.setVisibility('PROTECTED')
            classDef.getScope().add(str(j.getName()), j)
	# copy declaration entries
	# First, copy declaration elements into a temporary ST
	tempST=SymbolTable()
	for i in st.getDeclarations():
	    # self-defined deepcopy
	    j=i.deepCopy()
	    if visibility=='PROTECTED':
		j.setVisibility('PROTECTED')
	    tempST.add(str(j.getName()), j)
	# if there is no modif, just make copies
	if modif==None:
	    for i in tempST.getDeclarations():
            	if visibility=='PROTECTED':
                    i.setVisibility('PROTECTED')
            	classDef.getScope().add(str(i.getName()), i)
            # copy equations
            for i in baseClass.getInitEquations():
            	classDef.initEquationPart.append(i)
            for i in baseClass.getEquations():
            	classDef.equationPart.append(i)
            for i in baseClass.getInitAlgorithm():
            	classDef.initAlgorithmPart.append(i)
            for i in baseClass.getAlgorithm():
            	classDef.algorithmPart.append(i)
	else: # has class modificaions
	    # Append modif to the copies of declarations from base class
	    # Create attribute 'self'
	    for i in modif.getArgumentList():
	    	# i: ASTComponent::ElementModification or ElementRedeclaration
	    	if isinstance(i, ElementRedeclaration):
		    raise 'Error: \"redeclare\" not yet supported!'
	    	else: # ElementModification
		    compRef1=i.getCompRef()
		    firstIdent=compRef1.getOperands()[0].getOperands()[0]
		    # the first identifier of compRef is looked up in the 
		    # temporary ST
		    found=tempST.localLookup(str(firstIdent))
		    if found==None:
		    	raise 'Error: illegal modification of base class!'
		    # append modif 
		    if len(compRef1.getOperands())==1:
		    	if isinstance(i.getModification(), \
				ClassModification):
		    	    found.addModif(i.getModification())
		        elif isinstance(i.getModification(), \
				EqualsModification):
			    found.addModif(i.getModification())
		    	else:
			    raise 'Error: AssignModification not suppoted!'
		    else:
		    	# remove the first identifier in the CompRef in modif
		     	compRef2=copy.deepcopy(compRef1)
		    	del compRef2.operands[0]
		    	# make a copy of the modif using default shallow copy
		    	newEM=copy.copy(i)
		    	newEM.setCompRef(compRef2)
		    	newCM=ClassModification(-1,-1,[newEM],None)
		    	# append the modif
		    	found.addModif(newCM)
	    # copy declarations from the temporary ST 
	    # into the inheriating class
	    for i in tempST.getDeclarations():
            	if visibility=='PROTECTED':
                    i.setVisibility('PROTECTED')
                classDef.getScope().add(str(i.getName()), i)
	    # copy equations
	    for i in baseClass.getInitEquations():
            	classDef.initEquationPart.append(i)
            for i in baseClass.getEquations():
            	classDef.equationPart.append(i)
            for i in baseClass.getInitAlgorithm():
            	classDef.initAlgorithmPart.append(i)
            for i in baseClass.getAlgorithm():
            	classDef.algorithmPart.append(i)


    # lookup algorithm for the base class name in an extends clause
    # The base class name can be a simple name, e.g. A,
    # or composite name, e.g. A.B.C
    # sp: ASTScope::SymbolTable
    # exp: ASTExp::NameExp
    def baseClassLookup(self, sp, exp):
        operands=exp.getOperands()
        numOperands=len(operands)
        # look up the first identifier in a NameExp, eg, A in A.B.C
        firstIdent=operands[0]
        # names are first searched from pre-defined types
        found=self.predefinedScope.localLookup(str(firstIdent))
        if (not found==None) and numOperands==1 :
            return found
        if self.isEncapsulated:
            # can lookup among imported elements
            raise EncapsulatedInheritanceException(str(exp))
        # normal look up: lookup sequentially in each member of
        # the ordered set of parents
        else:
            found=sp.lookup(str(firstIdent))
            # the found entry must be a DefinitionEntry
            if isinstance(found, DefinitionEntry):
                localScope=found.getDefinition().getScope()
            else:
                raise NameNotDeclaredException(str(exp))
        # continue look up if exp is a composite name
        # eg, look up B, C in name A.B.C
        i=1
        while i<numOperands:
            isPackage=found.getDefinition().isPackage()
            found=localScope.localLookup(str(operands[i]))
            if isinstance(found, DefinitionEntry):
                if not isPackage:
                    # lookup is restricted to encapsulated elements only
                    if not found.getDefinition().isEncapsulated():
                        raise UnencapsulatedElementException(str(operands[i]))
                localScope=found.getDefinition().getScope()
            else:
                raise NameNotDeclaredException(str(exp))
            i=i+1
        return found

