# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University


import dircache

from AST.ASTDefinition import ShortClassDefinition
from AST.ASTScope import ComponentPrefix, DefinitionEntry
from AST.ASTScope import DeclarationEntry, SymbolTable
from AST.ASTExp import IdentExp
from FrontEnd.Exception.NameException import NameNotDeclaredException, \
					     UnencapsulatedElementException
from FrontEnd.Exception.ScopeException import EncapsulatedInheritanceException
from FrontEnd.Parser import Parser
from FrontEnd.Visitor import AbstractVisitor
from FrontEnd.Scoper import Scoper

# ---------------------------------------------------------------------

class Lookuper1(AbstractVisitor):
    "type specifier lookup"
    def __init__(self, ast, MODELICAPATH, parserPath, parserName):#, filePath):
	self.parser=Parser()
	self.ast=ast
	#self.scopingVisitor=Scoper()
	self.MODELICAPATH=MODELICAPATH # a list [] of String
	self.parserPath=parserPath
	self.parserName=parserName
	#self.filePath=filePath
	self.currentPath=None
	# the scope node which stores all builtin primitive types 
	self.builtinScope=ast.builtinScope
	# the scope node which stores all predefined primitive types
	self.predefinedScope=ast.predefined.classDefinitions[0].scope
	self.currentScope=None
	self.currentEntry=None
	# a flag indicating if the currentEntry is an Encapsulated class
	self.isEncapsulated=False
	# flag: indicates whether a ShortClassDefinition is being visited
        self.isVisitingShortClass=False
	# the debugging message
        self.msg=''


    def setCurrentEntry(self, e):
	self.currentEntry=e

    def setCurrentScope(self, scope):
	self.currentScope=scope

    def setBuiltinScope(self, scope):
	self.builtinScope=scope

    def setPredefinedScope(self, scope):
	self.predefinedScope=scope


    # -----------------------------------------------------------------
    def visitClassFile(self, classFile):
	#self.setPredefinedScope(classFile.getPredefinedScope())
	#self.setBuiltinScope(classFile.getBuiltinScope())
	#file.getScope().accept(self)
	#print "current path: ", classFile.dir
	self.currentPath=classFile.dir
	for i in classFile.getClassDefinitions():
                i.accept(self)

    def visitRegularClassDefinition(self, classDef):
	# visit extends clauses 
	for i in classDef.getExtendsClauses():
	    i.accept(self)
  	# visit symbol table
	classDef.getScope().accept(self)

    def visitShortClassDefinition(self, classDef):
	self.isVisitingShortClass=True
	classDef.getBaseName().accept(self)
	self.isVisitingShortClass=False

    def visitSymbolTable(self, st):
	self.setCurrentScope(st)
	for i in st.getTable().keys():
	    st.getTable()[i].accept(self)

    # type_specifiers are looked up
    def visitDeclarationEntry(self, e):
	self.setCurrentEntry(e)
	e.type.accept(self) # NameExp

    
    def visitDefinitionEntry(self, e):
	if str(e.getName())=='self':
	    return
	self.msg=self.msg+'\n In class: '+str(e.getName())
	#if isinstance(e.getDefinition(), ShortClassDefinition):
	# base class name look up
	if e.getDefinition().isEncapsulated():
	    self.isEncapsulated=True
	e.getDefinition().getScope().accept(self)
	self.isEncapsulated=False
	#else: # RegularClassDefinition & EnumClassDefinition
	#    if e.getDefinition().isEncapsulated():
	#    	self.isEncapsulated=True
	#    e.getDefinition().getScope().accept(self) # SymbolTable
	#    self.isEncapsulated=False
	self.msg=self.msg+'\n end of class: '+str(e.getName())

    def visitExtendsClause(self, extendsClause):
        found=self.baseClassLookup(self.currentScope, extendsClause.getName())
	if found == None or not isinstance(found, DefinitionEntry):
           raise NameNotDeclaredException(str(exp))
	#print "base class: "+str(extendsClause.getName())+" is found."

    def visitIdentExp(self, exp):
	# this method won't be called because 'type' attibute is of 
	# type NameExp 
	return
    
    # lookup for type specifiers
    def visitNameExp(self, exp):
	operands=exp.getOperands()
	# lookup of first identifier
	#if(str(exp)=='Generator'):
        #  print self.currentScope
	found=self.lookup(self.currentScope, operands[0])
	if found == None or not isinstance(found, DefinitionEntry):
	   raise NameNotDeclaredException(str(exp))
	# lookup of the remaining part of the name
	index=1
	numOperands=len(operands)
	while (index<numOperands):
	    # the rest of the name is looked up among declared classes
	    found=self.lookupInScope(found.getDefinition().getScope(), \
				operands[index])
	    if found == None or not isinstance(found, DefinitionEntry):
		raise NameNotDeclaredException(str(exp))
	    index=index+1
	if not self.isVisitingShortClass:
	    self.currentEntry.setDefinition(found.getDefinition())
	self.msg=self.msg+'\n Type Specifier: "'+str(exp)+'" is found;'


    # the lookup process of a simple name, starting from scope sp
    # sp: ASTScope::SymbolTable
    # exp: ASTExp::IdentExp
    def lookup(self, sp, exp):
	found=self.builtinScope.localLookup(str(exp))
        if not found ==None:
            return found
	# lookup in pre-defined scope
	found=self.predefinedScope.localLookup(str(exp))
	if not found ==None:
	    return found
	# lookup sequentially in the ordered set of parents
        while (True):
            found=self.lookupInScope(sp, exp)
            if not found==None:
                return found
            if not sp.getOuter()==None:
                # lookup stops if an encapsulated parent is found
                currentClass=sp.localLookup('self').getDefinition()
                if currentClass.isEncapsulated():
                    break
                sp=sp.getOuter()
            # lookup stops if reaching the top parent
            else:
                break
	# look up from external files
	return self.externalLookup(exp)

    def externalLookup(self, exp):
	#print "external lookup..."
	classFile=None
        scopingVisitor=Scoper(self.ast, self.MODELICAPATH, self.parserPath, self.parserName)
	# first lookup in currentPath
	fileList=dircache.listdir(self.currentPath)
	#print "file list: ", fileList
	if str(exp)+'.mo' in fileList:
	    classFile=self.parser.parse(self.parserPath, \
                        self.parserName, self.currentPath, \
                        str(exp)+'.mo', str(exp)+'.agl', str(exp)+'.mo')
        for i in self.MODELICAPATH:
	    # get all filenames in directory i
            a = dircache.listdir(i)
            if str(exp)+'.mo' in a:
                aglname = filename[:-3] + ".agl"
                filename = "lib/" + f
                classFile=self.parser.parse(self.parserPath, \
                        self.parserName, self.currentPath, \
                        str(exp)+'.mo', str(exp)+'.agl', str(exp)+'.mo')
                        #filename, aglname, filename)
                break
	    # if str(exp) is a directory name
            elif str(exp) in a:
		#print str(exp)+" is a dir."
                # load str(exp)/Package.mo
                classFile=None # TBD
                break
        if not classFile==None:
            classFile.accept(scopingVisitor)
            self.ast.addClassFile(classFile)
            return self.ast.topScope.localLookup(str(exp))
            #return classFile.getClassDefinitions()[0]
        return None

    # the lookup precess in a scope sp
    # sp: ASTScope::SymbolTable
    # exp: ASTExp::IdentExp
    def lookupInScope(self, sp, exp):
	# lookup from locally declared elements
	found=sp.localLookup(str(exp))
        if not found ==None:
            return found
	# if not found, then perform lookup from inherited elements
	if sp.getOuter()==None: # has reached the top scope
	    return None
	for i in sp.localLookup('self').getDefinition().getExtendsClauses():
	    baseClass=self.baseClassLookup(sp, i.getName()).getDefinition()
	    found=self.lookupInScope(baseClass.getScope(), exp)
	    if not found==None:
		return found
	return None

    # lookup algorithm for the base class name in an extends clause
    # The base class name can be a simple name, e.g. A,
    # or composite name, e.g. A.B.C
    # sp: ASTScope::SymbolTable
    # exp: ASTExp::NameExp
    def baseClassLookup(self, sp, exp):
        operands=exp.getOperands()
	numOperands=len(operands)
        # look up the first identifier in a NameExp, eg, A in A.B.C
        firstIdent=operands[0]
        # names are first searched from pre-defined types
	found=self.predefinedScope.localLookup(str(firstIdent))
	if (not found==None) and numOperands==1 :
	    return found
        if self.isEncapsulated:
	    # can lookup among imported elements
	    raise EncapsulatedInheritanceException(str(exp))
	# normal look up: lookup sequentially in each member of 
	# the ordered set of parents, 
	# or from external files
        else: 
            found=sp.lookup(str(firstIdent))
	    if found==None:
		# this only applies to simple names, need to be changed
		found=self.externalLookup(str(firstIdent))
            # the found entry must be a DefinitionEntry
            if isinstance(found, DefinitionEntry):
	    	localScope=found.getDefinition().getScope()
            else:
            	raise NameNotDeclaredException(str(exp))
	# continue look up if exp is a composite name
	# eg, look up B, C in name A.B.C
	i=1
        while i<numOperands:
            isPackage=found.getDefinition().isPackage()
            found=localScope.localLookup(str(operands[i]))
            if isinstance(found, DefinitionEntry):
                if not isPackage:
                    # lookup is restricted to encapsulated elements only
                    if not found.getDefinition().isEncapsulated():
                        raise UnencapsulatedElementException(str(operands[i]))
                localScope=found.getDefinition().getScope()
            else:
		raise NameNotDeclaredException(str(exp))
            i=i+1
	return found



