# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the visitor module


from FrontEnd.TypeChecker1 import TypeChecker
from AST.ASTEquations import *
from AST.ASTScope import *
# ---------------------------------------------------------------------
class TypeCheckVisitor:
    "type check visitor"

    def __init__(self, ast):
        # the AST (ClassFile)
        self.ast=ast
	self.msg=""

    def go(self):
	mainModel=self.ast.classFiles[0].classDefinitions[0]
	for i in mainModel.getFlattenedInitEqs():
	    if not self.getEqType(i)=='Void':
		#print str(i), ' --has type errors ...'
		self.msg=self.msg+str(i)+' --has type errors ...\n'
		# throw exception here
	    else:
		#print str(i), ' --is type correct ...'
		self.msg=self.msg+str(i)+' --is type correct ...\n'
	for i in mainModel.getFlattenedEqs():
	    if not self.getEqType(i)=='Void':
		#print str(i), ' --has type errors ...'
		# throw exception here
		self.msg=self.msg+str(i)+' --has type errors ...\n'
	    else:
		#print str(i), ' --is type correct ...'
		self.msg=self.msg+str(i)+' --is type correct ...\n'
	return self.msg

    # eq: AST::EquationStm
    def getEqType(self, eq):
	if isinstance(eq, RegularEqStm):
	    typeLHS=self.getExpType(eq.getLHS())
	    typeRHS=self.getExpType(eq.getRHS())
	    if typeLHS==typeRHS:
            	return 'Void'
            elif typeLHS=='RealType' and typeRHS=='IntegerType':
            	#type coercion
            	#newExp=ConversionExp([self.RHS])
            	#self.RHS=newExp
            	return 'Void'
	    elif typeLHS=='IntegerType' and typeRHS=='RealType':
            	#type coercion
            	#newExp=ConversionExp([self.LHS])
            	#self.LHS=newExp
            	return 'Void'
            else:
		return 'TypeError'
	elif isinstance(eq, ConnectEqStm):
	    # TBD
	    return 'TypeError'
	elif isinstance(eq, FunCallEqStm):
	    # TBD
            return 'TypeError'
	elif isinstance(eq, ConditionalEqStm):
	    # TBD
            return 'TypeError'
	elif isinstance(eq, ForEqStm):
	    # TBD
            return 'TypeError'
	elif isinstance(eq, WhenEqStm):
	    # TBD
            return 'TypeError'
	else:
	    return 'TypeError'

    def getExpType(self, exp):
	if isinstance(exp, IdentExp):
	    result=str(exp.getValue().getType())
            if result=='Real':
            	return 'RealType'
            elif result=='Integer':
            	return 'IntegerType'
            elif result=='String':
            	return 'StringType'
            elif result=='Boolean':
            	return 'BooleanType'
            else:
            	return 'TypeError'
	elif isinstance(exp, IntegerExp):
	    return 'IntegerType'
	elif isinstance(exp, RealExp):
	    return 'RealType'
	elif isinstance(exp, BoolExp):
	    return 'BooleanType'
	elif isinstance(exp, StringExp):
	    return 'StringType'
	elif isinstance(exp, NameExp):
	    raise 'unexpected NameExp in equation part ... abort!'
	elif isinstance(exp, OrExp) or isinstance(exp, AndExp):
	    operands=exp.getOperands()
	    if self.getExpType(operands[0])=='BooleanType' and \
               self.getExpType(operands[1])=='BooleanType':
                return 'BooleanType'
            return 'TypeError'
	elif isinstance(exp, NotExp):
	    operands=exp.getOperands()
	    if self.getExpType(operands[0])=='BooleanType':
		return 'BooleanType'
	    return 'TypeError'
	elif isinstance(exp, LessExp) or \
	     isinstance(exp, LessEQExp) or \
	     isinstance(exp, GreaterExp) or \
	     isinstance(exp, GreaterEQExp) or \
	     isinstance(exp, EQEQExp) or \
	     isinstance(exp, LessGtExp):
	    operands=exp.getOperands()
	    if self.getExpType(operands[0])=='IntegerType' and \
               self.getExpType(operands[1])=='IntegerType':
            	return 'BooleanType'
            elif self.getExpType(operands[0])=='RealType' and \
                 self.getExpType(operands[1])=='RealType':
            	return 'BooleanType'
            elif self.getExpType(operands[0])=='StringType' and \
                 self.getExpType(operands[1])=='StringType':
            	return 'BooleanType'
            elif self.getExpType(operands[0])=='BooleanType' and \
                 self.getExpType(operands[1])=='BooleanType':
            	return 'BooleaType'
	    # type coercion
            elif self.getExpType(operands[0])=='RealType' and \
                 self.getExpType(operands[1])=='IntegerType':
                #newExp=ConversionExp([self.operands[1]])
                #self.operands[1]=newExp
            	return 'BooleanType'
            elif self.getExpType(operands[0])=='IntegerType' and \
                 self.getExpType(operands[1])=='RealType':
            	#newExp=ConversionExp([self.operands[0]])
            	#self.operands[0]=newExp
            	return 'BooleanType'
            else:
            	return 'TypeError'
	elif isinstance(exp, SumExp):
	    operands=exp.getOperands()
            if self.getExpType(operands[0])=='IntegerType' and \
               self.getExpType(operands[1])=='IntegerType':
            	return 'IntegerType'
	    elif self.getExpType(operands[0])=='RealType' and \
                 self.getExpType(operands[1])=='RealType':
            	return 'RealType'
	    elif self.getExpType(operands[0])=='StringType' and \
                 self.getExpType(operands[1])=='StringType':
            	return 'StringType'
            # type coercion
	    elif self.getExpType(operands[0])=='RealType' and \
                 self.getExpType(operands[1])=='IntegerType':
            	#newExp=ConversionExp([self.operands[1]])
            	#self.operands[1]=newExp
            	return 'RealType'
	    elif self.getExpType(operands[0])=='IntegerType' and \
                 self.getExpType(operands[1])=='RealType':
            	#newExp=ConversionExp([self.operands[0]])
            	#self.operands[0]=newExp
            	return 'RealType'
            else:
            	return 'TypeError'
	elif isinstance(exp, MulExp) or \
	     isinstance(exp, SubtractExp) or \
	     isinstance(exp, DivExp) or \
	     isinstance(exp, PowerExp):
	    operands=exp.getOperands()
	    type1=self.getExpType(operands[0])
	    type2=self.getExpType(operands[1])
            if type1=='IntegerType' and type2=='IntegerType':
                return 'IntegerType'
            elif type1=='RealType' and type2=='RealType':
                return 'RealType'
	    # type coercion
            elif type1=='RealType' and type2=='IntegerType':
                #newExp=ConversionExp([self.operands[1]])
                #self.operands[1]=newExp
                return 'RealType'
            elif type1=='IntegerType' and type2=='RealType':
                #newExp=ConversionExp([self.operands[0]])
                #self.operands[0]=newExp
                return 'RealType'
            else:
                return 'TypeError'
	elif isinstance(exp, UnaryMinusExp):
	    operands=exp.getOperands()
	    if self.getExpType(operands[0])=='RealType':
            	return 'RealType'
	    elif self.getExpType(operands[0])=='IntegerType':
            	return 'IntegerType'
            else:
            	return 'TypeError'
	elif isinstance(exp, IfElseExp):
	    if not self.getExpType(exp.getIfTest())=='BooleanType':
            	return 'TypeError'
            # type might be determined at runtime
	elif isinstance(exp, FunCallExp):
	    # currently this only handles the 'der' operator (function)
	    operands=exp.getOperands()
	    if str(operands[0])=='der':
		args=operands[1].getOperands()
		if len(args)==1:
		    return self.getExpType(args[0])
	    else:
		return 'TypeError'
	    #else: # need to know the return type of a function
		#args=operands[1].getOperands()
		#return self.getExpType(args[len(args)-1])
	elif isinstance(exp, SingleCompRefExp):
	    return self.getExpType(exp.getOperands()[0])
            #if result=='Real':
            #	return 'RealType'
            #elif result=='Integer':
            #	return 'IntegerType'
            #elif result=='String':
		#return 'StringType'
            #elif result=='Boolean':
            	#return 'BooleanType'
            #else:
            #	return 'TypeError'
	elif isinstance(exp, CompRefExp):
	    return self.getExpType(exp.getLastOperand())
	elif isinstance(exp, ExpListExp):
	    operands=exp.getOperands()
	    if exp.getKind()=='PAR' and len(operands)==1:
            	return self.getExpType(operands[0])
	elif isinstance(exp, ConversionExp):
	    return 'RealType'
	else:
	    raise 'type check to be done in this kind of exp ...'
