# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University



# ---------------------------------------------------------------------

class TypeChecker:
    "Type Checker"
    def __init__(self):
	pass

    # ------------------Subtyping & Type Equivalence------------------
    # c1, c2: ASTScope::TableEntry
    # return True if c1 is subtype of c2
    def isClassSubtype(self, c1, c2):
	if self.isTypeEquivalent(c1, c2):
	    return True
	pubElems1=c1.getDefinition().getScope().getPublicDeclarations()
	pubElems2=c2.getDefinition().getScope().getPublicDeclarations()
	for i in pubElems1:
	    j=self.getSameNameElem(i, pubElems2)
	    if (j==None) or (not self.isComponentSubtype(i,j)):
		return False
	return True

    # c1, c2: ASTScope::TableEntry
    # return True if c1 is subtype of c2
    def isComponentSubtype(self, c1, c2):
	if not self.hasSameNumOfDim(c1,c2):
	    return False
	if not self.isClassSubtype(c1,c2):
	    return False
	# need dimension check here
	return True

    # c1, c2: ASTScope::DeclarationEntry
    # return True if c1 and c2 are type equivalent
    def isTypeEquivalent(self, c1, c2):
	if (str(c1.type)=='RealType' and str(c2.type)=='RealType') or \
	   (str(c1.type)=='IntegerType' and str(c2.type)=='IntegerType') or\
	   (str(c1.type)=='BooleanType' and str(c2.type)=='BooleanType') or\
	   (str(c1.type)=='StringType' and str(c2.type)=='StringType') or \
	   (str(c1.type)=='StateSelect' and str(c2.type)=='StateSelect'):
	    return True
	pubElems1=c1.getDefinition().getScope().getPublicDeclarations()
        pubElems2=c2.getDefinition().getScope().getPublicDeclarations()
	if not (len(pubElems1)==len(pubElems2)):
	    return False
	for i in pubElems1:
	    j=self.getSameNameElem(i, pubElems2)
	    if (j==None) or (not self.isTypeEquivalent(i,j)):
		return False
	return True

    # c1, c2: ASTScope::ClassDefinition
    # return True if c1 and c2 are type equivalent
    def isClassTypeEquivalent(self, c1, c2):
	pubElems1=c1.getScope().getPublicDeclarations()
	pubElems2=c2.getScope().getPublicDeclarations()
	if not (len(pubElems1)==len(pubElems2)):
            return False
	for i in pubElems1:
	    j=self.getSameNameElem(i, pubElems2)
	    if (j==None) or (not self.isTypeEquivalent(i,j)):
		return False
        return True


    #  c1, c2: ASTScope::TableEntry
    # return True if c1 and c2 are identical elements
    def isTypeIdentical(self, c1, c2):
	if not self.isTypeEquivalent(c1, c2):
	    return False
	isFinal1=c1.getComponentPrefix().getFinal()
	isFinal2=c2.getComponentPrefix().getFinal()
	if not isFinal1==isFinal2:
	    return False
	tp1=c1.getTypePrefix()
	tp2=c2.getTypePrefix()
	if (not tp1==None) and (not tp2==None):
	    if not tp1.isIdenticalTo(tp2):
		return False
	if self.isBuiltinType(c1) and self.isBuiltinType(c2):
	    if not str(c1.getType())==str(c2.getType()):
		return False
	    return True
	pubElems1=c1.getDefinition().getScope().getPublicDeclarations()
	pubElems2=c2.getDefinition().getScope().getPublicDeclarations()
	if not (len(pubElems1)==len(pubElems2)):
	    return False
	for i in pubElems1:
	    j=self.getSameNameElem(i, pubElems2)
	    if (j==None) or (not self.isTypeIdentical(i,j)):
		return False
	return True

    # ------------------------ auxilary methods ----------------------
    def getSameNameElem(self, elem, elemList):
	name1=str(elem.getName())
	for i in elemList:
	    name2=str(i.getName())
	    if name1==name2:
		return i
	return None

    # elem: ASTScope::DeclarationEntry
    def isBuiltinType(self, elem):
	if str(elem.getType())=='RealType' or \
		str(elem.getType())=='IntegerType' or \
		str(elem.getType())=='BooleanType' or \
		str(elem.getType())=='StringType' or \
		str(elem.getType())=='StateSelect':
		return True
	return False

    # ---------------Type check for equations and Exps---------------

    # types: RealType, IntegerType, BooleanType, StringType, Void, 
    #        TypeError
