# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

from AST.ASTEquations import *
from AST.ASTScope import *
# ---------------------------------------------------------------------
class CopyMaker:
    "create shallow and deep copies of AST enties"
    def __init__(self):
	pass

    def deepcopySTDecl(self, decl):
	pass
	raise "TBD"

    # make a deepcopy of an equation
    # eq: AST::EquationStm
    def deepcopyEqStm(self, eq):
	if isinstance(eq, RegularEqStm):
	    lhs=self.deepcopyExp(eq.getLHS())
	    rhs=self.deepcopyExp(eq.getRHS())
	    return RegularEqStm(eq.getLineNum(), eq.getColNum(), lhs, rhs)
	elif isinstance(eq, ConnectEqStm):
	    firstConnector=self.deepcopyExp(eq.getFirstConnector())
	    secondConnector=self.deepcopyExp(eq.getSecondConnector())
	    return ConnectEqStm(eq.getLineNum(), eq.getColNum(), \
			firstConnector, secondConnector)
	elif isinstance(eq, FunCallEqStm):
	    funCall=self.deepcopyExp(eq.getFunCall())
	    return FunCallEqStm(eq.getLineNum(), eq.getColNum(), funCall)
	elif isinstance(eq, ConditionalEqStm):
	    ifTest=self.deepcopyExp(eq.getIfTest())
	    ifBody=[]
	    for i in eq.getIfBody():
		ifBody.append(self.deepcopyEqStm(i))
	    elseifList=[]
	    for i in eq.getElseifList():
		elseifList.append(self.deepcopyEqStm(i))
	    elseBody=[]
	    for i in eq.getElseBody():
		elseBody.append(self.deepcopyEqStm(i))
            return ConditionalEqStm(eq.getLineNum(), eq.getColNum(),\
			ifTest, ifBody, elseifList, elseBody)
	elif isinstance(eq, ForEqStm):
	    forIndices=self.deepcopyExp(eq.getForIndices())
	    body=[]
	    for i in eq.getBody():
		body.append(self.deepcopyEqStm(i))
	    # the scope node is not copied
            return ForEqStm(eq.getLineNum(), eq.getColNum(), forIndices, body)
	elif isinstance(eq, WhenEqStm):
	    whenTest=self.deepcopyExp(eq.getWhenTest())
	    body=[]
	    for i in eq.getBody():
		body.append(self.deepcopyEqStm(i))
	    elsewhenList=[]
	    for i in eq.getElsewhenList():
		elsewhenList.append(self.deepcopyEqStm(i))
            return WhenEqStm(eq.getLineNum(), eq.getColNum(), \
			whenTest, body, elsewhenList)
	else:
	    raise "Exception: Not a legal EquationStm!"

    # make a deepcopy of an exp
    def deepcopyExp(self, exp):
	if isinstance(exp, IdentExp):
	    return IdentExp(exp.getLineNum(), exp.getColNum(), exp.getValue())
	elif isinstance(exp, IntegerExp):
	    return IntegerExp(exp.getLineNum(), exp.getColNum(), \
		exp.getValue())
	elif isinstance(exp, RealExp):
	    return RealExp(exp.getLineNum(), exp.getColNum(), exp.getValue())
	elif isinstance(exp, BoolExp):
	    return BoolExp(exp.getLineNum(), exp.getColNum(), exp.getValue())
	elif isinstance(exp, StringExp):
	    return StringExp(exp.getLineNum(), exp.getColNum(), exp.getValue())
	elif isinstance(exp, NameExp):
	    operands=[]
	    for i in exp.getOperands():
		operands.append(self.deepcopyExp(i))
	    return NameExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, OrExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return OrExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, AndExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return AndExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, NotExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return NotExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, LessExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return LessExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, LessEQExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return LessEQExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, GreaterExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return GreaterExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, GreaterEQExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return GreaterEQExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, EQEQExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return EQEQExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, LessGtExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return LessGtExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, SumExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return SumExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, MulExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return MulExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, SubtractExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return SubtractExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, DivExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return DivExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, PowerExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return PowerExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, UnaryMinusExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return UnaryMinusExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, RangeExp):
            operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return RangeExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, IfElseExp):
	    ifTest=self.deepcopyExp(exp.getIfTest())
	    ifBody=self.deepcopyExp(exp.getIfBody())
	    elseifList=[]
	    for i in exp.getElseifList():
		elseifList.append(self.deepcopyExp(i))
	    elseBody=self.deepcopyExp(exp.getElseBody())
	    return IfElseExp(exp.getLineNum(), exp.getColNum(), \
			ifTest, ifBody, elseifList, elseBody)
	elif isinstance(exp, FunCallExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return FunCallExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, FunArgsExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return FunArgsExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, ForFunArgExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ForFunArgExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, NamedArgExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return NamedArgExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, ConnectRefExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ConnectRefExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, SingleCompRefExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return SingleCompRefExp(exp.getLineNum(), exp.getColNum(), \
			operands)
	elif isinstance(exp, CompRefExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return CompRefExp(exp.getLineNum(), exp.getColNum(), operands)	
	elif isinstance(exp, ArrSubExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ArrSubExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, CommentExp):
            operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return CommentExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, AnnotationExp):
	    # AnnotationExp is not copied
	    return None 
	elif isinstance(exp, IdentInExpExp):
            operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return IdentInExpExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, ForIndicesExp):
            operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ForIndicesExp(exp.getLineNum(), exp.getColNum(), operands)
	elif isinstance(exp, ExpListExp):
	    kind=exp.getKind()
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ExpListExp(exp.getLineNum(), exp.getColNum(), \
		operands, kind)
	elif isinstance(exp, ConversionExp):
	    operands=[]
            for i in exp.getOperands():
                operands.append(self.deepcopyExp(i))
            return ConversionExp(exp.getLineNum(), exp.getColNum(), operands)
	else:
	    raise 'Exp can not be copied! '
