# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the pretty printer of canonical repesentation


from AST.ASTExp import *
from AST.ASTEquations import *
from AST.ASTScope import DeclarationEntry, SymbolTable

# ---------------------------------------------------------------------
class EqExpPrinter:
    def __init__(self):
	pass

    # currently only dump RegularEquationStm
    def dumpEq(self, eq):
	result=''
	if isinstance(eq, RegularEqStm):
	    return self.dumpExp(eq.getLHS())+"="+self.dumpExp(eq.getRHS())
	elif isinstance(eq, ConnectEqStm):
	    return str(eq)
	else:
	    print "Only Regular equations can be pretty printed!"
	    return result

    # indentLevel: int
    # indent: string of spaces
    def dumpIndentEq(self, eq, indentLevel, indent):
	indentation=indentLevel*indent
	if isinstance(eq, RegularEqStm):
	    return indentation+self.dumpExp(eq.getLHS())+"="+ \
			self.dumpExp(eq.getRHS())
			
        #TODO
        if eq.__class__.__name__ == "AtomicAlgStm1":
	#if isinstance(eq, AtomicAlgStm1):  #added by Hongyan Song, Mar 27, 2006
	    return eq.dump(indentLevel)

	elif isinstance(eq, ConnectEqStm):
	    return indentation+str(eq)
	elif isinstance(eq, FunCallEqStm):
	    return indentation+self.dumpExp(eq.getFunCall())
	elif isinstance(eq, ConditionalEqStm):
	    #print "Conditional"
	    result=indentation+'if '+self.dumpExp(eq.getIfTest()) \
		+' then\n'
	    indentLevel=indentLevel+1
	    for i in eq.getIfBody():
		res = self.dumpIndentEq(i, indentLevel, indent)
		if(res!=None):
		  result=result+self.dumpIndentEq(i, indentLevel, indent) \
		        + ';\n'
	    indentLevel=indentLevel-1
	    if len(eq.getElseifList())>0:
		#print eq.getElseifList()
		for j in eq.getElseifList():
		    tempEqStr=self.dumpIndentEq(j, indentLevel, indent)
		    result=result+indentation+'else\n'+ \
			tempEqStr[len(indent)*indentLevel:\
			 len(tempEqStr)-len(indent)*indentLevel-6]
	    if len(eq.getElseBody())>0:
		result=result+indentation+'else\n'
		indentLevel=indentLevel+1
		for k in eq.getElseBody():
		    #print k.__class
		    kr = self.dumpIndentEq(k, indentLevel, indent)
		    if(kr != None):
		      result=result+self.dumpIndentEq(k, indentLevel, indent) \
                          + ';\n'

		indentLevel=indentLevel-1
	    result=result+indentation+'end if'
	    if not eq.comment==None:
            	result=result+' '+str(eq.comment)
            return result
	elif isinstance(eq, ForEqStm):
	    pass
	elif isinstance(eq, WhenEqStm):
	    pass

    def dumpExp(self, exp):
	#print 'in: ', str(exp)
	if isinstance(exp, IdentExp) or \
           isinstance(exp, BoolExp) or \
           isinstance(exp, StringExp) or \
	   isinstance(exp, NameExp):
	    return str(exp)
	elif isinstance(exp, IntegerExp) or isinstance(exp, RealExp):
	    if exp.getValue()<0:
		return '('+str(exp)+')'
	    else:
		return str(exp)
	elif isinstance(exp, OrExp):
	    result=''
	    for i in exp.getOperands():
		insertPar=self.hasLowerPrecedence(i, exp)
		if insertPar:
		    result=result+'('+self.dumpExp(i)+')'+' or'
		else:
		    result=result+self.dumpExp(i)+' or'
	    return result[0:len(result)-3]
	elif isinstance(exp, AndExp):
	    result=''
	    for i in exp.getOperands():
		insertPar=self.hasLowerPrecedence(i, exp)
		if insertPar:
		    result=result+'('+self.dumpExp(i)+')'+' and'
		else:
		    result=result+self.dumpExp(i)+' and'
	    return result[0:len(result)-4]
	elif isinstance(exp, NotExp):
            insertPar=self.hasLowerPrecedence(exp.getOperands()[0], exp)
	    if insertPar:
	        return 'not ('+self.dumpExp(exp.getOperands()[0])+')'
	    else:
		return 'not '+self.dumpExp(exp.getOperands()[0])
	elif isinstance(exp, LessExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
	    if insertPar1:
	        result='('+self.dumpExp(exp.getOperands()[0])+') < '
	    else:
		result=self.dumpExp(exp.getOperands()[0])+' < '
	    insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
	    if insertPar2:
		result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
	    else:
		result=result+self.dumpExp(exp.getOperands()[1])
	    return result
	elif isinstance(exp, LessEQExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+') <= '
            else:
                result=self.dumpExp(exp.getOperands()[0])+' <= '
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	elif isinstance(exp, GreaterExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+') > '
            else:
                result=self.dumpExp(exp.getOperands()[0])+' > '
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	elif isinstance(exp, GreaterEQExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+') >= '
            else:
                result=self.dumpExp(exp.getOperands()[0])+' >= '
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	elif isinstance(exp, EQEQExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+') == '
            else:
                result=self.dumpExp(exp.getOperands()[0])+' == '
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	elif isinstance(exp, LessGtExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+') <> '
            else:
                result=self.dumpExp(exp.getOperands()[0])+' <> '
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	elif isinstance(exp, SumExp):
	    result=''
	    for i in exp.getOperands():
		insertPar=self.hasLowerPrecedence(i, exp)
		if insertPar:
		    result=result+'('+self.dumpExp(i)+')'+'+'
		else:
		    result=result+self.dumpExp(i)+'+'
	    return result[0:len(result)-1]
	elif isinstance(exp, MulExp):
            result=''
            for i in exp.getOperands():
		insertPar=self.hasLowerPrecedence(i, exp)
		if insertPar:
		    result=result+'('+self.dumpExp(i)+')'+'*'
		else:
                    result=result+self.dumpExp(i)+'*'
            return result[0:len(result)-1]
	elif isinstance(exp, SubtractExp):
	    #raise 'SubtractExp should not exist in canonical form!'
	    result=''
	    operands=exp.getOperands()
	    result=result+self.dumpExp(operands[0])+'-'
	    if self.hasLowerPrecedence(operands[1], exp) or \
		isinstance(operands[1], SubtractExp):
		result=result+'('+self.dumpExp(operands[1])+')'
	    else:
		result=result+self.dumpExp(operands[1])
	    return result
	elif isinstance(exp, DivExp):
	    #raise 'DivExp should not exist in canonical form!'
	    result=''
            for i in exp.getOperands():
                insertPar=self.hasLowerPrecedence(i, exp)
                if insertPar:
                    result=result+'('+self.dumpExp(i)+')'+'/'
                else:
                    result=result+self.dumpExp(i)+'/'
            return result[0:len(result)-1]
	elif isinstance(exp, UnaryMinusExp):
	    # need check precedence to place the ( here
	    insertPar=self.hasLowerPrecedence(exp.getOperands()[0], exp)
	    if insertPar:
	        return '(-('+self.dumpExp(exp.getOperands()[0])+'))'
	    else:
		return '(-'+self.dumpExp(exp.getOperands()[0])+')'
	elif isinstance(exp, PowerExp):
            insertPar1=self.hasLowerPrecedence(exp.getOperands()[0], exp)
            if insertPar1:
                result='('+self.dumpExp(exp.getOperands()[0])+')^'
            else:
                result=self.dumpExp(exp.getOperands()[0])+'^'
            insertPar2=self.hasLowerPrecedence(exp.getOperands()[1], exp)
            if insertPar2:
                result=result+'('+self.dumpExp(exp.getOperands()[1])+')'
            else:
                result=result+self.dumpExp(exp.getOperands()[1])
            return result
	    #return self.dumpExp(exp.getOperands()[0])+'^'+ \
            #    self.dumpExp(exp.getOperands()[1])
	elif isinstance(exp, RangeExp):
	    if len(exp.getOperands())==0:
		return ':'
	    elif len(exp.getOperands())==2:
		return self.dumpExp(exp.getOperands()[0])+':'+ \
                	self.dumpExp(exp.getOperands()[1])
	    elif len(exp.getOperands())==3:
		return self.dumpExp(exp.getOperands()[0])+':'+ \
			self.dumpExp(exp.getOperands()[1])+':'+ \
			self.dumpExp(exp.getOperands()[2])
	elif isinstance(exp, IfElseExp):
	    raise 'IfElseExp should not exist in canonical form!'
	elif isinstance(exp, FunCallExp):
	    operands=exp.getOperands()
	    return self.dumpExp(operands[0]) + '(' \
		+self.dumpExp(operands[1]) + ')'
	elif isinstance(exp, FunArgsExp):
	    result=''
	    for i in exp.getOperands():
		result=result+self.dumpExp(i)+', '
	    return result[0:len(result)-2]
	elif isinstance(exp, ForFunArgExp):
	    operands=exp.getOperands()
	    result=self.dumExp(operands[0])+' for '
	    for i in range(1, len(operands)):
            	result=result+self.dumpExp(operands[i])+', '
            return result[0:len(result)-2]
	elif isinstance(exp, NamedArgExp):
	    operands=exp.getOperands()
	    return self.dumpExp(operands[0])+'='+self.dumpExp(operands[1])
	elif isinstance(exp, ConnectRefExp):
	    raise 'ConnectRefExp should not exist in canonical form!'
	elif isinstance(exp, SingleCompRefExp):
	    operands=exp.getOperands()
	    if len(operands)==1:
            	result=self.dumpExp(operands[0])
            if len(operands)==2:
            	result=self.dumpExp(operands[0])+self.dumpExp(operands[1])
            return result
	elif isinstance(exp, CompRefExp):
	    operands=exp.getOperands()
	    result=''
            for i in operands:
            	result=result+self.dumpExp(i)+'.'
            return result[0:len(result)-1]
	elif isinstance(exp, ArrSubExp):
	    raise 'ArrSubExp should not exist in canonical form!'
	elif isinstance(exp, CommentExp):
	    pass
	elif isinstance(exp, AnnotationExp):
	    pass
	elif isinstance(exp, IdentInExpExp):
	    raise 'IdentInExpExp should not exist in canonical form!'
	elif isinstance(exp, ForIndicesExp):
	    raise 'ForIndicesExp should not exist in canonical form!'
	elif isinstance(exp, ExpListExp):
	    kind=exp.getKind()
	    operands=exp.getOperands()
	    if kind=='PAR':
            	begin='('
            	end=')'
            elif kind=='BRACK':
            	begin='['
            	end=']'
            elif kind=='BRACE':
            	begin='{'
            	end='}'
            result=begin
            if kind=='BRACK':
            	for i in operands:
                    result=result+self.dumpExp(i)+'; '
            	result=result[0:len(result)-2]
            else:
            	for i in operands:
                    result=result+self.dumpExp(i)+','
            	result=result[0:len(result)-1]
            return result+end
	elif isinstance(exp, ConversionExp):
	    return 'Real('+self.dumpExp(exp.getOperands()[0])+')'
	else:
	    return str(exp)


    # compare the precedence of the type of 2 exps
    # if the precedence of the first arg is lower than the second one,
    # return True
    def hasLowerPrecedence(self, exp1, exp2):
	# assign precedence to exp1
	if isinstance(exp1, IfElseExp):
	    p1=1
	elif isinstance(exp1, OrExp):
	    p1=2
	elif isinstance(exp1, AndExp):
	    p1=3
	elif isinstance(exp1, NotExp):
	    p1=4
	elif isinstance(exp1, LessExp) or isinstance(exp1, LessEQExp) or \
	     isinstance(exp1, GreaterExp) or isinstance(exp1, GreaterEQExp) \
	     or isinstance(exp1, EQEQExp) or isinstance(exp1, LessGtExp):
	    p1=5
	#elif (isinstance(exp1, RealExp) or isinstance(exp1, IntegerExp)) \
        #        and exp1.getValue()<0:
        #    p1=6
	elif isinstance(exp1, SumExp):
	    p1=8
	elif isinstance(exp1, SubtractExp):
	    p1=9
	elif isinstance(exp1, MulExp):
	    p1=10
	elif isinstance(exp1, DivExp):
	    p1=11
	elif isinstance(exp1, PowerExp):
	    p1=12
	elif isinstance(exp1, UnaryMinusExp):
            p1=13
	else: 
	    p1=100 # dont care, a higher preceednce is assigned
	# assign precedence to exp2
	if isinstance(exp2, IfElseExp):
            p2=1
        elif isinstance(exp2, OrExp):
            p2=2
        elif isinstance(exp2, AndExp):
            p2=3
        elif isinstance(exp2, NotExp):
            p2=4
        elif isinstance(exp2, LessExp) or isinstance(exp2, LessEQExp) or \
             isinstance(exp2, GreaterExp) or isinstance(exp2, GreaterEQExp) \
             or isinstance(exp2, EQEQExp) or isinstance(exp2, LessGtExp):
            p2=5
	#elif (isinstance(exp2, RealExp) or isinstance(exp2, IntegerExp)) \
        #        and exp2.getValue()<0:
        #    p2=6
        elif isinstance(exp2, SumExp):
            p2=8
        elif isinstance(exp2, SubtractExp):
            p2=9
        elif isinstance(exp2, MulExp):
            p2=10
        elif isinstance(exp2, DivExp):
            p2=11
        elif isinstance(exp2, PowerExp):
            p2=12
	elif isinstance(exp2, UnaryMinusExp):
            p2=13
        else:
	    p2=100 #dont care, a higher preceednce is assigned
	# compare
	if p1 < p2:
	    return True
	return False

