# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

from AST.ASTExp import IdentExp
from AST.ASTScope import ComponentPrefix, DefinitionEntry
from AST.ASTScope import DeclarationEntry, SymbolTable
from FrontEnd.Visitor import AbstractVisitor
from FrontEnd.printers.EqExpPrinter import EqExpPrinter

# ---------------------------------------------------------------------

class ModelicaPrinter(AbstractVisitor):
    "print out AST to original the Modelica model"
    def __init__(self):
	self.indentLevel=0
	self.indent='  '
	self.output=''
	self.eqPrinter=EqExpPrinter()
	
    def getOutput(self):
        self.output += ";"
	return self.output

    # write to a file
    def writeToFile(self, file):
        modelicaFile=open(file, 'w')
        modelicaFile.write(self.output + ";")

    # ---------------------------ClassDefinition-----------------------
    def visitAST(self, ast):
	for i in ast.classFiles:
	    i.accept(self)

    def visitClassFile(self, classFile):
	if not classFile.withinClause==None:
	    classFile.withinClause.accept(self)
	for i in classFile.classDefinitions:
	    i.accept(self)
	    self.output=self.output+';\n\n'

    def visitWithinClause(self, wClause):
	self.output=self.output+self.indentLevel*self.indent+str(wClause)+\
		';\n'

    def visitRegularClassDefinition(self, r):
	# dump class prefix, class name ...
	indentation=self.indentLevel*self.indent
	self.output=self.output+indentation+str(r.prefix)
	self.output=self.output+str(r.kind)+' '+str(r.className)
	if not r.stringComment==None:
            self.output=self.output+' '+str(r.stringComment)
	self.output=self.output+'\n'
	# dump class elements
	self.indentLevel=self.indentLevel+1
	indentation=self.indentLevel*self.indent
	if len(r.elementList)>0:
	    visibility='DEFAULT'
	    for i in r.elementList:
		v=i.getVisibility()
		if not (v==visibility):
		    if not v=='DEFAULT':
			if v=='PUBLIC':
			    self.output=self.output+indentation+'public\n'
			else:
			    self.output=self.output+indentation+'protected\n'
		    visibility=v
		i.accept(self)
		self.output=self.output+';\n'
	self.indentLevel=self.indentLevel-1
	# dump initial equations
	indentation=self.indentLevel*self.indent
	if len(r.initEquationPart)>0:
	    self.output=self.output+indentation+'initial equation \n'
	    self.indentLevel=self.indentLevel+1
	    for i in r.initEquationPart:
		i.accept(self)
		self.output=self.output+';\n'
	    self.indentLevel=self.indentLevel-1
	# dump equations
	indentation=self.indentLevel*self.indent
        if len(r.equationPart)>0:
            self.output=self.output+indentation+'equation \n'
            self.indentLevel=self.indentLevel+1
            for i in r.equationPart:
                i.accept(self)
                self.output=self.output+';\n'
            self.indentLevel=self.indentLevel-1
	# dump initial alg
	indentation=self.indentLevel*self.indent
        if len(r.initAlgorithmPart)>0:
            self.output=self.output+indentation+'initial algorithm \n'
            self.indentLevel=self.indentLevel+1
            for i in r.initAlgorithmPart:
                i.accept(self)
                self.output=self.output+';\n'
            self.indentLevel=self.indentLevel-1
	# dumpl alg
	indentation=self.indentLevel*self.indent
        if len(r.algorithmPart)>0:
##            self.output=self.output+indentation+'initial algorithm \n'
            self.output=self.output+indentation+'algorithm \n' #modified by Hongyan Song Mar 27, 2006
            self.indentLevel=self.indentLevel+1
            for i in r.algorithmPart:
                i.accept(self)
                self.output=self.output+';\n'
            self.indentLevel=self.indentLevel-1
	# dump 'end' & class name
	self.output=self.output+indentation+'end ' +str(r.className)

    def visitShortClassDefinition(self, s):
	indentation=self.indentLevel*self.indent
	self.output=self.output+indentation+str(s)

    def visitEnumClassDefinition(self, eClassDefinition):	
	# EnumClassDefinition has been transformed to 
	# RegularClassDefinition
	pass

    def visitComponentClause(self, componentClause):
	indentation=self.indentLevel*self.indent
	self.output=self.output+indentation+str(componentClause)

    def visitReplaceableClause(self, rc):
	indentation=self.indentLevel*self.indent
	if not rc.classDefinition==None:
	    rc.classDefinition.accept(self)
	elif not rc.componentClause==None:
	    rc.componentClause.accept(self)
	elif not rc.constrainingClause==None:
	    self.output=self.output+str(rc.constrainingClause)
	if not rc.comment==None:
	    self.output=self.output+str(rc.comment)
	
	
    def visitImportClause(self, importClause):
	indentation=self.indentLevel*self.indent
	self.output=self.output+indentation+str(importClause)

    def visitExtendsClause(self, extendsClause):
	indentation=self.indentLevel*self.indent
	self.output=self.output+indentation+str(extendsClause)

    # ------------------------------EqStm-----------------------------
    def visitRegularEqStm(self, eq):
	#Eindentation=self.indentLevel*self.indent
	#self.output=self.output+indentation+self.eqPrinter.dumpEq(eq)
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
		self.indentLevel, self.indent)

    def visitConnectEqStm(self, eq):
	#indentation=self.indentLevel*self.indent
	#self.output=self.output+indentation+str(eq)
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
                self.indentLevel, self.indent)

    def visitFunCallEqStm(self, eq):
	#indentation=self.indentLevel*self.indent
	#self.output=self.output+indentation+str(eq)
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
                self.indentLevel, self.indent)

    def visitConditionalEqStm(self, eq):
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
                self.indentLevel, self.indent)

    def visitForEqStm(self, eq):
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
                self.indentLevel, self.indent)

    def visitWhileAlgStm(self, eq):
	self.output=self.output+eq.dump(self.indentLevel)

    def visitWhenEqStm(self, eq):
	self.output=self.output+self.eqPrinter.dumpIndentEq(eq, \
                self.indentLevel, self.indent)

    def visitAtomicAlgStm1(self, eq):  #Added by Hongyan Song, Mar 27, 2006
	self.output=self.output+eq.dump(self.indentLevel)
	
    def visitAnnotationExp(self, exp): #Added by Hongyan Song, Mar 28, 2006
	#print dir(exp)
	self.output=self.output+exp.dump(self.indentLevel)

