connector Pin "pin of an electric component"
	Voltage v "Potential at the pin";
	flow Current i "Current flowing into the pin";
end Pin;

partial model OnePort "Superclass of Components with two electrical pins p and n"
	Voltage v "Voltage drop between p and n";
	Current i "Current flowing from p to n";
	Pin p;
	Pin n;
equation
	v = p.v - n.v;
	0 = p.i + n.i;
	i = p.i;
end OnePort;

model Resistor "Ideal linear electrical resistor"
	extends OnePort;
	parameter Real r(unit="Ohm") "Resistance";
equation
	r*i = v;
end Resistor;

model Capacitor "Ideal electrical Capacitor"
	extends OnePort;
	parameter Real c(unit="F") "Capacitance";
equation
	c*der(v) = i;	
end Capacitor;

model VsourceAC "sin-wave voltage source"
	extends OnePort;
	parameter Voltage VA = 220 "Amplitude";
	parameter Real f (unit="Hz") = 50 "Frequency";
	constant Real PI = 3.14159265;
equation
	v = VA*sin(2*PI*f*time);
end VsourceAC;

model Ground "Ground"
	Pin p;
equation  
	p.v = 0;
end Ground;

model circuit
	Resistor R1(r=10);
	Resistor R2(r=10);
	Capacitor C(c=1);
	VsourceAC AC;
	Ground G;
equation
	connect(AC.p, R1.p);
	connect(R1.n, R2.p);
	connect(R2.n, C.p);
	connect(C.n, AC.n);
	connect(AC.n, G.p);
end circuit;


