#include "antlr/CommonAST.hpp"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

using namespace std;

class xml_dumper
{
private:
    int id;
    int fromOrd;
    int fIndent;
    static char c;
    static int indentSize;

    std::ofstream outFile;
    std::ostringstream nodes;
    std::ostringstream edges;
    string text;
public:
  
    xml_dumper(char* fileName){
	outFile.open(fileName);
	id = 0;
	fromOrd = 0;
	fIndent = 0;
    }

    void indentNodes(int size)
    {
	for (int i=0; i<size; i++)
	{
	    nodes << c;
	}
    }

    void indentEdges(int size)
    {
        for (int i=0; i<size; i++)
        {
            edges << c;
        }
    }
    
    // dump tag <symb id=" " type="SymbolTable">
    void dumpSTPrefix()
    {
	indentNodes(2*indentSize);
	nodes << "<symb id=\"" << id << "\" type=\"SymbolTable\">" << std::endl;
    }

    // dump tag </symb>
    void dumpSTSuffix()
    {
	indentNodes(2*indentSize);
	nodes << "</symb>" << std::endl;
    }

    // dump tag <map key="type">
    void dumpMapTypePrefix()
    {
	indentNodes(3*indentSize);
	nodes << "<map key=\"type\">" << std::endl;
    }

    // dump tag <map key="text">
    void dumpMapTextPrefix()
    {
	indentNodes(3*indentSize);
        nodes << "<map key=\"text\">" << std::endl;
    }
    // dump tag </map>
    void dumpMapSuffix()
    {
	indentNodes(3*indentSize);
	nodes << "</map>" << std::endl;
    }

    // dump mapping of Type, eg. <int id="" type="Int" value="11"/>
    void dumpMapType(antlr::RefAST ast)
    {
	dumpMapTypePrefix();
	indentNodes(4*indentSize);
	nodes << "<int id=\"\" type=\"Int\" value=\"" << ast->getType() \
            << "\"/>" << std::endl;
	dumpMapSuffix();
    }

    //dump mapping of Text, 
    //eg. <string id="" type="String" value="connector"/>
    void dumpMapText(antlr::RefAST ast)
    {
        dumpMapTextPrefix();
	indentNodes(4*indentSize);
	text=ast->getText();
	if (text=="<")
	    //cout << "find <" <<std::endl;
	    text="LT";
	else 
	    if (text=="<=")
		text="LTEQ";
	    else
		if (text==">")
		    text="GT";
		else 
		    if (text==">=")
		    	text="GTEQ";
		    else 
			if (text=="<>")
			    text="GTLT"; 
        nodes << "<string id=\"\" type=\"String\" value=\"" \
            << text << "\"/>" << std::endl;
        dumpMapSuffix();
    }

    //dump the following tags:
    //<?xml version="1.0"?>
    //<!DOCTYPE agl SYSTEM "http://agl.dtd">
    //<!-- GraphElements -->
    void dumpXMLHeader()
    {
	nodes << "<?xml version=\"1.0\"?>" << std::endl;
	nodes << "<!DOCTYPE agl SYSTEM \"http://agl.dtd\">" << std::endl;
	nodes << "<!-- GraphElements -->" << std::endl;
    }

    // dump tag <agl>
    void dumpXMLBeginTag()
    {
	nodes << "<agl>" << std::endl;
    }

    // dump </agl>
    void dumpXMLEndTag()
    {
        edges << "</agl>" << std::endl;
    }

    // dump tag <graph id="filename" type="AST">
    void dumpGraphBeginTag(char* filename) 
    {
	indentNodes(indentSize);
	/*
	nodes << "<graph id=\"" << filename <<"\" type=\"" << filename \
              << "\">" << std::endl;
	*/
	nodes << "<graph id=\"AST\" type=\"AST\">" <<std::endl;
    }

    // dump tag </graph>
    void dumpGraphEndTag()
    {
	indentEdges(indentSize);
	edges << "</graph>" << std::endl;
    }

    void incrementID()
    {
	id++;
    }

    // dump an edge, eg. <edge from="286" to="293" fromOrd="1" toOrd="0"/>
    void dumpEdge(int fromID, int toID, int fromOrd, int toOrd)
    {
	indentEdges(2*indentSize);
	edges << "<edge from=\"" << fromID << "\" to=\"" << toID << \
              "\" fromOrd=\"" << fromOrd << "\" toOrd=\"" << toOrd \
              << "\"/>" << std::endl;
    }

    // recursively dump all nodes and edges 
    void dump(antlr::RefAST ast)
    {
	int localID = id;
	if (ast == 0)
	{
	    nodes << "<NULL>";
	}
	else
	{
	    dumpSTPrefix();
	    dumpMapText(ast);
	    dumpMapType(ast);
	    dumpSTSuffix();
	    if (ast->getFirstChild() != 0)
	    {
		incrementID();
		dumpEdge(localID, id, 0, 0);
		dump(ast->getFirstChild());
	    }
	    if (ast->getNextSibling() != 0)
	    {
		incrementID();
		dumpEdge(localID, id, 1, 0);
		dump(ast->getNextSibling());
	    }
	}
    }

    // write to file
    void go(antlr::RefAST ast, char* filename){
	dumpXMLHeader();
	dumpXMLBeginTag();
	dumpGraphBeginTag(filename);
	dump(ast);
	dumpGraphEndTag();
	dumpXMLEndTag();
	std::string text1 = nodes.str();
	outFile << text1;
	std::string text2 = edges.str();
        outFile << text2;
	outFile.close();
    }

};

