# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the AST module

from ASTScope import DefinitionEntry, SymbolTable
from ASTComponent import TypePrefix, ClassModification, Declaration
from ASTEquations import ConnectEqStm
from ASTExp import IdentExp, NameExp, ArrSubExp, CommentExp, AnnotationExp
import string

indent='  '

class AST:
    "the AST of a progarm"
    def __init__(self, classFilesP=[], predefinedP=None):
        self.classFiles=classFilesP
        # the top scope
        self.topScope=SymbolTable()
        # the built-in scope
        self.builtinScope=SymbolTable()
        self.addBuiltinTypes()
        # the classFile of pre-defined names
        # pre-defined names are deifined in the package "Predfined.mo"
        self.predefined=predefinedP # ClassFile

    # key: the key in symbol table (String)
    # This method also adds a definition entry into the symbol table
    def addClassFile(self, classFile):
        #lineNum=-1
        #colNum=-1
	#definition=classFile.getClassDefinitions()[0]
        #name=definition.getName() # IdentExp
        #visibility=definition.getVisibility()
	#entry=DefinitionEntry(lineNum, colNum, name, definition, visibility)
        #self.topScope.add(key, entry) # key may contain path e.g., A.B.C
        self.classFiles.append(classFile)

    def getScope(self):
        return self.topScope

    def setPredefined(self, predefined):
	self.predefined=predefined

    def getPredefinedScope(self):
	return self.predefined.getClassDefinitions()[0].getScope()

    def getBuiltinScope(self):
	return self.builtinScope

    def addBuiltinTypes(self):
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'RealType'), \
                        None, None)
        self.builtinScope.add('RealType', entry)
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'IntegerType'), \
                        None, None)
        self.builtinScope.add('IntegerType', entry)
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'BooleanType'), \
                        None, None)
        self.builtinScope.add('BooleanType', entry)
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'StringType'), \
                        None, None)
        self.builtinScope.add('StringType', entry)
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'EnumType'), \
                        None, None)
        self.builtinScope.add('EnumType', entry)
        entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'StateSelect'), \
                        None, None)
        self.builtinScope.add('StateSelect', entry)

    def accept(self, visitor):
        visitor.visitAST(self)

class ClassFile:
    "represents a Modelica source file"
    def __init__(self, fileNameP="", dir="", withinClauseP=None, \
		 classDefinitionsP=[]):
	self.fileName=fileNameP
	# String: the directory where the file is located
	self.dir=dir 
	self.withinClause=withinClauseP
	# list[] of classDefinition
	self.classDefinitions=classDefinitionsP
        self.scope=SymbolTable()
	#self.builtinScope=SymbolTable()
	# create built-in scope
        #entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'RealType'), \
        #                None, None)#, None)#, None)
        #self.builtinScope.add('RealType', entry)
        #entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'IntegerType'), \
        #                None, None)#, None)#, None)
        #self.builtinScope.add('IntegerType', entry)
        #entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'BooleanType'), \
        #                None, None)#, None)#, None)
        #self.builtinScope.add('BooleanType', entry)
        #entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'StringType'), \
        #                None, None)#, None)#, None)
        #self.builtinScope.add('StringType', entry)
        #entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'EnumType'), \
        #                None, None)#, None)#, None)
        #self.builtinScope.add('EnumType', entry)
	#entry=DefinitionEntry(-1, -1, IdentExp(-1, -1, 'StateSelect'), \
	#		None, None)#, None)#, None)
	#self.builtinScope.add('StateSelect', entry)
	# predefined scope (this need to be changed)
	#self.predefinedScope=SymbolTable()

    def mergePredefinedScope(self, classFile):
	self.predefinedScope=classFile.getScope()

    def getScope(self):
	return self.scope

    def getBuiltinScope(self):
	return self.builtinScope

    def getPredefinedScope(self):
	return self.predefinedScope

    def getClassDefinitions(self):
	return self.classDefinitions

    def __str__(self):
	result='// '+self.fileName+'\n'
	if not self.withinClause==None:
	    result=result+str(self.withinClause)+';\n\n'
	for i in self.classDefinitions:
	    result=result+str(i)+';\n\n'
	return result

    def dump(self):
	result=''
	if not self.withinClause==None:
	    result=result+str(self.withinClause)+';\n\n'
	for i in self.classDefinitions:
	    result=result+i.dump(0)+';\n\n'
	return result

    def dumpST(self, scope, indentLevel):
	# not yet able to dump STs of for-loops
	indent='  '
	for i in scope.table.keys():
	    if not isinstance(scope.table[i], DefinitionEntry):
		print indent*indentLevel+scope.table[i].dump()
	    #elif not (i=='Real' or i=='Integer' or i=='Boolean' or \
		#	i=='String' or i=='StateSelect'):
	    elif not i=='self':
		print indent*indentLevel+\
		    str(scope.table[i].getDefinition().getClassPrefix())\
		    +str(scope.table[i].getDefinition().getKind())+' '+\
		    str(scope.table[i].getDefinition().getName())
		nextScope=scope.table[i].definition.getScope()
		if not nextScope==None:
	            self.dumpST(nextScope, indentLevel+1)
		endStr=indent*indentLevel+'end '+\
		    str(scope.table[i].getDefinition().getName())+';'
		if indentLevel==0:
		    endStr=endStr+'\n'
		print endStr

    def accept(self, visitor):
	visitor.visitClassFile(self)


class WithinClause:
    "within [name];"
    def __init__(self, lineNumP, colNumP, nameP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.name=nameP # NameExp

    def accept(self, visitor):
        visitor.visitWithinClause(self)

    def __str__(self):
	return 'within ' + str(self.name)

# -------------------- element ----------------------
# abstract class Element
class Element:

    lineNum=-1
    colNum=-1

    def __init__(self, lineNumP, colNumP):
	raise "Abstract class Element can not be instantiated!"

    def getLineNum(self):
	return self.lineNum

    def getColNum(self):
	return self.colNum

# ------------------ class_definition---------------------
class ClassPrefix:
    # visibility: Enum{'DEFAULT', 'PUBLIC', 'PROTECTED'}
    # isFinal: Boolean
    # isEncapsulated: Boolean
    # isPartial: Boolean
    # innerOuter: Enum{None, 'INNER', 'OUTER'}
    # isReplaceable: Boolean
    def __init__(self, visibilityP='DEFAULT', isFinalP=False, \
		isEncapsulatedP=False, isPartialP=False, \
		innerOuterP=None, isReplaceableP=False):
	self.visibility=visibilityP
	self.isFinal=isFinalP
	self.isEncapsulated=isEncapsulatedP
	self.isPartial=isPartialP
	self.innerOuter=innerOuterP
	self.isReplaceable=isReplaceableP

    def getVisibility(self):
	return self.visibility

    def setVisibility(self, value):
	self.visibility=value

    def setFinal(self, value):
        self.isFinal=value

    def setEncapsulated(self, value):
        self.isEncapsulated=value

    def setPartial(self, value):
        self.isPartial=value

    def setInnerOuter(self, value):
	self.innerOuter=value

    def setReplaceable(self, value):
	self.isReplaceable=value

    def __str__(self):
	result=''
	if self.isFinal==True:
	    result=result+'final '
	if self.innerOuter=='INNER':
	    result=result+'inner '
	elif self.innerOuter=='OUTER':
	    result=result+'outer '
	if self.isReplaceable==True:
	    result=result+'replaceable '
	    # 'replaceable' will not be dumped again in 
	    # ReplaceableClause
	if self.isEncapsulated==True:
	    result=result+'encapsulated '
	if self.isPartial==True:
	    result=result+'partial '
	# visibility is dumped in 
	# ClassDefinition
	return result
	

# abstract class ClassDefinition
class ClassDefinition(Element):
    "represents a class definition in Modelica"

    prefix=None # ASTDefinition::ClassPrefix
    #Enum{'class','model','type','connector','function','block','package'}
    kind='class'
    className=None # ASTExp::IdentExp
    scope=None # ASTScope::SymbolTable

    def __init__(self, lineNumP, colNumP, nameP):
	raise "Abstract class ClassDefinition can not be instantiated!"

    def getClassPrefix(self):
        return self.prefix

    def getKind(self):
        return self.kind

    def getName(self):
        return self.className

    def getScope(self):
	return self.scope

    def setOuterScope(self, st):
        self.scope.setOuter(st)

    def isEncapsulated(self):
        return self.prefix.isEncapsulated

    def isPartial(self):
        return self.prefix.isPartial


class RegularClassDefinition(ClassDefinition):
    "defined by keyword 'class'"
    def __init__(self, lineNumP, colNumP, prefixP, kindP, classNameP,\
		 stringCommentP, elementListP=[], \
                 initEquationPartP=[], equationPartP=[], \
                 initAlgorithmPartP=[], algorithmPartP=[]):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.prefix=prefixP # ClassPrefix
	self.kind=kindP
	self.className=classNameP
	self.stringComment=stringCommentP
	self.elementList=elementListP
	self.initEquationPart=initEquationPartP
	self.equationPart=equationPartP
	self.initAlgorithmPart=initAlgorithmPartP
	self.algorithmPart=algorithmPartP
	self.scope=SymbolTable()
	#self.flattenedElements=[]
	self.flatScope=SymbolTable()
	self.flattenedInitEqs=[]
	self.flattenedEqs=[]

    def getVisibility(self):
	return self.prefix.visibility

    def setVisibility(self, value):
	self.prefix.setVisibility(value)

    def setFinal(self, value):
	self.prefix.setFinal(value)

    def setEncapsulated(self, value):
	self.prefix.setEncapsulated(value)

    def setPartial(self, value):
	self.prefix.setPartial(value)

    def setInnerOuter(self, value):
	self.prefix.setInnerOuter(value)

    def setReplaceable(self, value):
	self.prefix.setReplaceable(value)

    def getElementList(self):
	return self.elementList
    
    #-------------------------------------------------------
    #Helper functions for DEVS code generation
    #-----------------------------------------------------------
    def hasIDDeclaration(self):
      for el in self.elementList:
	if(isinstance(el,ComponentClause)):
	  for comp in el.componentList:
    	    if(isinstance(comp, Declaration)):
              if(string.strip(str(comp.comment))=='\"ID\"'):
		#print dir(comp)
    	        return True, comp.getName()
      return False, 'None'
	
    def getImports(self):
      impList = []
      for el in self.elementList:
	if isinstance(el, ImportClause):
	  impList.append(el)
      return impList
	
    def getExtends(self):
      extends = ''
      for el in self.elementList:
	if isinstance(el, ExtendsClause):
	  extends = str(el.getName())
      return extends
	
    def isDevsModel(self):
      extends = self.getExtends()
      if(extends=='AtomicDEVS' or extends=='CoupledDEVS'):
	return True
      else:
	return False
    
    def getAnnotations(self): #added by Hongyan Song, Mar 29, 2006
	annList = []
	for el in self.elementList:
	    if isinstance(el, AnnotationExp):
	        annList.append(el)
	return annList
	
    def getScope(self): #added by Hongyan Song, April 10, 2006
	return self.scope

    def getParameters(self):  #added by Hongyan Song, Mar 29, 2006
	paramList = []
	for el in self.elementList:
	    if isinstance(el, ComponentClause) and el.isParameter():
                paramList.append(el.getComponentList())
	return paramList

    def getVars(self):  #added by Yentl Van Tendeloo
	paramList = []
	for el in self.elementList:
	    if isinstance(el, ComponentClause) and el.isNormal():
                paramList.append(el.getComponentList())
	return paramList


    def getInputs(self):   #added by Hongyan Song, Mar 29, 2006
	inList = []
	for el in self.elementList:
	    #if(isinstance(el, ComponentClause) and el.isInput()):
            if el.__class__.__name__ == "ComponentClause" and el.isInput():
	        inList.append(el.getComponentList())
	return inList
	
    #return a list of output names defined in a function or class
    def getOutputs(self):   #added by Hongyan Song, Mar 29, 2006
	outList = []
	for el in self.elementList:
	    if(isinstance(el, ComponentClause) and el.isOutput()):
	        outList.append(el.getComponentList())
	return outList

    #return a list of constant names defined in a function or class
    def getConstants(self):   #added by Hongyan Song, Mar 29, 2006
	constList = []
	for el in self.elementList:
	    if(isinstance(el, ComponentClause) and el.isConstant()):
	        constList.append(el.getComponentList())
	return constList

    #get sequential states from an atomic devs model
    #the state names are returned in a list in the order of the states defined
    #in the corresponding enumeration
    def getSeqStates(self):
	sList = []
	for el in self.elementList:
	    if(isinstance(el, RegularClassDefinition)):
		if(str(el.kind)=='type' and "SeqStates" in str(el.className)):
		    sList = el.getConstants()
	return sList
    #get formalism name from class/function annotations
    #for example, if annotation=(formalism="DEVS"), the function returns "DEVS"
    #if there is no element called formalism in the annotation, "" is returned
    def getFormalism(self):
	fStr = ""
	annList = self.getAnnotations()
	if(len(annList)>0):
	    ann = annList[0]
	    for el in ann.getAnnotation().getArgumentList():
		if string.lower(str(el.getCompRef())) == 'formalism':
		    fStr=fStr + el.getModification().getExp().getValue()
        return fStr
		
    #get source information from function annotation
    #for example, if annotation=(sorce="random"), the function returns "random"
    #if there is no element called source in the annotation, "" is returned
    def getSource(self):
	sStr = ""
	annList = self.getAnnotations()
	if(len(annList)>0):
	    ann = annList[0]
	    for el in ann.getAnnotation().getArgumentList():
		if string.lower(str(el.getCompRef())) == 'source':
		    sStr=sStr + el.getModification().getExp().getValue()
        return sStr
		
    #get base class name from class annotation
    #for example, if annotation=(base="Atomic"), the function returns "Atomic"
    #if there is not element called base defined in the given annotation, "" is returned
    def getBase(self):
	bStr = ""
	annList = self.getAnnotations()
	if(len(annList)>0):
	    ann = annList[0]
	    for el in ann.getAnnotation().getArgumentList():
		if string.lower(str(el.getCompRef())) == 'base':
		    bStr=bStr + el.getModification().getExp().getValue()
        return bStr

    #Generate the __str__ function for a class
    def toString(self, indentUnit, indentLevel):
      indentation = indentUnit * indentLevel
      strRep = '\n'
      if(self.kind=='class'):
	strRep = strRep + indentation + 'strRep = \'\''
	for el in self.elementList:
	  if(isinstance(el, ComponentClause) and len(str(el.getTypePrefix()))==0):
	    strRep = strRep + '\n'
	    jNum = 0
            for cp in el.getComponentList():
		if(jNum>0):
		  strRep = strRep + '\n'
		jNum = jNum + 1
	        strType = str(el.getType())
		#found = False
                if(strType=='Integer' or strType=='Real' or strType=='Boolean' or strType=='String'):		
	          strRep = strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"' + ' + str(self.'+str(cp.getName())+')'
		elif(strType=='list'):
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"'
		  strRep = strRep + '\n' + indentation + 'for item in self.' + str(cp.getName()) + ':'
		  strRep = strRep + '\n' + indentation + indentUnit + 'strRep = strRep + ' + '\"\\n\\t\"' + ' + str(item)'		  
		else:
		  strRep =strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"' + ' + str(self.'+str(cp.getName())+')'
        strRep = strRep + '\n'+ indentation + 'return strRep'
      else:
	strRep = self.kind + ' has no toString function ...' 

      return strRep

    def toXML(self, indentUnit, indentLevel):
      #indentLevel = indentLevel + 1
      indentation = indentUnit * indentLevel
      strRep ='\n'
      if(self.kind=='class'):
	strRep = strRep + indentation + 'strRep = \'\''
	for el in self.elementList:
	  if(isinstance(el, ComponentClause) and len(str(el.getTypePrefix()))==0):
	    strRep = strRep + '\n'
	    jNum = 0
            for cp in el.getComponentList():
		if(jNum>0):
		  strRep = strRep + '\n'
		jNum = jNum + 1
	        strType = str(el.getType())
		found = string.find(strType, 'SeqStates')
#		if( found>(-1)):
#	          strRep = strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"' + ' + str(self.'+str(cp.getName())+')'	
#		else:

	        if(strType=='Integer' or strType=='Real' or strType=='Boolean' or strType=='String' or found!=-1):
	          strRep = strRep + indentation + 'strRep = strRep + \"\\n<attribute category=\\\"P\\\">\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<name>'+str(cp.getName())+'</name>\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<type>'+str(el.getType())+'</type>\"\n'
		  strRep = strRep + indentation  + 'strRep = strRep + \"\\n\\t<value>\"+'+ 'str(self.'+str(cp.getName())+')'+'+\"</value>\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n</attribute>\"'
				
	          #strRep = strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"' + ' + str(self.'+str(cp.getName())+')'
		elif(strType=='list'):
	          strRep = strRep + indentation + 'strRep = strRep + \"\\n<attribute category=\\\"P\\\">\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<name>'+str(cp.getName())+'</name>\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<type>'+str(el.getType())+'</type>\"\n'
		  strRep = strRep + indentation  + 'strRep = strRep + \"\\n\\t<value>\"\n'
		  strRep = strRep + indentation + 'for item in self.' + str(cp.getName()) + ':\n'
		  strRep = strRep + indentation + indentUnit + 'strRep = strRep + ' + '\"\\n\\t\"' + ' + item.toXML()\n'
		  strRep = strRep + indentation  + 'strRep = strRep + \"\\n\\t</value>\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n</attribute>\"'
		  #strRep = strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"'
		  #strRep = strRep + '\n' + indentation + 'for item in self.' + str(cp.getName()) + ':'
		  #strRep = strRep + '\n' + indentation + indentUnit + 'strRep = strRep + ' + '\"\\n\\t\"' + ' + str(item)'		  
		else:
	          strRep = strRep + indentation + 'strRep = strRep + \"\\n<attribute category=\\\"C\\\">\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<name>'+str(cp.getName())+'</name>\"\n'
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n\\t<type>'+str(el.getType())+'</type>\"\n'
		  strRep = strRep + indentation + 'if(self.' + str(cp.getName()) + '!=None):\n'
		  strRep = strRep + indentation + indentUnit + 'strRep = strRep + \"\\n\\t<value>\"+'+ 'self.'+str(cp.getName())+'.toXML()'+'+\"</value>\"\n'
		  strRep = strRep + indentation + 'else:\n'
		  strRep = strRep + indentation + indentUnit + 'strRep = strRep + \"\\n\\t<value>None</value>\"\n'
		  
		  strRep = strRep + indentation + 'strRep = strRep + \"\\n</attribute>\"'
		  
		  #strRep =strRep + indentation + 'strRep = strRep + \"\\n' + str(cp.getName()) + ': \"' + ' + str(self.'+str(cp.getName())+')'
		

	      #print cp
	      #print cp.__class__

	    #print el
	    #print el.__class__
	    #print el.getType()
	    #for cp in el.getComponentList():
		#print cp
		#print cp.__class__
		#print cp.getName()
		#print dir(cp)
	    #print dir(el)
        strRep = strRep + '\n'+ indentation + 'return strRep'
      else:
	strRep = self.kind + ' has no toString function ...' 
      return strRep

    #-----------------------------------------------------------------
    def getInitEquations(self):
	return self.initEquationPart

    def getEquations(self):
	return self.equationPart

    def getConnectEqStms(self):
	result=[]
	for eq in self.equationPart:
	    #if isinstance(eq, ConnectEqStm):
            if eq.__class__.__name__ == "ConnectEqStm":
		result.append(eq)
	return result

    def getInitAlgorithm(self):
	return self.initAlgorithmPart

    def getAlgorithm(self):
	return self.algorithmPart

    def getFlatScope(self):
	return self.flatScope

    def getFlattenedComponents(self):
	return self.flatScope.getAllElems()

    def addFlattenedComponent(self, id, decl):
	self.flatScope.add(id, decl)

    def getFlattenedInitEqs(self):
	return self.flattenedInitEqs

    def getFlattenedEqs(self):
	return self.flattenedEqs

    def getExtendsClauses(self):
	result=[]
	for i in self.elementList:
	    if isinstance(i, ExtendsClause):
		result.append(i)
	return result

    def isExpanded(self):
	for i in self.getExtendsClauses():
	    if not i.isExpanded():
		return False
	return True

    def isPackage(self):
	if not (self.kind=='class' or self.kind=='package'):
	    return False 
	for i in self.elementList:
	    if not isinstance(i, ClassDefinition):
		if isinstance(i, ReplaceableClause):
		    raise '\""Replaceable\" construct is not yet supported!'
		if isinstance(i, ComponentClause):
		    vbty=i.getTypePrefix().getVariability()
		    if not vbty=='CONSTANT':
			return False
	if not len(self.initEquationPart)==0:
	    return False
	if not len(self.equationPart)==0:
            return False
	if not len(self.initAlgorithmPart)==0:
            return False
	if not len(self.algorithmPart)==0:
            return False
	return True

    # verify whether this is a valid connector
    def isConnector(self):
	# need add one more attribute flattenedInitEqs0, flattenedEqs0,
     	# which store the references to eqs after flattening of inheritance
	if self.kind=='connector':
	    return True
	return False

    def __str__(self):
	result=str(self.prefix)
	result=result+str(self.kind)+' '+str(self.className)
	if not self.stringComment==None:
	    result=result+' '+str(self.stringComment)
	result=result+'\n'
	cursor='DEFAULT'
	if len(self.elementList)>0:
	    for i in self.elementList:
		v=i.getVisibility()
		if not (v==cursor):
		    if not v=='DEFAULT':
			if v=='PUBLIC':
			    result=result+'public\n'
			else:
			    result=result+'protected\n'
		    cursor=v
		result=result+str(i)+';\n'
	if len(self.initEquationPart)>0:
	    result=result+'initial equation \n'
	    for i in self.initEquationPart:
                result=result+str(i)+'\n'
        if len(self.equationPart)>0:
            result=result+'equation \n'
            for i in self.equationPart:
                result=result+str(i)+'\n'
        if len(self.initAlgorithmPart)>0:
            result=result+'initial algorithm \n'
            for i in self.initAlgorithmPart:
                result=result+str(i)+'\n'
        if len(self.algorithmPart)>0:
            result=result+'algorithm \n'
            for i in self.algorithmPart:
                result=result+str(i)+'\n'
	result=result+'end ' +str(self.className)
	return result

    def dump(self, indentLevel):
	result=indent*indentLevel+str(self.prefix)
	result=result+str(self.kind)+' '+str(self.className)
        if not self.stringComment==None:
            result=result+' '+str(self.stringComment)
        result=result+'\n'
	if len(self.elementList)>0:
	    cursor='DEFAULT'
	    for i in self.elementList:
                v=i.getVisibility()
                if not (v==cursor):
                    if not v=='DEFAULT':
                        if v=='PUBLIC':
                            result=result+indent*indentLevel+'public\n'
                        else:
                            result=result+indent*indentLevel+'protected\n'
                    cursor=v
		result=result+i.dump(indentLevel+1)+';\n'
        if len(self.initEquationPart)>0:
            result=result+indent*indentLevel+'initial equation \n'
            for i in self.initEquationPart:
		result=result+i.dump(indentLevel+1)+';\n'
        if len(self.equationPart)>0:
            result=result+indent*indentLevel+'equation \n'
            for i in self.equationPart:
		result=result+i.dump(indentLevel+1)+';\n'
        if len(self.initAlgorithmPart)>0:
            result=result+indent*indentLevel+'initial algorithm \n'
            for i in self.initAlgorithmPart:
		result=result+i.dump(indentLevel+1)+';\n'
        if len(self.algorithmPart)>0:
            result=result+indent*indentLevel+'algorithm \n'
            for i in self.algorithmPart:
		result=result+i.dump(indentLevel+1)+';\n'
	result=result+indent*indentLevel+'end ' +str(self.className)
	return result

    def dumpFlattenedVersion(self):
	# dump flattened version declarations
        print 'class '+str(self.className)
        for i in self.getFlattenedComponents():
            #vis=''
            #if i.typePrefix.getVisibility()=='PROTECTED':
            #    vis='protected '
            #print '  '+vis+str(i.getTypePrefix())+str(i.getType())+\
            #      ' '+str(i)+';'
	    print '  '+i.dumpEnv()
	if len(self.flattenedInitEqs)>0:
	    print 'initial equation'
	for i in self.flattenedInitEqs:
	    print '  '+str(i)+';'
	if len(self.flattenedEqs):
	    print 'equations'
	for i in self.flattenedEqs:
	    print '  '+str(i)+';'
        print 'end '+str(self.className)+';'

    def accept(self, visitor):
        visitor.visitRegularClassDefinition(self)


class ShortClassDefinition(ClassDefinition):
    "short form of class definition"
    def __init__(self, lineNumP, colNumP, classPrefixP, kindP, \
		classNameP, basePrefixP, baseNameP, arrSubP=None, \
		classModificationP=None, commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.prefix=classPrefixP # ClassPrefix
	self.kind=kindP
	self.className=classNameP
        self.basePrefix=basePrefixP # typePrefix
        self.baseName=baseNameP #NameExp
	self.arrSub=arrSubP
        self.classModification=classModificationP
        self.comment=commentP
	self.initEquationPart=[]
        self.equationPart=[]
	self.initAlgorithmPart=[]
        self.algorithmPart=[]
        self.scope=SymbolTable()
        #self.flattenedElements=[]
        self.flatScope=SymbolTable()
        self.flattenedInitEqs=[]
        self.flattenedEqs=[]
	# reference to the base-class definition
	self.baseClass=None 

    def getBaseName(self):
	return self.baseName

    def getVisibility(self):
        return self.prefix.getVisibility()

    def getModification(self):
	return self.classModification

    def setVisibility(self, elem):
        self.prefix.setVisibility(value)

    def setFinal(self, value):
        self.prefix.setFinal(value)

    def setEncapsulated(self, value):
        self.prefix.setEncapsulated(value)

    def setPartial(self, value):
        self.prefix.setPartial(value)

    def setInnerOuter(self, value):
        self.prefix.setInnerOuter(value)

    def setReplaceable(self, value):
        self.prefix.setReplaceable(value)

    def getBaseClass(self):
	return self.baseClass

    def setBaseClass(self, classDefinition):
	self.baseClass=classDefinition

    def getInitEquations(self):
        return self.initEquationPart

    def getEquations(self):
        return self.equationPart

    def getConnectEqStms(self):
        result=[]
        for eq in self.equationPart:
            #if isinstance(eq, ConnectEqStm):
            if eq.__class__.__name__ == "ConnectEqStm":
                result.append(eq)
        return result

    def getInitAlgorithm(self):
        return self.initAlgorithmPart

    def getAlgorithm(self):
        return self.algorithmPart

    def getFlatScope(self):
        return self.flatScope

    def getFlattenedComponents(self):
        return self.flatScope.getAllElems()

    def addFlattenedComponent(self, id, decl):
        self.flatScope.add(id, decl)

    def getFlattenedInitEqs(self):
        return self.flattenedInitEqs

    def getFlattenedEqs(self):
        return self.flattenedEqs

    def isPackage(self):
	if not (self.kind=='class' or self.kind=='package'):
            return False
	return self.baseClass.isPackage()

    # verify whether this is a valid connector
    def isConnector(self):
        # need add one more attribute flattenedInitEqs0, flattenedEqs0,
        # which store the references to eqs after flattening of inheritance
        if self.kind=='connector':
            return True
        return False

    # Always return True since ShortClassDef has no extends clauses
    def isExpanded(self):
        return True

    def __str__(self):
        result=str(self.prefix)
	result= result+str(self.kind)+' '+str(self.className)+' = '
        result=result+str(self.basePrefix) + str(self.baseName)
        if not self.arrSub==None:
            result=result+str(self.arrSub)
        if not self.classModification==None:
            result=result+str(self.classModification)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result

    def dump(self, indentLevel):
	return indent*indentLevel+str(self)

    def accept(self, visitor):
        visitor.visitShortClassDefinition(self)


class EnumClassDefinition(ClassDefinition):
    "represents enumeration class definition"
    def __init__(self, lineNumP, colNumP, prefixP, kindP, classNameP, \
		enumListP, commentP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
	self.prefix=prefixP # ClassPrefix
        self.kind=kindP
        self.className=classNameP
	self.enumList=enumListP
	self.comment=commentP
	self.scope=SymbolTable()

    def getVisibility(self):
        return self.prefix.visibility

    def setVisibility(self, value):
        self.prefix.setVisibility(value)

    def setFinal(self, value):
        self.prefix.setFinal(value)

    def setEncapsulated(self, value):
        self.prefix.setEncapsulated(value)

    def setPartial(self, value):
        self.prefix.setPartial(value)

    def setInnerOuter(self, value):
        self.prefix.setInnerOuter(value)

    def setReplaceable(self, value):
        self.prefix.setReplaceable(value)

    def __str__(self):
        result=str(self.prefix)
        result= result+str(self.kind)+' '+str(self.className)+\
		' = enumeration('
	temp=''
	for i in self.enumList:
	    temp=temp+str(i)+', '
	result= result+temp[0:len(temp)-2]+')'
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result

    def dump(self, indentLevel):
	return indent*indentLevel+str(self)

    def accept(self, visitor):
        visitor.visitEnumClassDefinition(self)


# ------------ import, extends and component clauses-----------------
class ImportClause(Element):
    "represents import_clause"
    def __init__(self, lineNumP, colNumP, nameP, identP=None, \
                 qualifiedP=False, commentP=None, visibilityP='DEFAULT'):
        self.lineNum=lineNumP
        self.colNum=colNumP
	self.visibility=visibilityP
        self.name=nameP
        self.ident=identP # IdentExp
        self.qualified=qualifiedP
        self.comment=commentP

    def getVisibility(self):
	return self.visibility

    def setVisibility(self, value):
	self.visibility=value

    def setOuterScope(self, st):
	pass

    def __str__(self):
        result='import '
        if not self.ident==None:
            result=result+str(self.ident)+'='
        result=result+str(self.name)
        if self.qualified==False:
            result=result+'.*'
        if not self.comment==None:
            result=result+str(self.comment)
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel+'import '
	if not self.ident==None:
            result=result+str(self.ident)+'='
        result=result+str(self.name)
        if self.qualified==False:
            result=result+'.*'
        if not self.comment==None:
            result=result+str(self.comment)
        return result

    def accept(self, visitor):
        visitor.visitImportClause(self)

	
class ExtendsClause(Element):
    "represents extends_clause"
    def __init__(self, lineNumP, colNumP, nameP, modificationP=None, \
		visibilityP='DEFAULT'):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.name=nameP
        self.modification=modificationP # ASTComponent::ClassModification
	self.visibility=visibilityP
	# flag: indiactes whether this inheritance  has been expanded
	self.expanded=False

    def isExpanded(self):
	return self.expanded

    def setExpanded(self, value):
	self.expanded=value

    def getName(self):
	return self.name

    def getModification(self):
	return self.modification

    def getVisibility(self):
        return self.visibility

    def setVisibility(self, value):
        self.visibility=value

    def setOuterScope(self, st):
        pass

    def __str__(self):
        result='extends '+str(self.name)
        if not self.modification==None:
            result=result+str(self.modification)
        return result

    def dump(self, indentLevel):
        result=indent*indentLevel+'extends '+str(self.name)
	if not self.modification==None:
            result=result+str(self.modification)
	return result

    def accept(self, visitor):
        visitor.visitExtendsClause(self)


class ComponentClause(Element):
    "declaration"
    def __init__(self, lineNumP, colNumP, typePrefixP, typeP, \
		componentListP, arrSubP=None, isFinalP=False, \
		innerOuterP=None, isReplaceableP=False):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.typePrefix=typePrefixP
        self.type=typeP # NameExp
        self.arrSub=arrSubP
        self.componentList=componentListP # [] of ASTComponent::Declaration
        self.isFinal=isFinalP # Boolean
        self.innerOuter=innerOuterP # Enum{None, 'INNER', 'OUTER'}
	self.isReplaceable=isReplaceableP

    def getTypePrefix(self):
	return self.typePrefix

    def getVisibility(self):
	return self.typePrefix.visibility

    def setVisibility(self, value):
	self.typePrefix.setVisibility(value)

    def getType(self):
	return self.type

    def getArrSub(self):
	return self.arrSub

    def getComponentList(self):
	return self.componentList

    def getFinal(self):
	return self.isFinal

    def setFinal(self, value):
	self.isFinal=value

    def getInnerOuter(self):
	return self.innerOuter

    def setInnerOuter(self, value):
	self.innerOuter=value

    def getReplaceable(self):
	return self.isReplaceable

    def setReplaceable(self, value):
	self.isReplaceable=value

    def setOuterScope(self, st):
        pass
    
    #Helper functions for DEVS code generation
    #Added by Hongyan Song, Mar 29, 2006

    def isParameter(self):
	if(str(self.getTypePrefix())=='parameter '):
	    return True
	else:
	    return False

    def isInput(self):
	if(str(self.getTypePrefix())=='input '):
	    return True
	else:
	    return False
    
    def isOutput(self):
	if(str(self.getTypePrefix())=='output '):
	    return True
	else:
	    return False
	
    def isConstant(self):
	if(str(self.getTypePrefix())=='constant '):
	    return True
	else:
	    return False

    def isNormal(self):
	if(str(self.getTypePrefix())==''):
	    return True
	else:
	    return False

    #---------------------------------------------------
    def __str__(self):
	result=''
	if self.isFinal==True:
	    result=result+'final '
	if self.innerOuter=='INNER':
	    result=result+'inner '
	elif self.innerOuter=='OUTER':
	    result=result+'outer '
	if self.isReplaceable==True:
	    result=result+'replaceable '
        result=result+str(self.typePrefix) + str(self.type)
        if not self.arrSub==None:
            result=result+str(self.arrSub)
	result=result+' '
	temp=''
        for i in self.componentList:
            temp=temp+str(i)+', '
	result=result+temp[0:len(temp)-2]
        return result

    def dump(self, indentLevel):
	return indent*indentLevel+str(self)

    def accept(self, visitor):
        visitor.visitComponentClause(self)

class ReplaceableClause(Element):
    "replaceable element"
    def __init__(self, lineNumP, colNumP, classDefinitionP, \
		componentClauseP, constrainingClauseP=None, \
		commentP=None, isFinalP=False, innerOuterP=None):
	# invariant: classDefinition and componentClause are 
	# mutually exclusive
	# Exactly one of them None and the other refer to a valid object
        self.lineNum=lineNumP
        self.colNum=colNumP
	self.classDefinition=classDefinitionP
	# invariant: classDefinition must be 'replaceable'
	# ie., classDefinition.prefix.isRepaceable=True
	self.componentClause=componentClauseP
	self.constrainingClause=constrainingClauseP
	self.comment=commentP
        self.isFinal=isFinalP # Boolean
        self.innerOuter=innerOuterP # Enum{None, 'INNER', 'OUTER'}

    def getVisibility(self):
	if not self.componentClause==None:
	    return self.componentClause.getVisibility()
	elif not self.classDefinition==None:
	    return self.classDefinition.getVisibility()

    def setVisibility(self, value):
	if not self.componentClause==None:
	    self.componentClause.setVisibility(value)
	elif not self.classDefinition==None:
	    self.classDefinition.setVisibility(value)

    def setFinal(self, value):
	self.isFinal=value

    def setInnerOuter(self, value):
	self.innerOuter=value

    def setOuterScope(self, st):
        pass

    def __str__(self):
	result=''
	if not self.classDefinition==None:
	    result=result+str(self.classDefinition)
	elif not self.componentClause==None:
	    result=result+str(self.componentClause)
	if not self.constrainingClause==None:
	    result=result+' '+str(self.constrainingClause)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
	return result	

    def dump(self, indentLevel):
	result=''
	if not self.classDefinition==None:
	    result=result+self.classDefinition.dump(indentLevel)
	elif not self.componentClause==None:
	    result=result+self.componentClause.dump(indentLevel)
	if not self.constrainingClause==None:
	    result=result+' '+str(self.constrainingClause)
	if not self.comment==None:
	    result=result+' '+str(self.comment)
        return result

    def accept(self, visitor):
        visitor.visitReplaceableClause(self)

# --------------------Enumeration & ClassModifier----------------------
class EnumLiteral:
    def __init__(self, lineNumP, colNumP, identP, commentP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.ident=identP
	self.comment=commentP

    def __str__(self):
	result=str(self.ident) 
	if not self.comment==None:
	    result=result+ ' ' + str(self.comment)
	return result

