# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the AST module

from ASTScope import SymbolTable

indent='  '
#-------------------Expression--------------------------
#abstract class Exp
class Exp:
    lineNum=-1
    colNum=-1

    def __init__(self, lineNumP, colNumP):
        raise "Abstract class Exp can not be instantiated!"

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

#-------------------Atomic Expression--------------------------
#abstract class AtomicExp
class AtomicExp(Exp):
    def __init__(self):
        raise "Abstract class AtomicExp can not be instantiated!"


class IdentExp(AtomicExp):
    def __init__(self, lineNumP, colNumP, valueP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.value=valueP #String

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getValue(self):
	return self.value

    def setValue(self, v):
	self.value=v

    def getOperands(self):
	return [self]

    def __str__(self):
	return str(self.value)

    def accept(self, v):
	v.visitIdentExp(self)

    # the 'value' field in IdentExp is set refered to DeclarationEntry
    def getType(self):
	#print "visit IdentExp getType:", str(self)
	result=str(self.value.getType())
	if result=='Real':
	    return 'RealType'
	elif result=='Integer':
	    return 'IntegerType'
	elif result=='String':
	    return 'StringType'
	elif result=='Boolean':
	    return 'BooleanType'
	else:
	    return 'TypeError'

class IntegerExp(AtomicExp):
    def __init__(self, lineNumP, colNumP, valueP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.value=valueP # Int

    def getLineNum(self):
	return self.lineNum

    def getColNum(self):
	return self.colNum

    def setValue(self, v):
	self.value=v

    def getValue(self):
	return self.value

    def __str__(self):
        return str(self.value)

    def getOperands(self):
        return [self]

    def accept(self, v):
        v.visitIntegerExp(self)

    def getType(self):
	return 'IntegerType'

class RealExp(AtomicExp):
    def __init__(self, lineNumP, colNumP, valueP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.value=valueP # float

    def setValue(self, v):
        self.value=v

    def getValue(self):
        return self.value

    def getOperands(self):
        return [self]

    def __str__(self):
        return str(self.value)

    def accept(self, v):
        v.visitRealExp(self)

    def getType(self):
	return 'RealType'

class BoolExp(AtomicExp):
    def __init__(self, lineNumP, colNumP, valueP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.value=valueP # Boolean

    def getOperands(self):
        return [self]

    def __str__(self):
	if self.value==True:
            return str('true')
	elif self.value==False:
	    return str('false')

    def accept(self, v):
        v.visitBoolExp(self)

    def getType(self):
	return 'BooleanType'

class StringExp(AtomicExp):
    def __init__(self, lineNumP, colNumP, valueP):
	self.lineNum=lineNumP
	self.colNum=colNumP
	self.value=valueP # String

    def getOperands(self):
        return [self]

    def __str__(self):
        return '\"'+str(self.value)+'\"'

    def accept(self, v):
        v.visitStringExp(self)

    def getType(self):
	return 'StringType'

    def getValue(self):  #added by Hongyan Song, Mar 29, 2006
	return self.value

#-------------------Composite Expression--------------------------
#abstract class AtomicExp
class CompositeExp(Exp):
    def __init__(self):
        raise "Abstract class CompositeExp can not be instantiated!"

    
class NameExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of IdentExp

    def getOperands(self):
	return self.operands

    def __str__(self):
	result=''
	temp=''
	for i in self.operands:
	    temp=temp+str(i)+'.'
	result=result+temp[0:len(temp)-1]
        return result

    def accept(self, v):
        v.visitNameExp(self)

    # these are no NameExps in equation part
    def getType(self):
	raise 'unexpected NameExp in equation part ... abort!'

class OrExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
	return str(self.operands[0]) + ' or ' + str(self.operands[1])

    def accept(self, v):
        v.visitOrExp(self)

    def getType(self):
	if self.operands[0].getType()=='BooleanType' and \
	   self.operands[1].getType()=='BooleanType':
	    return 'BooleanType'
	return 'TypeError'

class AndExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + ' and ' + str(self.operands[1])

    def accept(self, v):
        v.visitAndExp(self)

    def getType(self):
        if self.operands[0].getType()=='BooleanType' and \
           self.operands[1].getType()=='BooleanType':
            return 'BooleanType'
        return 'TypeError'

class NotExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return ' not ' + str(self.operands[0])

    def accept(self, v):
        v.visitNotExp(self)

    def getType(self):
	if self.operands[0].getType()=='BooleanType':
	    return 'BooleanType'
	return 'TypeError'

class LessExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '<' + str(self.operands[1])

    def accept(self, v):
        v.visitLessExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
	elif self.operands[0].getType()=='BooleanType' and \
	     self.operands[1].getType()=='BooleanType':
	    return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
	    #newExp=ConversionExp([self.operands[1]])
	    #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
	    #newExp=ConversionExp([self.operands[0]])
	    #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class LessEQExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '<=' + str(self.operands[1])

    def accept(self, v):
        v.visitLessEQExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
        elif self.operands[0].getType()=='BooleanType' and \
             self.operands[1].getType()=='BooleanType':
            return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
	    #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
	    #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class GreaterExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '>' + str(self.operands[1])

    def accept(self, v):
        v.visitGreaterExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
        elif self.operands[0].getType()=='BooleanType' and \
             self.operands[1].getType()=='BooleanType':
            return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
	    #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class GreaterEQExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '>=' + str(self.operands[1])

    def accept(self, v):
        v.visitGreaterEQExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
        elif self.operands[0].getType()=='BooleanType' and \
             self.operands[1].getType()=='BooleanType':
            return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class EQEQExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '==' + str(self.operands[1])

    def accept(self, v):
        v.visitEQEQExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
        elif self.operands[0].getType()=='BooleanType' and \
             self.operands[1].getType()=='BooleanType':
            return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class LessGtExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '<>' + str(self.operands[1])

    def accept(self, v):
        v.visitLessGtExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'BooleanType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'BooleanType'
        elif self.operands[0].getType()=='StringType' and \
             self.operands[1].getType()=='StringType':
            return 'BooleanType'
        elif self.operands[0].getType()=='BooleanType' and \
             self.operands[1].getType()=='BooleanType':
            return 'BooleaType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'BooleanType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'BooleanType'
        else:
            return 'TypeError'

class SumExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getOperands(self):
        return self.operands

    def setOperands(self, newOperands):
	self.operands=newOperands

    def __str__(self):
        #result='+('
        #for i in self.operands:
        #    result=result+str(i)+','
        #result=result+')'
	result=''
	for i in self.operands:
	    result=result+str(i)+'+'
        return result[0:len(result)-1]

    def accept(self, v):
        v.visitSumExp(self)

    def getType(self):
	if self.operands[0].getType()=='IntegerType' and \
	   self.operands[1].getType()=='IntegerType':
	    return 'IntegerType'
	elif self.operands[0].getType()=='RealType' and \
	     self.operands[1].getType()=='RealType':
	    return 'RealType'
	elif self.operands[0].getType()=='StringType' and \
	     self.operands[1].getType()=='StringType':
	    return 'StringType'
	# type coercion
	elif self.operands[0].getType()=='RealType' and \
	     self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
	    return 'RealType'
	elif self.operands[0].getType()=='IntegerType' and \
	     self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
	    return 'RealType'
	else:
	    return 'TypeError'

class MulExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getOperands(self):
        return self.operands

    def setOperands(self, newOperands):
        self.operands=newOperands

    def __str__(self):
	# as n-nary operator
	#result='*('
	#for i in self.operands:
	#    result=result+str(i)+','
	#result=result+')'
	#return result
	result=''
        for i in self.operands:
            result=result+str(i)+'*'
        return result[0:len(result)-1]

    def accept(self, v):
        v.visitMulExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'IntegerType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'RealType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'RealType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'RealType'
        else:
            return 'TypeError'

class SubtractExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def setOperand(self, index, exp):
	self.operands[index]=exp

    def __str__(self):
        return str(self.operands[0]) + '-' + str(self.operands[1])

    def accept(self, v):
        v.visitSubtractExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'IntegerType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'RealType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'RealType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'RealType'
        else:
            return 'TypeError'

class DivExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '/' + str(self.operands[1])

    def accept(self, v):
        v.visitDivExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'RealType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'RealType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            #newExp=ConversionExp([self.operands[1]])
            #self.operands[1]=newExp
            return 'RealType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            #newExp=ConversionExp([self.operands[0]])
            #self.operands[0]=newExp
            return 'RealType'
        else:
            return 'TypeError'

class UnaryMinusExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
        self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)        

    def getOperands(self):
        return self.operands

    def __str__(self):
	# '(' should be removed
        return '(-' + str(self.operands[0])+')'

    def accept(self, v):
        v.visitUnaryMinusExp(self)

    def getType(self):
	if self.operands[0].getType()=='RealType':
	    return 'RealType'
	elif self.operands[0].getType()=='Integerype':
	    return 'IntegerType'
	else:
	    return 'TypeError'

class PowerExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getOperands(self):
        return self.operands

    def setOperands(self, newOperands):
        self.operands=newOperands

    def __str__(self):
	if len(self.operands)==1:
	    return str(self.operands[0])
	else:
            return str(self.operands[0]) + '^' + str(self.operands[1])

    def accept(self, v):
        v.visitPowerExp(self)

    def getType(self):
        if self.operands[0].getType()=='IntegerType' and \
           self.operands[1].getType()=='IntegerType':
            return 'IntegerType'
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='RealType':
            return 'RealType'
        # type coercion
        elif self.operands[0].getType()=='RealType' and \
             self.operands[1].getType()=='IntegerType':
            return 'RealType'
        elif self.operands[0].getType()=='IntegerType' and \
             self.operands[1].getType()=='RealType':
            return 'RealType'
        else:
            return 'TypeError'

# operands [] can be empty, eg. Real a[:];
class RangeExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands
        
    def __str__(self):
        if len(self.operands)==0:
            return ':'
	if len(self.operands)==2:
            return str(self.operands[0]) + ':' + str(self.operands[1])
	elif len(self.operands)==3:
            return str(self.operands[0]) + ':' + str(self.operands[1]) \
			+ ':' + str(self.operands[2])

    def accept(self, v):
        v.visitRangeExp(self)

    def getType(self):
	# TBD
	pass

class IfElseExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, ifTestP, ifBodyP, \
		elseifListP=[], elseBodyP=None):
	self.lineNum=lineNumP
	self.colNum=colNumP
        #self.operands=operandsP # list [] of operands(Exp)
	self.ifTest=ifTestP
	self.ifBody=ifBodyP # an Exp
	self.elseifList=elseifListP
	# list [] of ifElseExp
	self.elseBody=elseBodyP # an Exp

    def getOperands(self):
	pass

    def getIfTest(self):
	return self.ifTest

    def getIfBody(self):
	return self.ifBody

    def setIfBody(self, newIfBody):
	self.ifBody=newIfBody

    def getElseifList(self):
	return self.elseifList

    def getElseBody(self):
	return self.elseBody

    def setElseBody(self, newElseBody):
	self.elseBody=newElseBody

    def __str__(self):
	result='if '+str(self.ifTest)+' then '+str(self.ifBody)
	if len(self.elseifList)>0:
	    for i in self.elseifList:
		result=result+' else'+str(i)[0:len(str(i))-1]
	if not self.elseBody==None:
	    result=result+' else '+str(self.elseBody)
	#result=result+';'
	return result

    def accept(self, v):
        v.visitIfElseExp(self)

    def getType(self):
        if not self.ifTest.getType()=='BooleanType':
            return 'TypeError'
	# type might be determined at runtime

class FunCallExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP
        # list [] of operands(Exp), the first elem is an 
	# CompRefExp or IdentExp, the second is an FunArgsExp

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '(' + str(self.operands[1]) + ')'
 
    def accept(self, v):
        v.visitFunCallExp(self)       

    def getType(self):
	# TBD
	pass

class FunArgsExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
	result=''
	for i in self.operands:
	    result=result + str(i) + ','
	return result[0:len(result)-1]

    def accept(self, v):
        v.visitFunArgsExp(self)    

    def getType(self):
        # TBD
        pass

class ForFunArgExp(CompositeExp):
    def  __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP
        # list [] of operands: the 1st elem is Exp
        # the others are IdentInExpExp
	self.scope=SymbolTable()

    def getOperands(self):
        return self.operands

    def getScope(self):
        return self.scope

    def setOuterScope(self, st):
        self.scope.setOuter(st)

    def __str__(self):
	result=str(self.operands[0])+' for '
        for i in range(1, len(self.operands)):
            result=result + str(self.operands[i])+', '
	return result[0:len(result)-2]

    def accept(self, v):
        v.visitForFunArgExp(self)    

    def getType(self):
        # TBD
        pass

class NamedArgExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + '=' + str(self.operands[1])

    def accept(self, v):
        v.visitNamedArgExp(self)

    def getType(self):
        # TBD
        pass

class ConnectRefExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(SingleCompRefExp)

    def getOperands(self):
        return self.operands

    def __str__(self):
	result=''
        if len(self.operands)>0:
            result=str(self.operands[0])
        if len(self.operands)==2:
            result=result+'.'+str(self.operands[1])
        return result

    def accept(self, v):
        v.visitConnectRefExp(self)

    def getType(self):
        # TBD
        pass

class SingleCompRefExp(CompositeExp):
    "represent eg, a[i,j]"
    def  __init__(self, lineNumP, colNumP, operandsP):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.operands=operandsP
        # at most 2 elems are allowed
        # the first one is an IdentExp
        # the second one is an ArrSubExp (optional)

    def getOperands(self):
        return self.operands

    def __str__(self):
        if len(self.operands)==1:
            result=str(self.operands[0])
        elif len(self.operands)==2:
            result=str(self.operands[0])+str(self.operands[1])
        return result

    def accept(self, v):
        v.visitSingleCompRefExp(self)

    def getType(self):
	result=str(self.operands[0].getType())
        if result=='Real':
            return 'RealType'
        elif result=='Integer':
            return 'IntegerType'
        elif result=='String':
            return 'StringType'
        elif result=='Boolean':
            return 'BooleanType'
        else:
            return 'TypeError'

class CompRefExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of SingleCompRefExp

    def getLineNum(self):
        return self.lineNum

    def getColNum(self):
        return self.colNum

    def getOperands(self):
        return self.operands

    # return: SingleCompRefExp
    def getLastOperand(self):
	return self.operands[len(self.operands)-1]

    def __str__(self):
	result=''
	for i in self.operands:
	    result=result+str(i)+'.'
	return result[0:len(result)-1]

    def accept(self, v):
        v.visitCompRefExp(self)

    def getType(self):
	return self.getLastOperand().getType()

class ArrSubExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        result='['
	temp=''
        for i in self.operands:
            temp=temp + str(i) + ','
	result=result+temp[0:len(temp)-1]+']'
        return result

    def accept(self, v):
        v.visitArrSubExp(self)

    def getType(self):
	# TBD
	pass

class CommentExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)
	# the first element is a StringExp
	# the second element is an AnnotationExp

    def getOperands(self):
        return self.operands

    def __str__(self):
	result=''
	if len(self.operands)>0:
            result=str(self.operands[0])
	if len(self.operands)==2:
	    result=result+' '+str(self.operands[1])
	return result

    def accept(self, v):
        v.visitCommentExp(self)

    def getType(self):
        # TBD
        pass

class AnnotationExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, annotationP, \
		visibilityP='DEFAULT'):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.annotation=annotationP # instance of ClassModification
	self.visibility=visibilityP

    def getOperands(self):
        pass
        
    def getAnnotation(self):  #added by Hongyan Song, Mar 29, 2006
	return self.annotation
    
    def getVisibility(self):
	return self.visibility

    def setVisibility(self, value):
	self.visibility=value

    def __str__(self):
        return 'annotation '+str(self.annotation)

    def dump(self, indentLevel):
	result=indent*indentLevel+str(self)
	return result

    def accept(self, v):
        v.visitAnnotationExp(self)

    def getType(self):
        # TBD
        pass

class IdentInExpExp(CompositeExp):
    def  __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP
        # list [] of operands(Exp), the first elem is an IdentExp,
        # the second is an Exp
        
    def getOperands(self):
        return self.operands

    def __str__(self):
        return str(self.operands[0]) + ' in ' + str(self.operands[1])

    def accept(self, v):
        v.visitIdentInExpExp(self)

    def getType(self):
        # TBD
        pass

class ForIndicesExp(CompositeExp):
    def  __init__(self, lineNumP, colNumP, operandsP):
	self.lineNum=lineNumP
	self.colNum=colNumP
        self.operands=operandsP
        # list [] of operands(IdentInExpExp)

    def getOperands(self):
	return self.operands

    def __str__(self):
	result='for '
	temp=''
        for i in self.operands:
            temp=temp + str(i) + ', '
	result=result+temp[0:len(temp)-2]
	result=result + ' loop'
        return result

    def accept(self, v):
        v.visitForIndicesExp(self)

    def getType(self):
        # TBD
        pass

class ExpListExp(CompositeExp):
    def  __init__(self, lineNumP, colNumP, operandsP, kindP=None):
        self.lineNum=lineNumP
        self.colNum=colNumP
	self.kind=kindP 
	# Enum(None, "PAR", "BRACK", "BRACE"), 
	# PAR '(', BRACK '[', BRACE '{', None means no brackets
	# '[' contains a list of ExpListExp
        self.operands=operandsP

    def getKind(self):
	return self.kind

    def getOperands(self):
        return self.operands

    def __str__(self):
	if self.kind=='PAR':
	    begin='('
	    end=')'
	elif self.kind=='BRACK':
	    begin='['
	    end=']'
	elif self.kind=='BRACE':
	    begin='{'
	    end='}'
        else:
                # add a default
                begin = '('
                end = ')'
	result=begin
	if self.kind=='BRACK':
	    for i in self.operands:
	    	result=result + str(i) + '; '
	    result=result[0:len(result)-2]
        else:
	    for i in self.operands:
            	result=result + str(i) + ','
	    result=result[0:len(result)-1]
	return result+end

    def accept(self, v):
        v.visitExpListExp(self)

    def getType(self):
	if self.kind=='PAR' and len(self.operands)==1:
	    return self.operands[0].getType()

# conversion from IntegerType to RealType. Unary exp.
class ConversionExp(CompositeExp):
    def __init__(self, lineNumP, colNumP, operandsP):
        self.lineNum=lineNumP
        self.colNum=colNumP
        self.operands=operandsP # list [] of operands(Exp)

    def getOperands(self):
        return self.operands

    def __str__(self):
        return 'Real(' + str(self.operands[0])+')'

    def accept(self, v):
        v.visitConversionExp(self)

    def getType(self):
        return 'RealType'
