import AST
from AST.ASTComponent import *
from AST.ASTExp import *
from AST.ASTDefinition import *
from AST.ASTEquations import *
from AST.ASTScope import *
from copy import deepcopy

class TempAtomic:
        def __init__(self, name, flattened):
                self.name = name
                self.flattened = flattened
                self.atomics = []
                self.variables = []
                self.functions = []
                self.inputs = []
                self.outputs = []
                self.connections = []
                self.intTransition = None
                self.extTransition = None
                self.outputFnc = None
                self.timeAdvance = None
                self.select = None
                self.mappingFunction = None

        def getType(self, name):
                for check in self.atomics:
                        if str(check[0]) == name:
                                return str(check[1])

        def getInfluencees(self, name):
                influencees = []
                for connection in self.connections:
                        if name == connection[0]:
                                influencees.append(connection[2])
                return influencees

        def generateMappings(self):
                for i in self.connections:
                        rule = deepcopy(self.mappingFunction.getAlgorithm()[1])
                        inst, port, inst2, port2 = i
                        if str(inst) != "":
                                inst += "_"
                        if str(inst2) != "":
                                inst2 += "_"
                        rule.ifTest.operands[1].operands[0].value = str(inst) + str(port)
                        rule.ifBody[0].compRef.operands[0].operands[1].operands[0].value = str(inst2) + str(port2)
                        rule.ifBody[0].exp.operands[0].operands[1].operands[0].value = str(inst) + str(port)
                        self.mappingFunction.algorithmPart.append(rule)
                self.mappingFunction.algorithmPart = [self.mappingFunction.algorithmPart[0]] + self.mappingFunction.algorithmPart[2:]

        def generateVariables(self):
                foundReal = False
                #elemlist = self.state.elementList
                elemlist = self.functions
                for i in self.functions:
                        if (i.__class__.__name__ == "ComponentClause") and ("NewAtomic_State" in str(i.type)):
                                i.type = self.name + "State"
                #self.state.className = self.name + "State"
                for var in elemlist:
                        if var.__class__.__name__ == "ComponentClause":
                                if str(var.type) == "list" and str(var.componentList[0].ident) == "atomics":
                                        prule = var.componentList[0].modification.exp.operands[1].operands
                                        loc = var.componentList[0].modification.exp.operands[1]
                                        loc.operands = []
                                        newval = ""
                                        first = True
                                        for atomic in self.atomics:
                                                atomic_name = atomic[0]
                                                timeLeft = atomic_name + "_timeLeft"
                                                state = atomic_name + "_state"
                                                if not first:
                                                        newval += ","
                                                else:
                                                        first = False
                                                newval += "list([" + timeLeft + "," + state + "])"
                                        newval = "[" + newval + "]"
                                        loc.operands.append(IdentExp(-1, -1, newval))
                                        # Shift this rule to the back
                                        elemlist.remove(var)
                                        elemlist.append(var)
                                        #self.state.elementList.append(var)
                                        break
                                if str(var.type) == "Real" and not foundReal:
                                        for atomic in self.atomics:
                                                atomic_name = atomic[0]
                                                atomic_type = str(atomic[1].className)

                                                # Init the state first, we will need it later on
                                                rule = deepcopy(var)
                                                rule.componentList[0].ident.value = atomic_name + "_state"
                                                rule.type.operands[0].value = atomic_type + "State"
                                                rule.componentList[0].modification.exp.value = atomic_type  + "State()"
                                                self.functions.append(rule)


                                                # The time left
                                                rule = deepcopy(var)
                                                rule.componentList[0].ident.value = atomic_name + "_timeLeft"
                                                rule.type.operands[0].value = "Real"
                                                rule.componentList[0].modification.exp.value = "timeAdvance_" + atomic_type + "(" + atomic_name + "_state)"
                                                self.functions.append(rule)

                                                # The elapsed time
                                                rule = deepcopy(var)
                                                rule.componentList[0].ident.value = atomic_name + "_elapsed"
                                                rule.type.operands[0].value = "Real"
                                                rule.componentList[0].modification.exp.value = "0"

                                                self.functions.append(rule)

                                                foundReal = True

        def generateGetModelKeys(self):
                for i in self.getModelKeys.algorithmPart:
                        if "while" in str(i):
                                new_body = []
                                for j in i.body:
                                        if "if" in str(j):
                                                for atomic in self.atomics:
                                                        atomic_name = atomic[0]
                                                        rule = deepcopy(j)
                                                        rule.ifBody[0].funCall.operands[0].value = StringExp(-1, -1, atomic_name)
                                                        rule.ifTest.operands[1].operands[0].value = StringExp(-1, -1, atomic_name)
                                                        new_body.append(rule)
                                        else:
                                                new_body.append(j)
                                i.body = new_body

        def generateGetDStar(self):
                new_algo = []
                stmtcounter = 0
                for i in self.get_dstar.algorithmPart:
                        stmtcounter += 1
                        if stmtcounter == 2:
                                # if instA_timeLeft < 1e-10 then 
                                #     elems.append(instA_state); 
                                # end if;
                                for atomic in self.atomics:
                                        atomic_name = atomic[0]
                                        rule = deepcopy(i)
                                        rule.ifBody[0].funCall.operands[0].value = atomic_name + "_state"
                                        rule.ifTest.operands[0].value = atomic_name + "_timeLeft"
                                        new_algo.append(rule)
                        else:
                                new_algo.append(i)

                self.get_dstar.algorithmPart = new_algo


        def generateIntTransition(self):
                """
                We want to generate a function that chooses between the different internal transitions,
                depending on which timeLeft became 0, also calls the external transition functions for all
                required models
                """
                self.intTransition.className = "intTransition"
                counter = 0
                new_algo = []
                stmtcounter = 0
                for i in self.intTransition.algorithmPart:
                        stmtcounter += 1
                        if stmtcounter == 3:
                                # instA_timeLeft := instA_timeLeft - elapsed
                                for atomic in self.atomics:
                                        atomic_name = atomic[0]
                                        rule = deepcopy(i)
                                        rule.exp.operands[0].value = atomic_name + "_timeLeft"
                                        rule.compRef.operands[0].operands[0].value = atomic_name + "_timeLeft"
                                        new_algo.append(rule)
                        elif stmtcounter == 4:
                                # instA_elapsed := instA_elapsed + elapsed
                                for atomic in self.atomics:
                                        atomic_name = atomic[0]
                                        rule = deepcopy(i)
                                        rule.exp.operands[0].value = atomic_name + "_elapsed"
                                        rule.compRef.operands[0].operands[0].value = atomic_name + "_elapsed"
                                        new_algo.append(rule)
                        elif str(i.__class__.__name__) == "ConditionalEqStm" and ("dStar" in str(i.ifTest)):
                                # The final if statement, which checks which element has its internal transition
                                for atomic in self.atomics:
                                        rule = deepcopy(i)
                                        atomic_name = str(atomic[0])
                                        atomic_type = str(atomic[1].className)
                                        # Update the test
                                        rule.ifTest.operands[1].value = atomic_name + "_state"
                                        # Update the body
                                        counter = 0
                                        newrule = []
                                        for stm in rule.ifBody:
                                                counter += 1
                                                if counter == 1:
                                                        # Y = outputFnc_model1(instantiationA_state)
                                                        stm.exp.operands[1].operands[0].value = atomic_name + "_state"
                                                        stm.exp.operands[0].operands[0].operands[0].value = "outputFnc_" + atomic_type
                                                elif counter == 2:
                                                        # instantiationA_state = intTransition_modelA(instantiationA_state)
                                                        stm.compRef.operands[0].operands[0].value = atomic_name + "_state"
                                                        stm.exp.operands[0].operands[0].operands[0].value = "intTransition_" + atomic_type
                                                        stm.exp.operands[1].operands[0].value = atomic_name + "_state"
                                                elif counter == 3:
                                                        # Y := renamePorts(Y, instantiationA_name)
                                                        stm.exp.operands[1].operands[1].value = '"' + atomic_name + '"'
                                                elif counter == 5:
                                                        # instA_timeLeft := timeAdvance_model(state)
                                                        stm.compRef.operands[0].operands[0].value = atomic_name + "_timeLeft"
                                                        stm.exp.operands[0].operands[0].operands[0].value = "timeAdvance_" + str(atomic_type)
                                                        stm.exp.operands[1].operands[0].value = atomic_name + "_state"
                                                elif counter == 6:
                                                        stm.compRef.operands[0].operands[0].value = atomic_name + "_elapsed"
                                                elif counter not in range(7):
                                                        break
                                                newrule.append(stm)
                                        influencees = self.getInfluencees(atomic_name)
                                        for influencee in influencees:
                                                influencee_name = str(influencee)
                                                influencee_type = self.getType(influencee)
                                                if influencee_type == None:
                                                        # Couldn't find an influencee, so this is an output
                                                        continue
                                                for a in self.atomics:
                                                        if a[0] == influencee_name:
                                                                influencee_type = str(a[1].className)
                                                rule_part = rule.ifBody[6:]
                                                counter = 0
                                                for rpart in rule_part:
                                                        if counter == 0:
                                                                # Y_inst = getInstances(Y, "inst_A")
                                                                rpart.exp.operands[1].operands[1].value = influencee_name
                                                        elif counter == 1:
                                                                # extTransition_modelB(instantiationB_state, Y_inst, elapsed)
                                                                rpart.compRef.operands[0].operands[0].value = influencee_name + "_state"
                                                                rpart.exp.operands[1].operands[0].value = influencee_name + "_state"
                                                                rpart.exp.operands[0].operands[0].operands[0].value = "extTransition_" + influencee_type
                                                        elif counter == 2:
                                                                # instB_timeLeft := timeAdvance_modelB(instantiationB_state)
                                                                rpart.compRef.operands[0].operands[0].value = influencee_name + "_timeLeft"
                                                                rpart.exp.operands[0].operands[0].operands[0].value = "timeAdvance_" + str(influencee_type)
                                                                rpart.exp.operands[1].operands[0].value = influencee_name + "_state"
                                                        elif counter == 3:
                                                                # instB_elapsed := 0
                                                                rpart.compRef.operands[0].operands[0].value = influencee_name + "_elapsed"
                                                        counter += 1
                                                newrule.extend(rule_part)
                                        rule.ifBody = newrule
                                        new_algo.append(rule)
                        else:
                                new_algo.append(i)
                self.intTransition.algorithmPart = new_algo

        def generateExtTransition(self):
                """
                We want to generate a function that triggers the external transitions of all influenced
                components, this can be found from the inputs list
                """
                algorithm = self.extTransition.algorithmPart
                new_algo = []
                for i in algorithm:
                        if "keys.__contains__" in str(i):
                                for j in self.atomics:
                                        rule = deepcopy(i)
                                        atomicName = str(j[0])
                                        atomicType = str(j[1].className)
                                        ##################################################################
                                        # Update the if test itself
                                        rule.getIfTest().operands[1].operands[0].value = atomicName
                                        ##################################################################
                                        # Update the fetching of the Y dictionary for the instance
                                        rule.getIfBody()[0].exp.operands[1].operands[1].value = atomicName
                                        ##################################################################
                                        # Update the call to the external transition function
                                        rule.getIfBody()[1].exp.operands[0].operands[0].operands[0].value = "extTransition_" + str(atomicType)
                                        rule.getIfBody()[1].exp.operands[1].operands[0].value = str(atomicName) + "_state"
                                        rule.getIfBody()[1].exp.operands[1].operands[2].value = str(atomicName) + "_elapsed"
                                        rule.getIfBody()[1].compRef.operands[0].operands[0].value = str(atomicName) + "_state"
                                        ##################################################################
                                        # Update the time left
                                        rule.getIfBody()[2].exp.operands[0].operands[0].operands[0].value = "timeAdvance_" + str(atomicType)
                                        rule.getIfBody()[2].compRef.operands[0].operands[0].value = atomicName + "_timeLeft"
                                        rule.getIfBody()[2].exp.operands[1].operands[0].value = str(atomicName) + "_state"
                                        ##################################################################
                                        # Update the elapsed time
                                        rule.getIfBody()[3].compRef.operands[0].operands[0].value = atomicName + "_elapsed"
                                        new_algo.append(rule)
                        elif "peek" in str(i):
                                for j in self.inputs:
                                        rule = deepcopy(i)
                                        portname = j[2]
                                        rule.compRef.operands[0].operands[1].operands[0].value = str(portname)
                                        rule.exp.operands[1].operands[0].value = portname
                                        new_algo.append(rule)
                        elif "instA_timeLeft" in str(i):
                                for j in self.atomics:
                                        rule = deepcopy(i)
                                        atomicName = str(j[0])
                                        rule.exp.operands[0].value = atomicName + "_timeLeft"
                                        rule.compRef.operands[0].operands[0].value = atomicName + "_timeLeft"
                                        new_algo.append(rule)
                        elif "instA_elapsed" in str(i):
                                for j in self.atomics:
                                        rule = deepcopy(i)
                                        atomic_name = str(j[0])
                                        rule.exp.operands[0].value = atomic_name + "_elapsed"
                                        rule.compRef.operands[0].operands[0].value = atomic_name + "_elapsed"
                                        new_algo.append(rule)
                        else:
                                new_algo.append(i)
                self.extTransition.algorithmPart = new_algo
                self.extTransition.className = "extTransition"

        def generateOutputFnc(self):
                """
                Generate a function that sends all different outputs from the internal dictionary. The
                values that have to be sent can be checked from the outputs list
                """
                algorithm = self.outputFnc.algorithmPart
                new_algo = []
                for i in algorithm:
                        if "outputFnc" in str(i):
                                for atomic in self.atomics:
                                        rule = deepcopy(i)
                                        atomic_name = str(atomic[0])
                                        atomic_type = str(atomic[1].className)
                                        # Update the test
                                        rule.ifTest.operands[1].value = atomic_name + "_state"
                                        # Update the body
                                        counter = 0
                                        newrule = []
                                        for stm in rule.ifBody:
                                                counter += 1
                                                if counter == 1:
                                                        # Y = outputFnc_model1(instantiationA_state)
                                                        stm.exp.operands[1].operands[0].value = atomic_name + "_state"
                                                        stm.exp.operands[0].operands[0].operands[0].value = "outputFnc_" + atomic_type
                                                elif counter == 2:
                                                        stm.exp.operands[1].operands[1].value = StringExp(-1, -1, atomic_name)
                                        new_algo.append(rule)
                        elif "poke" in str(i):
                                for j in self.outputs:
                                        rule = deepcopy(i)
                                        inst, port, port2 = j
                                        operands = rule.funCall.operands
                                        rule.funCall.operands[0].value = IdentExp(-1, -1, port2)
                                        prule = rule.funCall.operands[1].operands[1].operands[0].value = str(port2)
                                        new_algo.append(rule)
                        else:
                                new_algo.append(i)
                self.outputFnc.algorithmPart = new_algo
                self.outputFnc.className = "outputFnc"

        def generateFunctions(self):
                lst = self.flattened.getClassDefinitions()[0].getElementList()
                for i in lst:
                        if (i.__class__.__name__ == "RegularClassDefinition") and (str(i.getName()) == "NewAtomic"):
                                lst = i.getElementList()
                for i in lst:
                        #di = deepcopy(i)
                        di = i
                        if hasattr(i, "getKind") and str(i.getKind()) == "function":
                                name = str(i.getName())
                                if "timeAdvance" in name:
                                        self.timeAdvance = di
                                elif "outputFnc" in name:
                                        self.outputFnc = di
                                elif "intTransition" in name:
                                        self.intTransition = di
                                elif "extTransition" in name:
                                        self.extTransition = di
                                elif "mapConnections" in name:
                                        self.mappingFunction = di
                                elif "get_dstar" in name:
                                        self.get_dstar = di
                                elif "getModelKeys" in name:
                                        self.getModelKeys = di

                                # Only append the standard select function of no custom one is defined
                                if "select" in name and not self.customSelect:
                                        self.select = di
                                        self.functions.append(di)
                                elif "select" not in name:
                                        self.functions.append(di)
                        else:
                                self.functions.append(di)

        def generateTimeAdvance(self):
                """
                Return the least timeLeft value of all submodels
                """
                self.timeAdvance.className = "timeAdvance"
                algo = []
                old_algo = self.timeAdvance.algorithmPart
                algo.append(old_algo[0])
                for i in self.atomics:
                        rule = deepcopy(old_algo[1])
                        atomic_name = str(i[0])
                        rule.funCall.operands[0].value = atomic_name + "_timeLeft"
                        algo.append(rule)
                algo.extend(old_algo[2:])
                self.timeAdvance.algorithmPart = algo

        def generateSelect(self):
                for func in self.functions:
                        if str(func.__class__.__name__) == "RegularClassDefinition" and str(func.className) == "select":
                                self.select = func
                if self.select != None:
                        self.select.className = "select"

        def getASTElement(self, outerScope):
                elementFuncList = []
                elementVarList = []
                for func in self.functions:
                        if func.__class__.__name__ == "RegularClassDefinition":
                                elementFuncList.append(func)
                        elif func.__class__.__name__ == "ComponentClause":
                                elementVarList.append(func)
                        elif func.__class__.__name__ == "ExtendsClause":
                                elementVarList.insert(0, func)
                elementList = elementVarList + elementFuncList
                element = RegularClassDefinition(-1, -1, ClassPrefix(), "class", self.name, None, elementList)
                flatclass = self.flattened
                element.scope = flatclass.getClassDefinitions()[0].scope
                element.scope.outer = outerScope
                #return [element, self.state]
                return [element, None]

class DevsFlattener:
        def __init__(self, ast):
                self.ast = ast
                self.flatClass = self.ast.classFiles[1]
                self.classes = {}
                self.types = {}
                self.depends = {}

        def go(self):
                mainModel = self.ast.classFiles[0].classDefinitions[0]
                self.flattenClass(mainModel)
                mainModel = self.ast.classFiles[0].classDefinitions[0]
                return mainModel

        def fetchClassDef(self, name):
                for i in self.ast.classFiles[0].classDefinitions[0].elementList:
                        if (i.__class__.__name__ == "RegularClassDefinition") and (str(i.className) == name):
                                return i

        def instantiate(self, name, entry):
                if name != "":
                        ename = name + "_"
                else:
                        ename = name
                for i in entry.elementList:
                        if (i.__class__.__name__ == "ComponentClause") and (str(i.type) not in ["String", "DevsPort"]) and ("State" not in str(i.type)):
                                newname = ename + i.componentList[0].ident.value
                                newentry = self.fetchClassDef(str(i.type))
                                self.inst_elems.append([newname, str(i.type)])
                                self.instantiate(newname, newentry)
                for i in entry.getConnectEqStms():
                        begin = str(i.getFirstConnector()).split(".")
                        end = str(i.getSecondConnector()).split(".")
                        if len(begin) < 2:
                                self.inst_conns.append([name, begin[0], ename + end[0], end[1]])
                        elif len(end) < 2:
                                self.inst_conns.append([ename + begin[0], begin[1], name, end[0]])
                        else:
                                self.inst_conns.append([ename + begin[0], begin[1], ename + end[0], end[1]])

        def directConnect(self, root, coupleds):
                scopeEntry = root.getScope().lookup("rootModel").getDefinition()
                self.inst_elems = []
                self.inst_conns = []
                instantiations = self.instantiate("", scopeEntry)
                changed = True
                ignores = []
                while changed:
                        changed = False
                        for i in self.inst_conns:
                                for j in self.inst_conns:
                                        if i[0] == j[2] and i[1] == j[3]:
                                                ignores.append(i[0])
                                                # Create a direct connection
                                                i[0] = j[0]
                                                i[1] = j[1]
                                                changed = True
                                        elif i[2] == j[0] and i[3] == j[1]:
                                                ignores.append(i[2])
                                                # Create a direct connection
                                                i[2] = j[2]
                                                i[3] = j[3]
                                                changed = True
                                        if changed:
                                                self.inst_conns.remove(j)
                                                break
                                if changed:
                                        break

                root = scopeEntry
                # Clear the elements, only keep the first ones (extend and name)
                newlist = [root.elementList[0]]
                for i in root.elementList:
                        if (i.__class__.__name__ == "ComponentClause") and (str(i.type) == "DevsPort"):
                                newlist.append(i)
                root.elementList = newlist
                # Clear connections
                root.equationPart = []

                for i in self.inst_conns:
                        if i[0] != "":
                                first = i[0] + "." + i[1]
                        else:
                                first = i[1]
                        if i[2] != "":
                                second = i[2] + "." + i[3]
                        else:
                                second = i[3]
                        root.equationPart.append(ConnectEqStm(-1, -1, first, second))
                new_elems = []
                for i in self.inst_elems:
                        if i[0] not in ignores:
                                new_elems.append(i)
                self.inst_elems = new_elems
                for i in self.inst_elems:
                        ident_type = IdentExp(-1, -1, i[1])
                        nameexp = NameExp(-1, -1, [ident_type])
                        declaration = IdentExp(-1, -1, i[0])
                        elem = ComponentClause(-1, -1, TypePrefix(), nameexp, [declaration])
                        root.elementList.append(elem)

        def flattenClass(self, classDef):
                atomics = []
                atomics_str = []
                coupleds = []
                coupled_defs = []
                coupleds_str = []
                states = []
                for i in classDef.getScope().getDefinitions():
                        definition = i.getDefinition()
                        extends = definition.getExtends()
                        if str(extends) == "AtomicDEVS":
                                atomics.append(i)
                                atomics_str.append(str(i))
                                self.classes[str(definition.className)] = definition
                        elif str(extends) == "CoupledDEVS":
                                coupleds.append(i)
                                self.depends[i] = []
                                coupled_defs.append(definition)
                                coupleds_str.append(str(i))
                        elif "State" in str(definition.className) and "DevsState" not in str(definition.className):
                                states.append(i)
                        elif "Experiment" in str(definition.className) and "DevsExperiment" not in str(definition.className):
                                root = definition
                                # Strip away the name parameter
                                root.elementList[1].componentList[0].modification = "()"
                                roottype = str(root.getScope().lookup("rootModel").getDefinition().className)
                self.directConnect(root, coupleds)

                for i in atomics:
                        self.flattenAtomicCode(i)

                for i in states:
                                definition = i.getDefinition()
                                name = str(definition.className)[:-5]
                                foundclass = self.classes[name]
                                newlist = []
                                for defin in definition.elementList:
                                        newname = name + "_SeqStates"
                                        defin.type = NameExp(-1, -1, [newname])
                                        start = defin.componentList[0].modification.argumentList[0].modification.exp.operands
                                        newstart = start[2:]
                                        defin.componentList[0].modification.argumentList[0].modification.exp.operands = newstart
                                        newlist.append(self.types[newname])
                                        newlist.append(defin)
                                definition.elementList = newlist
                        
                elems = classDef.getElementList()

                scopeEntry = root.getScope().lookup("rootModel")

                #for i in coupleds:
                for i in [scopeEntry]:
                        [atomic, state] = self.flattenCoupledCode(i, atomics_str, coupleds_str)
                        atomic.className = roottype
                        elems = classDef.getElementList()
                        for j in range(len(elems)):
                                elem = elems[j]
                                if str(elem) == str(i.definition):
                                        elems[j] = atomic
                        #classDef.elementList.append(state)
                elems = []

                for i in classDef.getElementList():
                        # Strip away all OLD atomic models
                        if (i.__class__.__name__ == "RegularClassDefinition") and (("AtomicDEVS" in i.getExtends()) or ("CoupledDEVS" in i.getExtends())):
                                continue
                        else:
                                elems.append(i)
                classDef.elementList = elems

        def searchDefinition(self, name):
                for i in self.ast.classFiles[0].classDefinitions[0].elementList:
                        if str(i.__class__.__name__) != "RegularClassDefinition":
                                continue
                        if name == str(i.className):
                                return i

        def flattenCoupledCode(self, coupled, atomics, coupleds):
                definition = coupled.getDefinition()
                elems = definition.elementList
                submodels = []
                ports = []
                for i in elems:
                        if (i.__class__.__name__ == "ComponentClause") and (str(i.type) not in ["dict", "list", "Real", "DevsPort", "String"]):
                                definition = self.searchDefinition(str(i.type))
                                name = str(i.componentList[0])
                                namelist = name.split("(")
                                name = namelist[0]
                                submodels.append([str(i.componentList[0]), definition])
                        elif (i.__class__.__name__ == "ComponentClause") and (str(i.type) == "DevsPort"):
                                ports.append(i)
                atomic = TempAtomic(str(coupled), self.flatClass)
                atomic.customSelect = False
                for func in definition.elementList:
                        if str(func.__class__.__name__) == "RegularClassDefinition" and str(func.className) == "select":
                                atomic.functions.append(func)
                                atomic.customSelect = True
                for port in ports:
                        atomic.functions.append(port)
                for var in submodels:
                        atomic.atomics.append(var)
                        self.addAtomic(atomic, var)
                for connection in coupled.getDefinition().getConnectEqStms():
                        begin = str(connection.getFirstConnector()).split(".")
                        end = str(connection.getSecondConnector()).split(".")
                        if len(begin) < 2:
                                # An External Input Connection
                                atomic.inputs.append([begin[0], end[0], end[1]])
                                atomic.connections.append(["", begin[0], end[0], end[1]])
                        elif len(end) < 2:
                                # An External Output Connection
                                atomic.outputs.append([begin[0], begin[1], end[0]])
                                atomic.connections.append([begin[0], begin[1], "", end[0]])
                        else:
                                # An Internal Connection
                                atomic.connections.append([begin[0], begin[1], end[0], end[1]])
                atomic.generateFunctions()
                atomic.generateVariables()
                atomic.generateMappings()
                atomic.generateGetDStar()
                atomic.generateGetModelKeys()
                atomic.generateIntTransition()
                atomic.generateExtTransition()
                atomic.generateTimeAdvance()
                atomic.generateOutputFnc()
                atomic.generateSelect()
                #definition.getScope().add(str(coupled), DefinitionEntry(-1, -1, str(coupled), definition, "PUBLIC"))
                outerScope = definition.getScope()
                [el, state] = atomic.getASTElement(outerScope)
                # Set the scope to the original one
                el.scope = coupled.getDefinition().scope
                return [el, state]

        def addAtomic(self, atomic, var):
                submodelname = var[0]
                definition = var[1]
                scope = definition.getScope().getTable()
                for var in definition.getParameters():
                        name = submodelname + "_" + str(var[0])
                        defentry = scope[str(var[0])]
                        typ = defentry.getDefinition().getName()
                        atomic.variables.append([name, typ])
                for elem in definition.getElementList():
                        #defentry = scope[str(elem.getName())]
                        if elem.__class__.__name__ == "ComponentClause":
                                continue
                        else:
                                if elem.__class__.__name__ != "RegularClassDefinition":
                                        atomic.functions.append(elem)
                                        continue
                                found = False
                                for i in atomic.functions:
                                        if str(i.__class__.__name__) != "RegularClassDefinition":
                                                continue
                                        if str(i.className) == str(elem.className):
                                                found = True
                                                break
                                if not found:
                                        atomic.functions.append(elem)
                name = submodelname + "_state"
                #defentry = scope["state"]
                #typ = defentry.getDefinition().getName()
                typ = submodelname + "State"
                atomic.variables.append([name, typ])
                name = submodelname + "_state"
                # Ports shouldn't be copied, since they are no longer needed due to the X and Y variables

        def flattenAtomicCode(self, atomic):
                definition = atomic.getDefinition()
                variables = definition
                stmts = definition.getElementList()
                for i in stmts:
                        if i.__class__.__name__ == "RegularClassDefinition" and i.getKind() == "function":
                                data = i.getAlgorithm()
                                self.changePeekAndPoke(variables, data)
                                self.modify_transitions(i, atomic)
                # Now also add these variables to the input/output list
                elemlist = []
                for i in definition.elementList:
                        if i.__class__.__name__ == "ComponentClause" and i.isInput():
                                continue
                        if i.__class__.__name__ == "ComponentClause" and i.isOutput():
                                continue
                        if i.__class__.__name__ == "RegularClassDefinition" and i.getKind() == "function":
                                addState = True
                                addX = False
                                addY = False
                                addElapsed = False
                                if "intTransition" in i.className:
                                        pass
                                elif "extTransition" in i.className:
                                        addX = True
                                        addElapsed = True
                                elif "outputFnc" in i.className:
                                        addY = True
                                elif "timeAdvance" in i.className:
                                        pass
                                ###################################################################
                                if addState:
                                        declaration = Declaration(-1, -1, IdentExp(-1, -1, "state"))
                                        state_type = str(atomic) + "State"
                                        input_state = ComponentClause(-1, -1, TypePrefix(causalityP="INPUT"), NameExp(-1, -1, [IdentExp(-1, -1, state_type)]), [declaration])
                                        i.elementList.append(input_state)
                                        RHS_name = "state"
                                        RHS_definition = DefinitionEntry(-1, -1, RHS_name, definition, 'PUBLIC')
                                        i.getScope().add(RHS_name, RHS_definition)
                                ###################################################################
                                if addX:
                                        declaration = Declaration(-1, -1, IdentExp(-1, -1, "X"))
                                        input_X = ComponentClause(-1, -1, TypePrefix(causalityP="INPUT"), NameExp(-1, -1, [IdentExp(-1, -1, "dict")]), [declaration])
                                        i.elementList.append(input_X)
                                        RHS_name = "X"
                                        RHS_definition = DefinitionEntry(-1, -1, RHS_name, definition, 'PUBLIC')
                                        i.getScope().add(RHS_name, RHS_definition)
                                ###################################################################
                                if addElapsed:
                                        declaration = Declaration(-1, -1, IdentExp(-1, -1, "elapsed"))
                                        input_state = ComponentClause(-1, -1, TypePrefix(causalityP="INPUT"), NameExp(-1, -1, [IdentExp(-1, -1, "Integer")]), [declaration])
                                        i.elementList.append(input_state)
                                        RHS_name = "elapsed"
                                        RHS_definition = DefinitionEntry(-1, -1, RHS_name, definition, 'PUBLIC')
                                        i.getScope().add(RHS_name, RHS_definition)
                                ###################################################################
                                if addY:
                                        declaration = Declaration(-1, -1, IdentExp(-1, -1, "Y"), modificationP="=dict()")
                                        input_state = ComponentClause(-1, -1, TypePrefix(causalityP="OUTPUT"), NameExp(-1, -1, [IdentExp(-1, -1, "dict")]), [declaration])
                                        i.elementList.append(input_state)
                                        RHS_name = "Y"
                                        RHS_definition = DefinitionEntry(-1, -1, RHS_name, definition, 'PUBLIC')
                                        i.getScope().add(RHS_name, RHS_definition)
                                ###################################################################
                                elemlist.append(i)
                        elif i.__class__.__name__ == "RegularClassDefinition" and i.getKind() == "type":
                                typename = str(atomic) + "_" + str(i.className)
                                i.className = typename
                                self.types[typename] = i
                        else:
                                elemlist.append(i)
                definition.elementList = elemlist

        def modify_transitions(self, function, parent):
                function.className = str(function.className) + "_" + str(parent)
                for i in function.algorithmPart:
                        if "SeqStates" in str(i):
                                i.ifTest.operands[1].operands[0].operands[0].value = str(parent) + "State"
                                for j in i.ifBody:
                                        if "SeqStates" in str(j):
                                                j.exp.operands[0].operands[0].value = str(parent) + "State"

        def changePeekAndPoke(self, variables, algorithm):
                tocheck = []
                for i in algorithm:
                        if "peek" in str(i) or "poke" in str(i):
                                tocheck.append(i)
                for i in tocheck:
                        found = False
                        for attrName in dir(i):
                                attr = getattr(i, attrName)
                                if type(attr) != type([]):
                                        continue
                                newattr = []
                                for j in attr:
                                        if "peek" in str(j) or "poke" in str(j):
                                                oldattr = attr
                                                newattr.extend(self.changePeekAndPoke(variables, attr))
                                                setattr(i, attrName, attr)
                                                found = True
                                        else:
                                                newattr.append(j)
                                setattr(i, attrName, newattr)
                        if not found:
                                if "peek" in str(i.exp):
                                        LHS = i.compRef
                                        RHS_name = 'X.get("' + str(str(i.exp)[5:-1]) + '", None)'
                                        RHS_identifier = IdentExp(-1, -1, RHS_name)
                                        #RHS_definition = DefinitionEntry(-1, -1, RHS_name, variables, 'PUBLIC')
                                        newExpr = AtomicAlgStm1(-1, -1, LHS, RHS_identifier)
                                        #variables.getScope().add(RHS_name, RHS_definition)
                                        result = [newExpr]
                                else:
                                        LHS_name = 'Y["' + str(i.funCall).split(",")[0] + '"]'
                                        LHS_identifier = IdentExp(-1, -1, LHS_name)
                                        #LHS_definition = DefinitionEntry(-1, -1, LHS_name, variables, 'PUBLIC')
                                        RHS_name = str(i.funCall).split(",")[1]
                                        RHS_identifier = IdentExp(-1, -1, RHS_name)
                                        newExpr = AtomicAlgStm1(-1, -1, LHS_identifier, RHS_identifier)
                                        #variables.getScope().add(LHS_name, LHS_definition)
                                        # We now have output to send, so also set the flag
                                        #LHS_name = "isSendMsg"
                                        #LHS_identifier = IdentExp(-1, -1, LHS_name)
                                        #sendFlag = AtomicAlgStm1(-1, -1, LHS_identifier, BoolExp(-1, -1, True))
                                        #result = [newExpr, sendFlag]
                                        result = [newExpr]
                                return result
