# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University



from AST.ASTExp import IdentExp
from AST.ASTScope import ComponentPrefix, DefinitionEntry
from AST.ASTScope import DeclarationEntry, SymbolTable
from FrontEnd.Visitor import AbstractVisitor

# ---------------------------------------------------------------------

class DevsScoper(AbstractVisitor):
    "add declared identifiers into symboltable"

    topScope=None
    currentScope=None

    def __init__(self, ast):
	self.topScope=ast.topScope
	self.currentScope=ast.topScope

    def setCurrentScope(self, scope):
	self.currentScope=scope


    # ---------------------------AST------------------------------
    def visitAST(self, ast):
   	self.topScope=ast.topScope
	self.setCurrentScope(ast.topScope)
	ast.predefined.accept(self)
        for i in ast.classFiles:
                i.accept(self)

    # ---------------------------ClassDefinition-----------------------
    def visitClassFile(self, classFile):
	#self.setCurrentScope(classFile.getScope())
	for i in classFile.classDefinitions:
	    i.accept(self)
	#print "Scoper has visited: ", classFile.fileName
	#print "Scope node: ", classFile.classDefinitions[0].scope.table.keys()

    def visitRegularClassDefinition(self, rClassDefinition):
	lineNum=-1
	colNum=-1
	name=rClassDefinition.getName() # IdentExp
	definition=rClassDefinition
	visibility=rClassDefinition.getVisibility()
	#originalScope=self.currentScope
	#modif=[]
	entry=DefinitionEntry(lineNum, colNum, name, definition, visibility)
	flag=self.currentScope.add(str(name), entry)
	if not flag:
	    raise 'Error: 2 elements have the same name in a lexical scope!'
	rClassDefinition.setOuterScope(self.currentScope)
	# set new currentScope
	self.setCurrentScope(rClassDefinition.getScope())
	# add 'self'
	selfName=IdentExp(-1,-1,'self')
	selfEntry=DefinitionEntry(lineNum, colNum, selfName, definition, \
                        visibility)#, originalScope)#, modif)
	flag=rClassDefinition.getScope().add('self', selfEntry)
	if not flag:
            raise 'Error: 2 elements have the same name in a lexical scope!'
	# add DeclarationEntry
	for i in rClassDefinition.elementList:
	    i.accept(self)
        for i in rClassDefinition.initEquationPart:
	    i.accept(self)
        for i in rClassDefinition.equationPart:
	    i.accept(self)
	# reset currentScope
	self.setCurrentScope(self.currentScope.getOuter())

    def visitShortClassDefinition(self, sClassDefinition):
	lineNum=-1
        colNum=-1
        name=sClassDefinition.getName()
        definition=sClassDefinition
	visibility=sClassDefinition.getVisibility()
	#originalScope=self.currentScope
	#modif=[]
	#modif.append(sClassDefinition.getModification())
        entry=DefinitionEntry(lineNum, colNum, name, definition, \
			visibility)#, originalScope)#, modif)
	entry.addModif(sClassDefinition.getModification())
        flag=self.currentScope.add(str(name), entry)
	if not flag:
            raise 'Error: 2 elements have the same name in a lexical scope!'
	# add 'self'
        selfName=IdentExp(-1,-1,'self')
        selfEntry=DefinitionEntry(lineNum, colNum, selfName, definition, \
                        visibility)#, originalScope)#, modif)
        flag=sClassDefinition.getScope().add('self', selfEntry)
	sClassDefinition.setOuterScope(self.currentScope)

    def visitEnumClassDefinition(self, eClassDefinition):	
	# EnumClassDefinition has been transformed to 
	# RegularClassDefinition
	pass

    def visitComponentClause(self, componentClause):
	# should check that a constant declaration has been given a value
	lineNum=-1
        colNum=-1
	definition=None
	#originalScope=self.currentScope
	isFinal=componentClause.getFinal()
	innerOuter=componentClause.getInnerOuter()
	isReplaceable=componentClause.getReplaceable()
	componentPrefix=ComponentPrefix(isFinal, innerOuter, isReplaceable)
	typePrefix=componentClause.getTypePrefix()
	type=componentClause.getType()
	#value=None
	typeArrSub=componentClause.getArrSub()
	for i in componentClause.getComponentList():
	    name=i.getName()
	    nameArrSub=i.getArrSub()
	    #modif=[]
	    #if not i.getModification()==None:
		#modif.append(i.getModification())
	    entry=DeclarationEntry(lineNum, colNum, name, definition, \
			#originalScope, 
			componentPrefix, typePrefix, type,\
			typeArrSub, nameArrSub)#, modif)#, value)
	    if not i.getModification()==None:
		entry.addModif(i.getModification())
	    flag=self.currentScope.add(str(name), entry)
	    if not flag:
            	raise 'Error: 2 elements have the same name in a lexical scope!'

    def visitReplaceableClause(self, replaceableClause):
	raise "\"Replaceable\" found: not supported! Abort ... "

    def visitImportClause(self, importClause):
	raise "\"Import clause\" found: not yet supported! Abort ..."

    def visitExtendsClause(self, extendsClause):
	pass

    # ------------------------------EqStm-----------------------------
    def visitRegularEqStm(self, eqStm):
	eqStm.LHS.accept(self)
	eqStm.RHS.accept(self)

    def visitConnectEqStm(self, eqStm):
	eqStm.first.accept(self)
	eqStm.second.accept(self)

    def visitFunCallEqStm(self, eqStm):
	eqStm.funCall.accept(self)

    def visitConditionalEqStm(self, eqStm):
	eqStm.getIfTest().accept(self)
	for i in eqStm.getIfBody():
	    i.accept(self)
	for i in eqStm.getElseifList():
	    i.accept(self)
	for i in eqStm.getElseBody():
	    i.accept(self)

    def visitForEqStm(self, eqStm):
	lineNum=-1
        colNum=-1
	definition=None
	#originalScope=self.currentScope
	componentPrefix=None
        typePrefix=None
        type=None
        typeArrSub=None
	nameArrSub=None
	#modif=[]
	#value=None
	# enter new scope
	eqStm.setOuterScope(self.currentScope)
	self.setCurrentScope(eqStm.getScope())
	forIndices=eqStm.getForIndices() # ForIndicesExp
	operands=forIndices.getOperands() # list[] of IdentInExp
	for i in operands: # i: IdentInExpExp
	    name=i.getOperands()[0] # IdentExp
	    entry=DeclarationEntry(lineNum, colNum, name, definition, \
                        #originalScope, 
			componentPrefix, typePrefix, type,\
                        typeArrSub, nameArrSub)#, modif)#, value)
	    flag=self.currentScope.add(str(name), entry)
	    if not flag:
            	raise 'Error: 2 elements have the same name in a lexical scope!'
	for i in eqStm.getBody():
	    i.accept(self)
	# reset currentScope
        self.setCurrentScope(self.currentScope.getOuter())
	
    def visitWhenEqStm(self, eqStm):
	eqStm.getWhenTest().accept(self)
	for i in eqStm.getBody():
	    i.accept(self)
	for i in eqSTm.getElsewhenList():
	    i.accept(self)


    # ------------------------------Exp-----------------------------
    def visitIdentExp(self, exp):
	pass

    def visitRealExp(self, exp):
        pass

    def visitIntegerExp(self, exp):
        pass

    def visitBoolExp(self, exp):
        pass

    def visitStringExp(self, exp):
        pass

    def visitNameExp(self, exp):
	pass

    def visitOrExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitAndExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitNotExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitLessExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def vistLessEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitGreaterExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitGreaterEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitEQEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitLessGtExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSumExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitMulExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSubtractExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitDivExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitUnaryMinusExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitPowerExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitRangeExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitIfElseExp(self, exp):
	exp.ifTest.accept(self)
	exp.ifBody.accept(self)
	for i in exp.elseifList:
	    i.accept(self)
	exp.elseBody.accept(self)

    def visitFunCallExp(self, exp):
	exp.getOperands()[1].accept(self) # FunArgsExp

    def visitFunArgsExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitForFunArgExp(self, exp):
	lineNum=-1
        colNum=-1
        definition=None
	#originalScope=self.currentScope
        componentPrefix=None
        typePrefix=None
        type=None
        typeArrSub=None
        nameArrSub=None
	#modif=[]
        #value=None
        # enter new scope
        exp.setOuterScope(self.currentScope)
        self.setCurrentScope(exp.getScope())
        operands=exp.getOperands() # list[] of Exp & IdentInExp
	operands[0].accept(self)
	j=1
	numOfOperands=len(operands)
	while j<numOfOperands:
	    i=operands[j]
            name=i.getOperands()[0] # IdentExp
	    entry=DeclarationEntry(lineNum, colNum, name, definition, \
                        #originalScope, 
			componentPrefix, typePrefix, type,\
                        typeArrSub, nameArrSub)#, modif)#, value)
            flag=self.currentScope.add(str(name), entry)
	    if not flag:
            	raise 'Error: 2 elements have the same name in a lexical scope!'
	    j=j+1
        # reset currentScope
        self.setCurrentScope(self.currentScope.getOuter())

    def visitNamedArgExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitConnectRefExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSingleCompRefExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitCompRefExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitArrSubExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitCommentExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitAnnotationExp(self, exp):
        pass

    def visitIdentInExpExp(self, exp):
        exp.getOperands()[1].accept(self)

    def visitForIndicesExp(self, exp):
        for i in exp.getOperands():
            # i: IdentInExpExp
            i.accept(self)

    def visitExpListExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    # -------------------------auxilially-------------------
    def isPredefined(self, name):
	if name=='Real' or name=='Integer' or name=='Boolean' or \
	   name=='String' or name=='StateSelect':
	    return True
	else:
	    return False

