
from xml.dom.minidom     import *
from sets                import Set

class MappedStructure(object):
  def __init__(self, mapName, mapPackage):
    self.mapName = mapName
    self.mapPackage = mapPackage

  def __str__(self):
    return self.mapPackage+'.'+self.mapName
#--------------------------------------------------------------
#Parse the whole function mapping file
#-----------------------------------------------------------------------
class MappingParser(object):
  def __init__(self, fileName):
    self.__fileName     = fileName
    self.__funcSet      = Set()
    self.__funcMap      = {}
    self.__packMap      = {}
    self.__initialize()

  def __initialize(self):
    doc     = parse(self.__fileName)
    mapping = doc.getElementsByTagName("mappings")[0]
    self.__parsePackages(mapping)
    self.__parseFunctions(mapping)

#--------------------------------------------------------------
# traverse the mapping file
# save function mappings to a dictionary
#------------------------------------------------------------------------
  def __parseFunctions(self, mapping):  
    functions = mapping.getElementsByTagName("function-mapping")
    for function in functions:
      funcName = function.getElementsByTagName("name")[0]
      txtName = funcName.childNodes[0].data      
      mapFuncName = function.getElementsByTagName("mapname")[0]
      txtMapFuncName = mapFuncName.childNodes[0].data
      mapPackName = function.getElementsByTagName("mappackage")[0]
      txtMapPackName = mapPackName.childNodes[0].data
      m = MappedStructure(txtMapFuncName, txtMapPackName)
      self.__funcMap[txtName] = m

  def getFunctionMap(self):
    return self.__funcMap

#--------------------------------------------------------------
# traverse the mapping file
# save package mappings to a dictionary
#------------------------------------------------------------------------
  def __parsePackages(self, mapping):  
    packages = mapping.getElementsByTagName("package-mapping")
    for package in packages:
      packName = package.getElementsByTagName("name")[0]
      txtName = packName.childNodes[0].data      
      mapPackName = package.getElementsByTagName("mappackage")[0]
      txtMapPackName = mapPackName.childNodes[0].data
      self.__packMap[txtName] = txtMapPackName

  def getPackageMap(self):
    return self.__packMap


if __name__ == '__main__':
  import sys  
  if(len(sys.argv)<2):
    print "\nUsage: python mappingparser.py XMLFILE"
    sys.exit(2)

  tParser = MappingParser(sys.argv[1])
  fMap = tParser.getFunctionMap()
  for f in fMap.keys():
    print f, fMap[f]

