
import sys
import os.path
sys.path.append(os.path.expanduser('~/DEVSEnv/pydevs/'))
from DEVS import *
from whrandom import randint
from simulator import *

class GeneratorState:
  def __init__(self):
    self.seqState = Generator.G_IDLE

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nseqState: " + str(self.seqState)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>seqState</name>"
    strRep = strRep + "\n\t<type>Generator.SeqStates</type>"
    strRep = strRep + "\n\t<value>"+str(self.seqState)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Generator( AtomicDEVS ):
  G_IDLE = 'G_IDLE'
  G_GENERATING = 'G_GENERATING'
  def __init__(self, ia, ib, szl, szh, name):
    AtomicDEVS.__init__(self, name)
    self.ia = ia
    self.ib = ib
    self.szl = szl
    self.szh = szh
    self.name = name
    self.g_out = self.addOutPort("g_out")
    self.state = GeneratorState()
    
    
  def intTransition( self ):
    if(self.state.seqState == Generator.G_IDLE):
      self.state.seqState = Generator.G_GENERATING
    elif(self.state.seqState == Generator.G_GENERATING):
      self.state.seqState = Generator.G_IDLE
    return self.state

  
  def outputFnc( self ):
    evt = None
    if(self.state.seqState == Generator.G_GENERATING):
      evt = Job(self.szl, self.szh)
      self.poke(self.g_out, evt)


  
  def timeAdvance( self ):
    if(self.state.seqState == Generator.G_IDLE):
      timespan = randint(self.ia, self.ib)

    if(self.state.seqState == Generator.G_GENERATING):
      timespan = 0
    return timespan


  def __str__(self):
    strRep = ''
    strRep = strRep + "\nstate: " + str(self.state)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>state</name>"
    strRep = strRep + "\n\t<type>GeneratorState</type>"
    if(self.state!=None):
      strRep = strRep + "\n\t<value>"+self.state.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class ProcessorState:
  def __init__(self):
    self.currentJob = None
    self.queue = []
    self.queueSize = 0
    self.timeElapsed = 0.0
    self.seqState = Processor.P_IDLE

  def __str__(self):
    strRep = ''
    strRep = strRep + "\ncurrentJob: " + str(self.currentJob)
    strRep = strRep + "\nqueue: "
    for item in self.queue:
      strRep = strRep + "\n\t" + str(item)
    strRep = strRep + "\nqueueSize: " + str(self.queueSize)
    strRep = strRep + "\ntimeElapsed: " + str(self.timeElapsed)
    strRep = strRep + "\nseqState: " + str(self.seqState)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>currentJob</name>"
    strRep = strRep + "\n\t<type>Job</type>"
    if(self.currentJob!=None):
      strRep = strRep + "\n\t<value>"+self.currentJob.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>queue</name>"
    strRep = strRep + "\n\t<type>list</type>"
    strRep = strRep + "\n\t<value>"
    for item in self.queue:
      strRep = strRep + "\n\t" + item.toXML()
    strRep = strRep + "\n\t</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>queueSize</name>"
    strRep = strRep + "\n\t<type>Integer</type>"
    strRep = strRep + "\n\t<value>"+str(self.queueSize)+"</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>timeElapsed</name>"
    strRep = strRep + "\n\t<type>Real</type>"
    strRep = strRep + "\n\t<value>"+str(self.timeElapsed)+"</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>seqState</name>"
    strRep = strRep + "\n\t<type>Processor.SeqStates</type>"
    strRep = strRep + "\n\t<value>"+str(self.seqState)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Processor( AtomicDEVS ):
  P_IDLE = 'P_IDLE'
  P_BUSY = 'P_BUSY'
  P_DISCARD = 'P_DISCARD'
  def __init__(self, qSize, name):
    AtomicDEVS.__init__(self, name)
    self.qSize = qSize
    self.name = name
    self.p_in = self.addInPort("p_in")
    self.p_out = self.addOutPort("p_out")
    self.p_discard = self.addOutPort("p_discard")
    self.state = ProcessorState()
    
    
    self.initialization()
  def initialization( self ):
    self.state.queueSize = self.qSize


  
  def extTransition( self ):
    if(self.state.seqState == Processor.P_IDLE):
      self.state.currentJob = self.peek(self.p_in)
      self.state.timeElapsed = 0
      self.state.seqState = Processor.P_BUSY
    elif(self.state.seqState == Processor.P_BUSY):
      self.state.timeElapsed = self.state.timeElapsed + self.elapsed
      if(len(self.state.queue) < self.state.queueSize):
        self.state.queue.append(self.peek(self.p_in))
        self.state.seqState = Processor.P_BUSY
      elif(len(self.state.queue) == self.state.queueSize):
        self.state.seqState = Processor.P_DISCARD
    return self.state

  
  def intTransition( self ):
    if(self.state.seqState == Processor.P_BUSY):
      if(len(self.state.queue) == 0):
        self.state.timeElapsed = 0
        self.state.currentJob = None
        self.state.seqState = Processor.P_IDLE
      elif(len(self.state.queue) > 0):
        self.state.timeElapsed = 0
        self.state.currentJob = self.state.queue.pop(0)
        self.state.seqState = Processor.P_BUSY
    elif(self.state.seqState == Processor.P_DISCARD):
      self.state.seqState = Processor.P_BUSY
    return self.state

  
  def outputFnc( self ):
    evt = None
    if(self.state.seqState == Processor.P_BUSY):
      self.poke(self.p_out, self.state.currentJob)
      self.state.currentJob = None

    if(self.state.seqState == Processor.P_DISCARD):
      self.poke(self.p_discard, self.peek(self.p_in))


  
  def timeAdvance( self ):
    if(self.state.seqState == Processor.P_IDLE):
      timespan = INFINITY

    if(self.state.seqState == Processor.P_BUSY):
      timespan = self.state.currentJob.size - self.state.timeElapsed

    if(self.state.seqState == Processor.P_DISCARD):
      timespan = 0
    return timespan


  def __str__(self):
    strRep = ''
    strRep = strRep + "\nstate: " + str(self.state)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>state</name>"
    strRep = strRep + "\n\t<type>ProcessorState</type>"
    if(self.state!=None):
      strRep = strRep + "\n\t<value>"+self.state.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Root( CoupledDEVS ):
  def __init__(self, ria, rib, rszl, rszh, rsize, name):
    CoupledDEVS.__init__(self, name)
    self.ria = ria
    self.rib = rib
    self.rszl = rszl
    self.rszh = rszh
    self.rsize = rsize
    self.name = name
    self.r_out = self.addOutPort("r_out")
    self.r_discard = self.addOutPort("r_discard")
    self.g1 = Generator(ia = self.ria, ib = self.rib, szl = self.rszl, szh = self.rszh, name='g1')
    self.addSubModel( self.g1 )
    self.p1 = Processor(qSize = self.rsize, name='p1')
    self.addSubModel( self.p1 )
    self.p2 = Processor(qSize = self.rsize, name='p2')
    self.addSubModel( self.p2 )
    self.p3 = Processor(qSize = self.rsize, name='p3')
    self.addSubModel( self.p3 )  
    self.connectPorts( self.g1.g_out, self.p1.p_in)  
    self.connectPorts( self.p1.p_out, self.r_out)  
    self.connectPorts( self.p1.p_discard, self.p2.p_in)  
    self.connectPorts( self.p2.p_out, self.r_out)  
    self.connectPorts( self.p2.p_discard, self.p3.p_in)  
    self.connectPorts( self.p3.p_out, self.r_out)  
    self.connectPorts( self.p3.p_discard, self.r_discard)

  def __str__(self):
    strRep = ''
    strRep = strRep + "\ng1: " + str(self.g1)
    strRep = strRep + "\np1: " + str(self.p1)
    strRep = strRep + "\np2: " + str(self.p2)
    strRep = strRep + "\np3: " + str(self.p3)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>g1</name>"
    strRep = strRep + "\n\t<type>Generator</type>"
    if(self.g1!=None):
      strRep = strRep + "\n\t<value>"+self.g1.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p1</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p1!=None):
      strRep = strRep + "\n\t<value>"+self.p1.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p2</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p2!=None):
      strRep = strRep + "\n\t<value>"+self.p2.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p3</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p3!=None):
      strRep = strRep + "\n\t<value>"+self.p3.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Job( object ):
  idNumber = 0
  def __init__(self, szl, szh):
    self.szl = szl
    self.szh = szh
    self.id = Job.idNumber + 1
    Job.idNumber = Job.idNumber + 1
    self.size = randint(self.szl, self.szh)

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nid: " + str(self.id)
    strRep = strRep + "\nsize: " + str(self.size)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>id</name>"
    strRep = strRep + "\n\t<type>Integer</type>"
    strRep = strRep + "\n\t<value>"+str(self.id)+"</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>size</name>"
    strRep = strRep + "\n\t<type>Integer</type>"
    strRep = strRep + "\n\t<value>"+str(self.size)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class RootExperiment( object ):
  def __init__(self):
    self.rootModel = Root(ria = 2, rib = 5, rszl = 3, rszh = 10, rsize = 2, name = "RootExperiment")
    self.sim = Simulator(model = self.rootModel)




def terminate_whenEndTimeReached(model, clock, end_time=1000):
  if clock >= end_time:
    return True
  else:
    return False
if __name__ == '__main__':
  experiment = RootExperiment()
  experiment.sim.simulate(termination_condition=terminate_whenEndTimeReached, verbose=True, trace=True)