package queue
  import externalfunctions.*;
  import devs.*;
  class Job
    extends DevsEvent;
    parameter Integer szl;
    parameter Integer szh;
    Integer id=0;
    Integer size=randint(szl,szh);
  end Job;
  class JobPort
    extends DevsPort;
    Job msg;
  end JobPort;
  class GeneratorState
    extends DevsState;
    Generator.SeqStates seqState(start=Generator.SeqStates.idle);
  end GeneratorState;
  class Generator "A Generator"
    extends AtomicDEVS;
    parameter Integer ia;
    parameter Integer ib;
    parameter Integer szl;
    parameter Integer szh;
    parameter String name;
    type SeqStates
      EnumType value;
      parameter StringType quantity="";
      parameter EnumType min=idle, max=generating;
      parameter EnumType start=idle;
      parameter BooleanType fixed=false;
      parameter BooleanType enable=true;
      constant EnumType idle;
      constant EnumType generating;
    end SeqStates;
    GeneratorState state();
    input JobPort inport;
    output JobPort out;
    function intTransition
    algorithm 
      if state.seqState == SeqStates.idle then
        state.seqState := SeqStates.generating;
      else
        state.seqState := SeqStates.idle;
      end if;
    end intTransition;
    function timeAdvance
      output Integer timespan;
    algorithm 
      timespan := randint(ia,ib);
    end timeAdvance;
    function outputFnc
    algorithm 
      poke(out,Job(szl,szh));
    end outputFnc;
  end Generator;
  class ProcessorState
    extends DevsState;
    parameter Integer n(start=0);
    list queue();
    Job currentJob;
    Real timeElapsed(start=0.0);
    Processor.SeqStates seqState(start=Processor.SeqStates.idle);
  end ProcessorState;
  class Processor "A Processor Definition"
    extends AtomicDEVS;
    parameter Integer qn=0;
    parameter String name;
    constant Integer nonf(start=1000000);
    type SeqStates
      EnumType value;
      parameter StringType quantity="";
      parameter EnumType min=idle, max=discarding;
      parameter EnumType start=idle;
      parameter BooleanType fixed=false;
      parameter BooleanType enable=true;
      constant EnumType idle;
      constant EnumType busy;
      constant EnumType discarding;
    end SeqStates;
    ProcessorState state(n=qn);
    input JobPort inport;
    output JobPort out;
    output JobPort discard;
    function extTransition
      Job msg=null;
    algorithm 
      msg := peek(inport);
      if msg <> null then
        state.queue.append(msg);
        msg := null;
      end if;
      if state.seqState == SeqStates.idle then
        state.currentJob := state.queue[0];
        state.queue.remove(state.queue[0]);
        state.seqState := SeqStates.busy;
      else
        state.timeElapsed := state.timeElapsed+elapsed;
        if len(state.queue) > qn then
          state.seqState := SeqStates.discarding;
        else
          state.seqState := SeqStates.busy;
        end if;
      end if;
    end extTransition;
    function intTransition
    algorithm 
      if state.seqState == SeqStates.busy then
        state.timeElapsed := 0.0;
        if len(state.queue) < 1 then
          state.currentJob := null;
          state.seqState := SeqStates.idle;
        else
          state.currentJob := state.queue[0];
          state.queue.remove(state.queue[0]);
          state.seqState := SeqStates.busy;
        end if;
      end if;
      if state.seqState == SeqStates.discarding then
        state.seqState := SeqStates.busy;
      end if;
    end intTransition;
    function timeAdvance
      output Integer timespan;
    algorithm 
      if state.seqState == SeqStates.idle then
        timespan := INFINITY;
      else
if state.seqState == SeqStates.busy then
        timespan := state.currentJob.size-state.timeElapsed;
      else
        timespan := 0;
      end if;
    end timeAdvance;
    function outputFnc
      Job msg=null;
    algorithm 
      if state.seqState == SeqStates.busy then
        poke(out,state.currentJob);
      end if;
      if state.seqState == SeqStates.discarding then
        msg := state.queue.pop();
        poke(discard,msg);
      end if;
    end outputFnc;
  end Processor;
  class Root
    extends CoupledDEVS;
    parameter Integer pqs, pia, pib, psa, psb;
    parameter String name;
    output JobPort out;
    output JobPort discard;
    Generator generator(ia=pia, ib=pib, szl=psa, szh=psb);
    Processor p1(qn=pqs), p2(qn=pqs), p3(qn=pqs);
  equation 
    connect(generator.out, p1.inport);
    connect(p1.out, out);
    connect(p1.discard, p2.inport);
    connect(p2.out, out);
    connect(p2.discard, p3.inport);
    connect(p3.out, out);
    connect(p3.discard, discard);
  end Root;
end queue;

