import os,sys

from string import split

from XMLTools.XMLParser import XMLParser

from GraphRepr.Graph import Graph
from GraphRepr.Node import Node

from GraphRepr.Types.SimpleType.SimpleType import SimpleType
from GraphRepr.Types.SimpleType.Int import Int
from GraphRepr.Types.SimpleType.Bool import Bool
from GraphRepr.Types.SimpleType.Float import Float
from GraphRepr.Types.SimpleType.String import String

from GraphRepr.Types.CollectionType.CollectionType import CollectionType
from GraphRepr.Types.CollectionType.List import List

from GraphRepr.Types.MappingType.MappingType import MappingType
from GraphRepr.Types.MappingType.SymbolTable import SymbolTable

from GraphManager.AGLLoader.Error.AGLLoaderException import LoadErrorException

from string import replace
from string import split


class AGLLoader:

    def __init__(self, fileMapping, catalogPath):
        self.mapping=fileMapping
        self.catalogPath=catalogPath
        self.nodeToDo={}
        self.edgeToDo=[]
        self.loaded={}

    def getLocalId(self, fullPath, depth):
        localPath=fullPath[depth+1]
        for elem in fullPath[depth+2:]:
            localPath+="."+elem
        return localPath

    def __getGraph(self, tag, globalId, parentGraph=None):

        ID=tag.attrs["id"]
        TYPE=tag.attrs["type"]

        if tag.name=="graph":
            graph=None
            if parentGraph!=None:
                return self.load(globalId+"."+ID)
            else:
                graph = Graph(ID, type=TYPE)
            for element in tag.tags:
                if element.name=="edge":
                    FROM=element.attrs["from"]
                    TO=element.attrs["to"]
                    FROMORD=element.attrs["fromOrd"]
                    TOORD=element.attrs["toOrd"]
                    graph.connect(FROM, TO, int(FROMORD), int(TOORD))
                else:
                    graph.addElement(self.__getGraph(element, globalId, graph))
            return graph
        elif tag.name=="node":
            if ID==TYPE:
                return Node(ID)
            else:
                return Node(ID,type=TYPE)
        elif tag.name=="string":
            VAL=tag.attrs["value"]
            VAL=replace(VAL, '\\n', '\n')
            VAL=replace(VAL, '\\t', '\t')
            return String(VAL,ID)
        elif tag.name=="int":
            VAL=tag.attrs["value"]
            return Int(int(VAL), ID)
        elif tag.name=="float":
            VAL=tag.attrs["value"]
            return Float(float(VAL), ID)
        elif tag.name=="bool":
            VAL=tag.attrs["value"]
            if VAL=="true" or VAL=="false":
                return Bool(VAL, ID)
            else:
                return Bool(int(VAL), ID)
        elif tag.name=="list":
            lst = List(ID)
            for elem in tag.tags:
                lst.addElement(self.__getGraph(elem, globalId))
            return lst
        elif tag.name=="symb":
            symb = SymbolTable(ID)
            for mapping in tag.tags:
                key = mapping.attrs["key"]
                if len(mapping.tags)==1:
                    symb[key]=self.__getGraph(mapping.tags[0], globalId)
                else:
                    symb[key]=None
            return symb

    def load(self, graphId):
        if self.loaded.has_key(graphId):
            return self.loaded[graphId]
        tree = XMLParser(self.mapping[graphId], self.catalogPath).parseXML()
        graph=self.__getGraph(tree.root.tags[0], graphId)
        self.loaded[graphId]=graph
        return graph


