from GraphRepr.Graph import Graph
from GraphRepr.Node import Node
from GraphManager.AGLGenerator.AGLGenerator import AGLGenerator
from GraphManager.AGLLoader.AGLLoader import AGLLoader

from GraphRepr.Types.SimpleType.SimpleType import SimpleType
from GraphRepr.Types.SimpleType.Int import Int
from GraphRepr.Types.SimpleType.Bool import Bool
from GraphRepr.Types.SimpleType.String import String
from GraphRepr.Types.SimpleType.Float import Float

from GraphRepr.Types.CollectionType.CollectionType import CollectionType
from GraphRepr.Types.CollectionType.List import List

from GraphRepr.Types.MappingType.MappingType import MappingType
from GraphRepr.Types.MappingType.SymbolTable import SymbolTable

import unittest
import os,sys
import random

class AGLGeneratorSuccess(unittest.TestCase):

    def g(self):
        self.graph=Graph("BlackWhite")
        self.graph.addElement(Node("White"))
        self.graph.addElement(Node("Black"))
        self.idHost=0

    def gnb(self):
        self.graph.addElement(Node(str(self.idHost), type="Black"))
        self.idHost+=1

    def gnw(self):
        self.graph.addElement(Node(str(self.idHost), type="White"))
        self.idHost+=1

    def gc(self, a, b, aOrd, bOrd):
        self.graph.connect(a,b,aOrd,bOrd)

    #creates a random graph having between minNodes and maxNodes nodes
    #and between minConn and maxConn connections.
    def randomGraph(self, minNodes, maxNodes, minConn, maxConn, blackQty, noSelfLoop=0):
        nodes=random.randint(minNodes, maxNodes)
        conns=random.randint(minConn, maxConn)

        self.g()

        for i in xrange(nodes):
            if random.random()<blackQty:
                self.gnb()
            else:
                self.gnw()
        for i in xrange(conns):
            nodeA=random.randint(0, nodes-1)
            nodeB=random.randint(0, nodes-1)

            while 1:
                if (self.graph.getElement(str(nodeA)).outC().hasElement("BlackWhite."+str(nodeB))) or\
                   (noSelfLoop and nodeA==nodeB):
                    nodeA=random.randint(0, nodes-1)
                    nodeB=random.randint(0, nodes-1)
                else:
                    break
            self.gc(str(nodeA), str(nodeB), random.randint(0, nodes-1), random.randint(0, nodes-1))
        return [nodes, conns]



    def createCatalog(self):

        pathname = os.path.dirname(sys.argv[0])
        rootDir = os.path.abspath(pathname)

        sep=os.sep

        from string import replace

        rootDir = replace(rootDir, sep+"GraphManager"+sep+"Test", "")
        self.catalog=rootDir+sep+"XMLTools"+sep+"catalog"+sep+"catalog.txt"

    def createMapping(self):

        self.mapping={}

        pathname = os.path.dirname(sys.argv[0])
        rootDir = os.path.abspath(pathname)

        sep=os.sep

        testDir = rootDir+sep+"files"+sep

        self.mapping["root"]=testDir+"root.agl"
        self.mapping["root.A"]=testDir+"rootA.agl"
        self.mapping["root.B"]=testDir+"rootB.agl"
        self.mapping["root.A.C"]=testDir+"rootAC.agl"
        self.mapping["root.B.D"]=testDir+"rootBD.agl"
        self.mapping["root.A.C.E"]=testDir+"rootACE.agl"
        self.mapping["root.SimpleType"]=testDir+"simpletype.agl"

    def createBasicGraph(self):

        self.root=Graph("root")

        self.root.addElement(Graph("A"))
        self.root.addElement(Graph("B"))

        self.root.addElement(Graph("C"),"A")
        self.root.addElement(Graph("D"),"B")

        self.root.addElement(Graph("E"),"A.C")
        self.root.addElement(Node("F"), "B.D")

        self.root.connect("A.C", "B.D",0,0)
        self.root.connect("A", "A.C.E",10,0)
        self.root.connect("A", "A", 1,0)
        self.root.connect("A", "A.C", 2,0)
        self.root.connect("A", "B.D",3,10)

    def addTypes(self):

        self.root.addElement(Int(5, "IntA"))
        self.root.addElement(String("marc", "StringA"))
        self.root.addElement(Float(4.5, "FloatA"))
        self.root.addElement(Bool("false", "BoolA"))

        self.root.addElement(Int(5, "IntB"), "B.D")

        lst=List("ListA")
        lst.addElement(Int(5))
        lst.addElement(String("yes!"))
        lst.addElement(Float(4.5))
        lst.addElement(Bool("false"))

        self.root.addElement(lst)

        symb=SymbolTable("SymbA")
        symb["1"]=Int(10)
        symb["2"]=String("no!")
        symb["3"]=Float(123.23)
        symb["4"]=Bool("true")

        self.root.addElement(symb, "A.C.E")


    def testBasic(self):
        """ testing AGLGenerator combined with AGLLoader:basic
        """
        self.createBasicGraph()
        self.createMapping()
        self.createCatalog()

        gen=AGLGenerator(self.root, self.mapping)
        gen.genCode()

        loader=AGLLoader(self.mapping, self.catalog)

        root=loader.load("root")
        self.failIf(not root.equals(self.root))

    def testWithTypes(self):
        """ Testing with hardcoded types
        """

        self.createBasicGraph()
        self.createMapping()
        self.createCatalog()
        self.addTypes()

        gen=AGLGenerator(self.root, self.mapping)
        gen.genCode()

        loader=AGLLoader(self.mapping, self.catalog)

        root=loader.load("root")

        self.failIf(not root.equals(self.root))

    def testRandom(self):
        """Testing with a random graph (no hierarchy)
        """
        self.createCatalog()

        pathname = os.path.dirname(sys.argv[0])
        rootDir = os.path.abspath(pathname)
        mapp={}

        sep=os.sep

        testDir = rootDir+sep+"files"+sep

        mapp["BlackWhite"]=testDir+"BlackWhite.agl"

        self.randomGraph(100,200,75,150, 0.5)

        gen=AGLGenerator(self.graph, mapp)
        gen.genCode()

        loader=AGLLoader(mapp, self.catalog)

        graph=loader.load("BlackWhite")
        self.failIf(not graph.equals(self.graph))





if __name__ == "__main__":
    unittest.main()