class GraphException(Exception):
    """Base class for exception in this module"""
    pass

class TypeCompatibilityException(GraphException):

    def __init__(self, typeA, typeB):
        self.typeA=typeA
        self.typeB=typeB

    def __str__(self):
        return "\nIncompatible types: \n" +\
                str(self.typeA) + " != " + str(self.typeB)

class TypeInvalidityException(GraphException):

    def __init__(self, wrongType, expectedTypes):
        self.wrongType=wrongType
        self.expectedTypes=expectedTypes

    def __str__(self):
        answer = "\nInvalid type: " + str(self.wrongType) + "\n"+\
                 "Expected types: "

        for val in self.expectedTypes:
            answer += str(val) + " "

        return answer

class NonUniqueIdException(GraphException):

    def __init__(self, ID):
        self.ID=ID

    def __str__(self):
        return "\nID " + str(self.ID) + " is not unique\n"

class InvalidIdException(GraphException):

    def __init__(self, ID):
        self.ID=ID

    def __str__(self):
        return "\nID " + str(self.ID) + " does not exists in current graph\n"

class AlreadyConnectedException(GraphException):

    def __init__(self, elemA, elemB):
        self.elemA=elemA
        self.elemB=elemB

    def __str__(self):
        return "\n"+self.elemA+" is already connected to "+self.elemB+"\n"

class InvalidPathException(GraphException):

    def __init__(self, ID, msg=""):
        self.ID=ID
        self.msg=msg

    def __str__(self):
        return "\nID " + str(self.ID) + " does not exists in current graph.\n" + str(self.msg)

class NotConnectedException(GraphException):

    def __init__(self, elemA, elemB):
        self.elemA=elemA
        self.elemB=elemB

    def __str__(self):
        return "\n"+self.elemA+" is not connected to "+self.elemB+"\n"

class InvalidElementException(GraphException):

    def __init__(self, ID, msg=""):
        self.ID=ID
        self.msg=msg

    def __str__(self):
        return "\nID " + str(self.ID) + " is invalid.\n" + str(self.msg)