from GraphRepr.GraphUtils.ConnIterator import ConnIterator
from GraphRepr.GraphUtils.ElemIterator import ElemIterator

from GraphRepr.GraphUtils.GraphElemCollection import GraphElemCollection


from GraphRepr.GraphUtils.Error.IteratorException\
                            import NoMoreElementException,\
                                   TypeInvalidityException,\
                                   InvalidKeyException

from GraphRepr.Graph import Graph
from GraphRepr.Node import Node


import unittest

class IteratorSuccess(unittest.TestCase):
    """class testing Iterators for functionnality
    """

    listIteratorSampleSuccess= ( [4,5,6,7,8,9],
                                 ['a', 'b', 'c'])

    elemIteratorSampleSuccess= [ [{'a':[1,2,3], 'b':[4,5,6], 'c':[], 'd':[7]},
                                 [1,2,3,4,5,6,7],

                                 [[['a'], [1,2,3]],
                                  [['b'], [4,5,6]],
                                  [['c'], []],
                                  [['f'], []]
                                 ],


                               ]]




    def testElemIteration(self):
        """A elemIterator shall iterate properly
        """
        for AElem in self.elemIteratorSampleSuccess:
            length = len(AElem[1])
            listIter = ElemIterator(AElem[0])
            newlist=[]
            while(not listIter.isDone()):
                currElement=listIter.currentElement()
                self.failIf(currElement not in AElem[1],
                            "Iterating an item not in the original elem:"+\
                                                              str(currElement))
                self.failIf(currElement in newlist,
                            "Iterating an item more than once!")
                newlist.append(listIter.currentElement())
                listIter.next()

            self.failIf(length != len(newlist),
                        "Iterated more elements than in the original elem "+\
                        str(length) +" != "+ str(len(newlist)))

            for testElem in AElem[2]:


                listIter = ElemIterator(AElem[0], testElem[0])
                length = len(testElem[1])
                newlist=[]

                while(not listIter.isDone()):
                    currElement=listIter.currentElement()
                    self.failIf(currElement not in testElem[1],
                                "Iterating an item not in the original elem:"+\
                                                                  str(currElement))
                    self.failIf(currElement in newlist,
                                "Iterating an item more than once!")
                    newlist.append(listIter.currentElement())
                    listIter.next()

                self.failIf(length != len(newlist),
                            "Iterated more elements than in the original elem "+\
                            str(length) +" != "+ str(len(newlist)))

    def testElemFirst(self):
        """A elemIterator first() method shall behave correctly
        """
        for AElem in self.elemIteratorSampleSuccess:
            length = len(AElem[1])
            elemIter = ElemIterator(AElem[0])
            newlist=[]
            for i in range(int(length/2)):
                currElement=elemIter.currentElement()
                self.failIf(currElement not in AElem[1],
                            "Iterating an item not in the original elem:"+\
                                                              str(currElement))
                self.failIf(currElement in newlist,
                            "Iterating an item more than once!")
                newlist.append(elemIter.currentElement())
                elemIter.next()

            elemIter.first()
            newlist=[]

            while(not elemIter.isDone()):
                currElement=elemIter.currentElement()
                self.failIf(currElement not in AElem[1],
                            "Iterating an item not in the original elem:"+\
                                                              str(currElement))
                self.failIf(currElement in newlist,
                            "Iterating an item more than once!")
                newlist.append(elemIter.currentElement())
                elemIter.next()

            self.failIf(length != len(newlist),
                        "Iterated more elements than in the original elem "+\
                        str(length) +" != "+ str(len(newlist)))

class IteratorFailure(unittest.TestCase):
    """Class testing that appropriate errors are raised
       on bad inputs
    """

    elemTypeErrorSample = [("asd",None), ({},5.6), (6.7,None), ([],None)]
    listTypeErrorSample = [{}, 4, 3.4]

    elemKeyErrorSample = [({}, "sda"), ({'a':5}, "df")]

    listIteratorSampleSuccess= ( [4,5,6,7,8,9],
                                 ['a', 'b', 'c'])

    elemIteratorSampleSuccess= [ [{'a':[1,2,3], 'b':[4,5,6], 'c':[], 'd':[7]},
                                 [1,2,3,4,5,6,7]]]

    def testElemNoMoreElementSample(self):
        """A elemIterator shall raise a NoMoreElement after iteration
        """

        for testSample in self.elemIteratorSampleSuccess:
            elemIter=ElemIterator(testSample[0])
            while(not elemIter.isDone()):
                elemIter.next()

            self.assertRaises(NoMoreElementException, elemIter.next)
            self.assertRaises(NoMoreElementException, elemIter.currentElement)

class CollectionSuccess(unittest.TestCase):

    elemSampleSuccess = [ [ {'a1':1,'a2':2,'a3':3,'a4':4,'a5':5,'a6':6,'a7':7},
                            {'a':[1,2,3], 'b':[4,5], 'c':[], 'd':[6,7]}
                          ]
                        ]
    TOP = Graph("TOP")
    ER = Graph("ER",parent=TOP)

    ER.addElement(Node("Entity"))

    FSA = Graph("FSA", parent=TOP, type="ER")

    print FSA.getType().getGlobalId(), "*******"

    listSampleSuccess = [ [[Node("State", type="Entity"),
                            Node("Place", type="Entity"),
                            Node("Transition", type="Entity"),
                            Node("Relationship")
                           ],
                           ["Entity", "Relationship"]
                          ]
                        ]

    def testGraphElemCollection(self):
        """ Testing GraphElemCollection for success
        """

        for sample in self.elemSampleSuccess:

            myElem = GraphElemCollection(sample[0], sample[1])

            for elem in sample[0].keys():
                self.failIf(not myElem.hasElement(elem))
                collectionElem = myElem.getElement(elem)
                self.failIf(sample[0][elem] != collectionElem,
                            "Elements retrieved from the collection are not"+
                            " correct")

            for elem in sample[1].keys():
                collectionElem = myElem.getElementsOfType(elem)
                self.failIf(sample[1][elem] != collectionElem,
                            "Elements retrieved from the collection are not"+
                            " correct")










if __name__ == "__main__":
    unittest.main()







