#Node class.
#Must be member of an High graph
#Author: Marc Provost
#Version: 1.1
#Date: August 7th, 2003

# **** NOTE : Since python does not support protected attributes,
# GraphConnent Implementation had to be copy / pasted instead of
# being inherited...  ******

from GraphRepr.GraphElement import GraphElement

from GraphRepr.GraphUtils.GraphConnCollection import GraphConnCollection

import weakref

class Node(GraphElement):

    def __init__(self, ID, parent=None, type=None):

        self.__id=ID

        if type==None:
            self.__typeId=ID
        else:
            self.__typeId=type
        self.__type=None

        self.__parent=None
        self.__root=None
        self.__globalId=str(ID)
        self.__depth=0

        if parent != None:
            parent.addElement(self)

    #the following methods are meant to be protected
    def setRoot(self, root):
        if root==None:
            self.__root=root
        else:
            self.__root=weakref.ref(root)

    def setGlobalId(self, path):
        self.__globalId=path

    def setDepth(self, depth):
        self.__depth=depth

    def setParent(self, parent):
        if parent==None:
            self.__parent=parent
        else:
            self.__parent=weakref.ref(parent)

    def createConnInfo(self):
        conns=self.getRoot().getConnInfo()
        conns[self.getGlobalId()]=[{}, {}]

    #the following methods are accessible members of the interface
    def setType(self, newTypeId):
        self.__typeId=newTypeId
        self.__type=None

    def getTypeLocalId(self):
        if self.__typeId==None:
            return self.getLocalId()
        return self.__typeId

    def getType(self):
        if self.__type!=None:
            return self.__type()
        if self.getParent()==None:
            return self
        typeGraphId=self.getParent().getType().getGlobalId()
        TYPE=self.getRoot().getElementGlobal(typeGraphId+"."+self.__typeId)
        self.__type=weakref.ref(TYPE)
        return self.__type()

    def getTypeGlobalId(self):
        if self.getParent()==None:
            return self.__id
        return self.getParent().getType().getGlobalId()+"."+self.__typeId

    def getDepth(self):
        return self.__depth

    def getGlobalId(self, relativeToDepth=0):
        if not relativeToDepth:
            return self.__globalId
        ind=0
        i=0
        for j in range(len(self.__globalId)):
            if self.__globalId[j]==".":
                ind=j
                i+=1
            if i==relativeToDepth:
                break
        return self.__globalId[(ind+1):]

    def getRoot(self):
        if self.__root==None:
            return self.__root
        else:
            return self.__root()

    def getParent(self):
        if self.__parent==None:
            return self.__parent
        else:
            return self.__parent()

    def getLocalId(self, relativeToParent=0):
        if not relativeToParent:
            return self.__id
        if relativeToParent >= self.getDepth():
            return self.__globalId
        answer=self.__id
        curParent=self
        for i in range(relativeToParent):
            curParent=curParent.getParent()
            answer = curParent.getLocalId()+"."+answer
        return answer

    def hasElementAnyDepth(self, globalId):
        conns=self.getRoot().getConnInfo()
        return conns.has_key(globalId)

    def outC(self):
        conns=self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        return GraphConnCollection(conns[globalId][1])

    def inC(self):
        conns=self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        return GraphConnCollection(conns[globalId][0])

    def disconnectAllRecurse(self):
        conns = self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        for fromKey in conns[globalId][0].keys():
            fromElem=conns[globalId][0][fromKey][0]
            fromElemId=fromElem.getGlobalId()
            del conns[fromElemId][1][globalId]
            if globalId==fromElemId:
                del conns[fromElemId][0][globalId]

        for toKey in conns[globalId][1].keys():
            toElem=conns[globalId][1][toKey][0]
            toElemId=toElem.getGlobalId()
            del conns[toElemId][0][globalId]
        del conns[globalId]


    def __str__(self):
        return str(self.__id) + " :: " + str(self.getType().getLocalId())

    def hasLocalId(self, ID):
        raise NotImplementedError("Method not implemented for Node")

    def getElement(self, elem_ID):
        raise NotImplementedError("Method not implemented for Node")

    def getElements(self):
        raise NotImplementedError("Method not implemented for Node")

    def addElement(self, elem, sources=[], targets=[]):
        raise NotImplementedError("Method not implemented for Node")

    def removeElement(self, ID):
        raise NotImplementedError("Method not implemented for Node")

    def connect(self, id_A, id_B):
        raise NotImplementedError("Method not implemented for Node")

    def disconnect(self, id_A, id_B):
        raise NotImplementedError("Method not implemented for Node")

    def getElemInC(self, elem_ID):
        raise NotImplementedError("Method not implemented for Node")

    def getElemOutC(self, elem_ID):
        raise NotImplementedError("Method not implemented for Node")

    # enter action perfomed by a visitor on this element
    def enter(self, visitor):
        visitor.enterNode(self)

    # exit action perfomed by a visitor on this element
    def exit(self, visitor):
        visitor.exitNode(self)

    def equals(self, elem):
        if not isinstance(elem,Node):
            return 0

        if self.getLocalId() != elem.getLocalId():
            return 0

        if not (self.getType()==self and elem.getType()==elem):
            if not self.getType().equals(elem.getType()):
                return 0

        return 1

    def clone(self):
        return Node(self.getLocalId(),
                    parent=None,
                    type=self.getTypeLocalId())

    def __repr__(self):
        return str(self.__globalId)+"::"+str(self.getTypeLocalId())

    def __del__(self):
        #print "del node", self
        del self.__id
        del self.__type
        del self.__parent
        del self.__depth
        del self.__globalId
        del self.__root
