#from GraphRepr.Graph import Graph
#from GraphRepr.Node import Node

from GraphRepr.Graph import Graph
from GraphRepr.Node import Node

#from GraphRepr.Types.SimpleType.Int import Int
#from GraphRepr.Types.SimpleType.Bool import Bool
#from GraphRepr.Types.SimpleType.String import String
#from GraphRepr.Types.SimpleType.Float import Float

from GraphRepr.Error.GraphException import\
                                InvalidPathException,\
                                NonUniqueIdException,\
                                InvalidIdException,\
                                NotConnectedException,\
                                AlreadyConnectedException

import unittest
import sys
import random

class GraphElementSuccess(unittest.TestCase):
    """Class testing GraphElement common implementation
       for Nodes and Graphs
    """

    def testGraphElementSuccess(self):
        """Testing GraphElement for success
        """

        TOP = Graph("TOP")
        ER = Graph("ER", parent=TOP)
        FSA = Graph("FSA", parent=TOP, type="ER")
        entity = Node("Entity",parent=ER)
        state = Node("State",parent=FSA, type="Entity")

        graphElementSampleSuccess= [[TOP, "TOP", "TOP", TOP, None, TOP],
                               [ER, "ER", "TOP.ER", TOP, TOP, ER],
                               [FSA,"FSA","TOP.FSA",TOP,TOP,ER],
                               [entity,"Entity","TOP.ER.Entity",TOP,ER,entity],
                               [state,"State","TOP.FSA.State",TOP,FSA,entity]
                              ]

        for sample in graphElementSampleSuccess:

            #verify that id is correct
            self.failIf(sample[0].getLocalId()!=sample[1],
                        "id do not correspond to returned value\n"+\
                        sample[0].getLocalId()+" != "+sample[1])

            #verify that the full pathId is correct
            self.failIf(sample[0].getGlobalId()!=sample[2],
                        "pathId do not correspond to returned value\n"+\
                        sample[0].getGlobalId()+" != "+sample[2])

            #verify that the root is correct
            self.failIf(sample[0].getRoot()!=sample[3],
                        "root do not correspond to returned value\n"+\
                        str(sample[0].getRoot())+" != "+str(sample[3]))

            #verify that the parent is correct
            self.failIf(sample[0].getParent()!=sample[4],
                        "parent do not correspond to returned value\n"+\
                        str(sample[0].getParent())+" != "+str(sample[4]))

            #verify that the type is correct
            self.failIf(sample[0].getParent()!=sample[4],
                        "type do not correspond to returned value\n"+\
                        str(sample[0].getType())+" != "+str(sample[5]))

    def testGraphElementSuccess2(self):
        """Testing GraphElement for success
        """

        TOP = Graph("TOP")

        ER = Graph("ER")
        TOP.addElement(ER)

        FSA = Graph("FSA", type="ER")
        TOP.addElement(FSA)

        entity = Node("Entity")
        ER.addElement(entity)

        state = Node("State",type="Entity")
        FSA.addElement(state)

        graphElementSampleSuccess= [[TOP, "TOP", "TOP", TOP, None, TOP],
                               [ER, "ER", "TOP.ER", TOP, TOP, ER],
                               [FSA,"FSA","TOP.FSA",TOP,TOP,ER],
                               [entity,"Entity","TOP.ER.Entity",TOP,ER,entity],
                               [state,"State","TOP.FSA.State",TOP,FSA,entity]
                              ]

        for sample in graphElementSampleSuccess:

            #verify that id is correct
            self.failIf(sample[0].getLocalId()!=sample[1],
                        "id do not correspond to returned value\n"+\
                        sample[0].getLocalId()+" != "+sample[1])

            #verify that the full pathId is correct
            self.failIf(sample[0].getGlobalId()!=sample[2],
                        "pathId do not correspond to returned value\n"+\
                        sample[0].getGlobalId()+" != "+sample[2])

            #verify that the root is correct
            self.failIf(sample[0].getRoot()!=sample[3],
                        "root do not correspond to returned value\n"+\
                        str(sample[0].getRoot())+" != "+str(sample[3]))

            #verify that the parent is correct
            self.failIf(sample[0].getParent()!=sample[4],
                        "parent do not correspond to returned value\n"+\
                        str(sample[0].getParent())+" != "+str(sample[4]))

            #verify that the type is correct
            self.failIf(sample[0].getParent()!=sample[4],
                        "type do not correspond to returned value\n"+\
                        str(sample[0].getType())+" != "+str(sample[5]))


class GraphElementFailure(unittest.TestCase):
    """Test that expected execptions are raised
       for GraphElement common implementation for
       Nodes and Graphs
    """
    pass


class NodeSuccess(unittest.TestCase):
    """Class testing Node specific functionality
       for success
    """
    #nothing specific yet (node is entirely described by GraphElement)
    pass

class NodeFailure(unittest.TestCase):
    """Class testing Node specific functionality
       for failure
    """
    #nothing specific yet (node is entirely described by GraphElement)
    pass

class GraphSuccess(unittest.TestCase):
    """Class testing Graph specific functionality
       for success
    """

    def g(self):
        self.graph=Graph("BlackWhite")
        self.graph.addElement(Node("White"))
        self.graph.addElement(Node("Black"))
        self.idHost=0

    def gnb(self):
        self.graph.addElement(Node(str(self.idHost), type="Black"))
        self.idHost+=1

    def gnw(self):
        self.graph.addElement(Node(str(self.idHost), type="White"))
        self.idHost+=1

    def gc(self, a, b, aOrd, bOrd):
        self.graph.connect(a,b,aOrd,bOrd)

    #creates a random graph having between minNodes and maxNodes nodes
    #and between minConn and maxConn connections.
    def randomGraph(self, minNodes, maxNodes, minConn, maxConn, blackQty, noSelfLoop=0):
        nodes=random.randint(minNodes, maxNodes)
        conns=random.randint(minConn, maxConn)

        self.g()

        for i in xrange(nodes):
            if random.random()<blackQty:
                self.gnb()
            else:
                self.gnw()
        for i in xrange(conns):
            nodeA=random.randint(0, nodes-1)
            nodeB=random.randint(0, nodes-1)

            while 1:
                if (self.graph.getElement(str(nodeA)).outC().hasElement("BlackWhite."+str(nodeB))) or\
                   (noSelfLoop and nodeA==nodeB):
                    nodeA=random.randint(0, nodes-1)
                    nodeB=random.randint(0, nodes-1)
                else:
                    break
            self.gc(str(nodeA), str(nodeB), random.randint(0, nodes-1), random.randint(0, nodes-1))
        return [nodes, conns]


    def checkConnAllThere(self, connMap):
        #checking that every connection is there..
        for conn in connMap:
            outIter=conn[0].outC()
            self.failIf(not outIter.hasElement(conn[1].getGlobalId()),
                        "Some out connections " + str(conn[1].getGlobalId()) +\
                        "are not present in " + str(conn[0].outC()))
            inIter=conn[1].inC()
            self.failIf(not inIter.hasElement(conn[0].getGlobalId()),
                        "Some in connections " + str(conn[0].getGlobalId()) +\
                        "are not present in " + str(conn[1].inC()))

    def checkNoMoreConn(self, connMap, root):
        #checking that there are no unexpected connections
        elemIter = root.getElements().iterate()
        while not elemIter.isDone():
            cur=elemIter.currentElement()
            if cur.__class__==Graph:
                self.checkNoMoreConn(connMap, cur)

            #checking out conn of current element
            outCIter=cur.outC().iterate()
            while not outCIter.isDone():
                curOut=outCIter.currentElement()
                check=0
                for conn in connMap:
                    if conn[1]==curOut:
                        check=1
                self.failIf(not check, "Found an unexpected connection")
                outCIter.next()

            #checking in conn of current element
            inCIter=cur.inC().iterate()
            while not inCIter.isDone():
                curIn=inCIter.currentElement()
                check=0
                for conn in connMap:
                    if conn[0]==curIn:
                        check=1
                self.failIf(not check, "Found an unexpected connection")
                inCIter.next()
            elemIter.next()

    def createGraphs(self):
        #first graph... top

        self.top=Graph("top")

        #top children

        self.g1 = Graph("g1", parent=self.top)
        self.g2 = Graph("g2", parent=self.top)
        self.g3 = Graph("g3", parent=self.top)
        self.n1 = Node("n1", parent=self.top)
        self.n2 = Node("n2", parent=self.top)

        #g1 children

        self.g11 = Graph("g11", parent=self.g1)
        self.n11 = Node("n11", parent=self.g1)

        #g11 children
        self.g111 = Graph("g111", parent=self.g11)

        #g2 children
        self.g21 = Graph("g21", parent=self.g2)
        self.g22 = Graph("g22", parent=self.g2)

        #g21 children
        self.n211 = Node("n211", parent=self.g21)

        #g3 children
        self.g31 = Graph("g31", parent=self.g3)

        #g31 children
        self.g311 = Graph("g311", parent=self.g31)

        #g311 children
        self.g3111 = Graph("g3111", parent=self.g311)

        self.connMap = [ (self.g1, self.g2),
                         (self.g21, self.g2),
                         (self.g3111, self.g22),
                         (self.n2, self.g311),
                         (self.g1, self.g3111),
                         (self.g1, self.n2),
                         (self.g1, self.n11),
                         (self.n211, self.n11),
                         (self.g3, self.n11),
                         (self.n11, self.g3),
                         (self.g3111, self.g11),
                         (self.g3111, self.g1),
                         (self.g311, self.g311),
                         (self.g3111, self.g3111)
                       ]


        for conn in self.connMap:
            self.top.connect(conn[0].getGlobalId(1), conn[1].getGlobalId(1))

        #second graph ... top2

        self.top2 = Graph("top2")

        self.g4 = Graph("g4", self.top2)
        self.g5 = Graph("g5", self.top2)

        self.n41 = Node("n41", parent=self.g4)
        self.g6 = Node("g5", parent=self.g4)

        self.connMap2 = [ (self.g4, self.n41),
                          (self.g6, self.g5)
                        ]

        for conn in self.connMap2:
            self.top2.connect(conn[0].getGlobalId(1), conn[1].getGlobalId(1))


    def testGraphSuccessConn(self):
        """Testing Graph for success (connections)
        """
        #connection mapping

        self.createGraphs()

        self.checkConnAllThere(self.connMap)
        self.checkNoMoreConn(self.connMap, self.top)

    def testGraphSuccessRemove(self):
        """Testing Graph for success (remove+add)
        """
        self.createGraphs()

        self.g31.removeElement(self.g311.getLocalId())

        newConnMap=[]
        oldConnMap=[]
        for conn in self.connMap:
            if(conn[0]!=self.g311 and conn[0]!=self.g3111):
                if(conn[1]!=self.g311 and conn[1]!=self.g3111):
                    newConnMap.append(conn)
                else:
                    oldConnMap.append(conn)

        self.checkConnAllThere(newConnMap)
        self.checkNoMoreConn(newConnMap, self.top)


    def testGraphSuccessPop(self):
        """Testing Graph for success (pop)
        """
        self.createGraphs()

        newElem= Graph("123")

        newElem.addElement(Node("1"))
        newElem.addElement(Node("2"))
        newElem.addElement(Graph("3"))
        newElem2 = newElem.clone()
        newElem2.connect("1","2")

        self.top.addElement(newElem)

        self.top.connect("123", "g1")
        self.top.connect("g1.g11", "123")
        self.top.connect(self.n211.getGlobalId(1), "123")
        self.top.connect("123.1", "g1")
        self.top.connect("123.1", "123.2")
        self.top.connect("g1", "123.1")

        newElem4=self.top.popElement("123.3")

        newElem3=self.top.popElement("123")

        newElem3.addElement(newElem4)

        self.failIf(not newElem3.equals(newElem2))
        self.checkConnAllThere(self.connMap)
        self.checkNoMoreConn(self.connMap, self.top)

    def testGraphSuccessAddElements(self):
        """Testing Graph for success (addElements)
        """

        self.createGraphs()

        newElem = Graph("123")
        newElem.addElement(Node("1"))
        newElem.addElement(Node("2"))
        newElem.addElement(Graph("3"))
        newElem.addElement(Node("4"), "3")
        newElem.connect("1", "2")
        newElem.connect("1", "3.4")
        newElem.connect("1", "1")
        newElem.connect("3.4", "3")
        newElem.connect("2", "3")
        newElem.connect("3", "2")

        newElem2 = newElem.clone()

        self.top.addElements(newElem)

        graphs=[]
        newGraphs=[]
        graphs.append(newElem2)

        while len(graphs)!=0:
            for graph in graphs:
                elemIter=graph.getElements().iterate()
                while not elemIter.isDone():
                    elem=elemIter.currentElement()
                    elemIter.next()
                    self.failIf(not self.top.hasLocalId(elem.getGlobalId(1)))
                    if elem.__class__==Graph:
                        newGraphs.append(elem)
            graphs=[]
            for graph in newGraphs:
                graphs.append(graph)
            newGraphs=[]

        topConns=self.top.getConnTuples()
        newConns=newElem2.getConnTuples()

        for newConn in newConns:
            check=0
            for topConn in topConns:
                if newConn[0].equals(topConn[0]):
                    if newConn[1].equals(topConn[1]):
                        check=1
            self.failIf(not check)

        self.failIf(newElem.equals(newElem2))

    def testGraphConnOrdRand(self):
        """Testing Graph connections ordering with random graph
        """

        self.randomGraph(100,200,100,200,0.5)

        first=1
        order=1

        gIT=self.graph.getElements().iterate()
        while not gIT.isDone():
            curEl=gIT.currentElement()
            gIT.next()

            first=1
            order=1

            outC=curEl.outC()
            outIT=outC.iterate()

            while not outIT.isDone():
                curOut=outIT.currentElement()
                outIT.next()
                newOrder=outC.getOrd(curOut.getGlobalId())
                if first:
                    first=0
                    order=newOrder
                    continue
                self.failIf(newOrder<order)
                order=newOrder

    def testGraphConnOrd(self):
        """Testing Graph connections with ordering
        """

        sampleElems=[Node("1"),Node("2"),Node("3"),Node("4")]
        sampleConns=[["1","1",0,0],["1","2",2,10],
                     ["1","3",1,0],["3","1",10,1],
                     ["3","2",3,5]]

        g=Graph("root")

        for elem in sampleElems:
            g.addElement(elem)
        for conn in sampleConns:
            g.connect(conn[0],
                      conn[1],
                      conn[2],
                      conn[3])



        localC=g.getLocalConn()
        connTup=g.getConnTuples()

        for conn1 in localC:
            for conn2 in sampleConns:
                if conn1[0]==conn2[0]:
                    if conn1[1]==conn2[1]:
                        self.failIf(conn1[2]!=conn2[2])
                        self.failIf(conn1[3]!=conn2[3])
                        break

        for conn1 in connTup:
            for conn2 in sampleConns:
                if conn1[0]==conn2[0]:
                    if conn1[1]==conn2[1]:
                        self.failIf(conn1[2]!=conn2[2])
                        self.failIf(conn1[3]!=conn2[3])
                        break

        first=1
        order=0
        for elem in sampleElems:
            el=g.getElement(elem.getLocalId())
            it=el.outC().iterate()
            while not it.isDone():
                curOut=it.currentElement()
                it.next()

                for conn in sampleConns:
                    if conn[0]==el.getLocalId():
                        if conn[1]==curOut.getLocalId():
                            newOrder=conn[2]

                if first:
                    first=0
                    order=newOrder
                    continue

                self.failIf(newOrder<order)
                order=newOrder

        first=1
        order=0
        for elem in sampleElems:
            el=g.getElement(elem.getLocalId())
            it=el.inC().iterate()
            while not it.isDone():
                curOut=it.currentElement()
                it.next()

                for conn in sampleConns:
                    if conn[0]==el.getLocalId():
                        if conn[1]==curOut.getLocalId():
                            newOrder=conn[3]

                if first:
                    first=0
                    order=newOrder
                    continue

                self.failIf(newOrder<order)
                order=newOrder



























if __name__ == "__main__":
    unittest.main()







