from GraphRepr.Types.SimpleType.Int import Int
from GraphRepr.Types.SimpleType.Bool import Bool
from GraphRepr.Types.SimpleType.String import String
from GraphRepr.Types.SimpleType.Float import Float

from GraphRepr.Types.SimpleType.Error.SimpleTypeException\
                                 import TypeInvalidityException,ValueException

import unittest

class SimpleTypeSuccess(unittest.TestCase):

    GetValueSample = ((Int(5), 5),
                         (Float(3.4), 3.4),
                         (Float(3), 3),
                         (String("roger"), "roger"),
                         (String("true"), "true"),
                         (Bool("true"), 1),
                         (Bool("false"), 0),
                         (Bool(1), 1),
                         (Bool(3), 0))

    SetValueSample = ((Int(5), 7, Int(7)),
                         (Float(3.4), 5, Float(5)),
                         (Float(3), 3, Float(3)),
                         (String("roger"), "marie", String("marie")),
                         (String("true"), "true", String("true")),
                         (Bool("true"), 1, Bool(1)),
                         (Bool("false"), "true", Bool("true")),
                         (Bool(1), "false", Bool("false")),
                         (Bool(3), 0, Bool(0)))

    def testGetValue(self):
        """A getValue shall return the expected answer
        """
        for aTuple in self.GetValueSample:
            self.failIf(aTuple[0].getValue()!=aTuple[1],
                        str(aTuple[0].getValue())+" != "+str(aTuple[1]))

    def testSetValue(self):
        """A setValue shall set a value appropriately
        """
        for aTuple in self.SetValueSample:
            aTuple[0].setValue(aTuple[1])
            self.failIf(aTuple[0].getValue()!=aTuple[2].getValue(),
                        str(aTuple[0].getValue())+" != "+\
                        str(aTuple[2].getValue()))

    def testEquals(self):
        """An equals shall behave as expected
        """
        for aTuple in self.SetValueSample:
            aTuple[0].setValue(aTuple[1])
            self.failIf(not aTuple[0].equals(aTuple[2]),
                        str(aTuple[0].getValue())+" != "+\
                        str(aTuple[2].getValue()))

    def testClone(self):
        """A clone shall behave as expected
        """
        for aTuple in self.SetValueSample:
            cloned=aTuple[0].clone()
            self.failIf(not cloned.equals(aTuple[0]),
                        "Cloned object does not equal the original!")
            if(aTuple[1] != aTuple[0].getValue()):
                cloned.setValue(aTuple[1])
                self.failIf(cloned.equals(aTuple[0]),
                        "Cloned objects are the same object")

    def testGetType(self):
        """getType shall function properly
        """
        pass


if __name__ == "__main__":
    unittest.main()
