from XMLTools.xmltool.xml.parsers.xmlproc import xmlval,catalog

from string import *

#level of debugging
_DEBUG=0

def printDebug(msg):
    if _DEBUG>=1:
        print msg

class XMLParserApplication(xmlval.Application):
    """ Class that implements the Application interface of xmlproc
        Receives messages (events) from the parser and handles them
    """

    class XMLTag:
        def __init__(self, name, attrs):
            self.attrs = attrs
            self.name = name
            self.data=""

            #Tags inside this tag
            self.tags = []

    class XML:

        def __init__(self):

            self.root=None
            self.rootName=None
            self.pubID=None
            self.sysID=None
            self.xmlver=None
            self.enc=None
            self.sddecl=None

            #for prettyprinter
            self.__output=None
            self.__indentLevel=0
            self.__indentSize="    "

    def __init__(self):
        self.xml=self.XML()
        self.tagStack=[]
        self.level=0

    def push(self, tag):
        self.tagStack.append(tag)

    def pop(self):
        answer = self.tagStack[-1]
        self.tagStack = self.tagStack[0:-1]
        return answer

    def set_locator(self,locator):
        pass

    def doc_start(self):
        pass

    def doc_end(self):
        pass

    def handle_comment(self,data):
        pass

    def handle_start_tag(self,name,attrs):
        if str(name)==self.xml.rootName:
            self.xml.root = self.XMLTag(name, attrs)
            self.tagStack.append(self.xml.root)
        else:
            tag = self.XMLTag(name, attrs)
            self.tagStack[self.level].tags.append(tag)
            self.level+=1
            self.push(tag)

        printDebug("Entered " + str(name))
        printDebug("With data " + str(attrs))

    def handle_end_tag(self,name):
        printDebug("Exited " + str(name))

        tag = self.pop()
        self.level-=1

    def handle_data(self,data,start,end):
        printDebug("Data: " + data[start:end])
        self.tagStack[self.level].data += data[start:end]

    def handle_ignorable_data(self,data,start,end):
        pass

    def handle_pi(self,target,data):
        pass

    def handle_doctype(self,root,pubID,sysID):
        self.xml.rootName=root
        self.xml.pubID=pubID
        self.xml.sysID=sysID

    def set_entity_info(self,xmlver,enc,sddecl):
        self.xml.xmlver=xmlver
        self.xml.enc=enc
        self.xml.sddecl=sddecl


class XMLParser:

    def __init__(self, xmlFile, catalogDir):
        #data file
        self.xmlFile=xmlFile

        #validating parser (via the XML dtd)
        self.parser=xmlval.XMLValidator()

        #setting the application
        self.application=XMLParserApplication()
        self.parser.set_application(self.application)

        #catalog file mapping href -> local files (i.e. for the dtd)
        self.catalogFile=catalogDir

        #creating a catalog object
        self.catalog=catalog.xmlproc_catalog(self.catalogFile,\
                                             catalog.CatParserFactory())

        #setting the parser to use the defined catalog
        self.parser.set_pubid_resolver(self.catalog)

    def parseXML(self):
        self.parser.parse_resource(self.xmlFile)
        return self.application.xml






