
package test
  import devs.*;
  import externalfunctions.*;
  import simulator.*;

  class GeneratorState
    Generator.SeqStates seqState(start=Generator.SeqStates.generating);
  end GeneratorState;

  class Generator
    extends AtomicDEVS;
    parameter String name;
    output DevsPort p_out;
    GeneratorState state();
    type SeqStates = enumeration(generating);

    function intTransition
    algorithm
      if( state.seqState == SeqStates.generating ) then
        // Use Modelica syntax
        state.seqState := SeqStates.generating;
      end if;
    end intTransition;

    function outputFnc
      DevsEvent evt = null;
    algorithm
      if( state.seqState == SeqStates.generating) then
        evt := Job(0.3);
        poke(p_out, evt);
      end if;
    end outputFnc;

    function timeAdvance
      output Integer timespan;
    algorithm
      if( state.seqState == SeqStates.generating) then
        timespan := 1;
      end if;
    end timeAdvance;

  end Generator;

  class ProcessorState
    Processor.SeqStates seqState(start=Processor.SeqStates.idle);
  end ProcessorState;

  class Processor
    extends AtomicDEVS;
    parameter String name;
    input DevsPort p_in;
    output DevsPort p_out;
    ProcessorState state();
    type SeqStates = enumeration(idle, state_2);

    function extTransition
    algorithm
      if( state.seqState == SeqStates.idle ) then
        // Use Modelica syntax
        state.seqState := SeqStates.state_2;
      event := peek(p_in);
      end if;
    end extTransition;


    function intTransition
    algorithm
      // Use Modelica syntax
      if( state.seqState == SeqStates.state_2 ) then
        // Use Modelica syntax
        state.seqState := SeqStates.idle;
      end if;
    end intTransition;

    function outputFnc
      DevsEvent evt = null;
    algorithm
      if( state.seqState == SeqStates.state_2) then
        poke(p_out, event);
      end if;
    end outputFnc;

    function timeAdvance
      output Integer timespan;
    algorithm
      if( state.seqState == SeqStates.idle) then
        timespan := INFINITY;
      end if;
      if( state.seqState == SeqStates.state_2) then
        timespan := event.jobSize;
      end if;
    end timeAdvance;

  end Processor;

  class Root
    extends CoupledDEVS;
    parameter String name;
    output DevsPort p_out;
    Generator ins_0();
    Processor ins_1(name="ins_1");
    Processor ins_2(name="ins_2");
    Processor ins_3(name="ins_3");
    Processor ins_4(name="ins_4");
    Processor ins_5(name="ins_5");
    Processor ins_6(name="ins_6");
    Processor ins_7(name="ins_7");
    Processor ins_8(name="ins_8");
    Processor ins_9(name="ins_9");
    Processor ins_10(name="ins_10");
    Processor ins_11(name="ins_11");
    Processor ins_12(name="ins_12");
    Processor ins_13(name="ins_13");
    Processor ins_14(name="ins_14");
    Processor ins_15(name="ins_15");
    Processor ins_16(name="ins_16");
    Processor ins_17(name="ins_17");
    Processor ins_18(name="ins_18");
    Processor ins_19(name="ins_19");
    Processor ins_20(name="ins_20");
    Processor ins_21(name="ins_21");
    Processor ins_22(name="ins_22");
    Processor ins_23(name="ins_23");
    Processor ins_24(name="ins_24");
    Processor ins_25(name="ins_25");
    Processor ins_26(name="ins_26");
    Processor ins_27(name="ins_27");
    Processor ins_28(name="ins_28");
    Processor ins_29(name="ins_29");
    Processor ins_30(name="ins_30");
    Processor ins_31(name="ins_31");
    Processor ins_32(name="ins_32");
    Processor ins_33(name="ins_33");
    Processor ins_34(name="ins_34");
    Processor ins_35(name="ins_35");
    Processor ins_36(name="ins_36");
    Processor ins_37(name="ins_37");
    Processor ins_38(name="ins_38");
    Processor ins_39(name="ins_39");
    Processor ins_40(name="ins_40");
    Processor ins_41(name="ins_41");
    Processor ins_42(name="ins_42");
    Processor ins_43(name="ins_43");
    Processor ins_44(name="ins_44");
    Processor ins_45(name="ins_45");
    Processor ins_46(name="ins_46");
    Processor ins_47(name="ins_47");
    Processor ins_48(name="ins_48");
    Processor ins_49(name="ins_49");
    Processor ins_50(name="ins_50");
  equation
    connect(ins_0.p_out, ins_1.p_in);
    connect(ins_1.p_out, ins_2.p_in);
    connect(ins_2.p_out, ins_3.p_in);
    connect(ins_3.p_out, ins_4.p_in);
    connect(ins_4.p_out, ins_5.p_in);
    connect(ins_5.p_out, ins_6.p_in);
    connect(ins_6.p_out, ins_7.p_in);
    connect(ins_7.p_out, ins_8.p_in);
    connect(ins_8.p_out, ins_9.p_in);
    connect(ins_9.p_out, ins_10.p_in);
    connect(ins_10.p_out, ins_11.p_in);
    connect(ins_11.p_out, ins_12.p_in);
    connect(ins_12.p_out, ins_13.p_in);
    connect(ins_13.p_out, ins_14.p_in);
    connect(ins_14.p_out, ins_15.p_in);
    connect(ins_15.p_out, ins_16.p_in);
    connect(ins_16.p_out, ins_17.p_in);
    connect(ins_17.p_out, ins_18.p_in);
    connect(ins_18.p_out, ins_19.p_in);
    connect(ins_19.p_out, ins_20.p_in);
    connect(ins_20.p_out, ins_21.p_in);
    connect(ins_21.p_out, ins_22.p_in);
    connect(ins_22.p_out, ins_23.p_in);
    connect(ins_23.p_out, ins_24.p_in);
    connect(ins_24.p_out, ins_25.p_in);
    connect(ins_25.p_out, ins_26.p_in);
    connect(ins_26.p_out, ins_27.p_in);
    connect(ins_27.p_out, ins_28.p_in);
    connect(ins_28.p_out, ins_29.p_in);
    connect(ins_29.p_out, ins_30.p_in);
    connect(ins_30.p_out, ins_31.p_in);
    connect(ins_31.p_out, ins_32.p_in);
    connect(ins_32.p_out, ins_33.p_in);
    connect(ins_33.p_out, ins_34.p_in);
    connect(ins_34.p_out, ins_35.p_in);
    connect(ins_35.p_out, ins_36.p_in);
    connect(ins_36.p_out, ins_37.p_in);
    connect(ins_37.p_out, ins_38.p_in);
    connect(ins_38.p_out, ins_39.p_in);
    connect(ins_39.p_out, ins_40.p_in);
    connect(ins_40.p_out, ins_41.p_in);
    connect(ins_41.p_out, ins_42.p_in);
    connect(ins_42.p_out, ins_43.p_in);
    connect(ins_43.p_out, ins_44.p_in);
    connect(ins_44.p_out, ins_45.p_in);
    connect(ins_45.p_out, ins_46.p_in);
    connect(ins_46.p_out, ins_47.p_in);
    connect(ins_47.p_out, ins_48.p_in);
    connect(ins_48.p_out, ins_49.p_in);
    connect(ins_49.p_out, ins_50.p_in);
     connect(ins_50.p_out, p_out);
  end Root;

  class Job
    extends DevsEvent;
    parameter Real jobSize;
  end Job;

  class Experiment
    extends DevsExperiment;
    Root rootModel(name="Experiment");
    Simulator sim(simModel=rootModel);
    Boolean verbose = False;
    Boolean trace = False;
    Real end_time = 1000;
  end Experiment;

end test;
