//type Voltage = Real(unit="V");

//class InnerPin
  //Real v1, v2;
//end InnerPin;

connector Pin "pin of an electric component"
	Real v "Potential at the pin";
	flow Real i "Current flowing into the pin";
end Pin;

partial model OnePort "Superclass of Components with two electrical pins p and n"
	Real v "Voltage drop between p and n";
	Real i "Current flowing from p to n";
	Pin p;
	Pin n;
equation
	v = p.v - n.v;
	0 = p.i + n.i;
	i = p.i;
end OnePort;

model Resistor "Ideal linear electrical resistor"
	extends OnePort;
	parameter Real r(unit="Ohm") "Resistance";
equation
	r*i = v;
end Resistor;

model TwoResistor
  Pin p, n;
  Resistor R1, R2;
equation
  connect(p, R1.p);
  connect(R1.n, R2.p);
  connect(R2.n, n);
end TwoResistor;

model Capacitor "Ideal electrical Capacitor"
	extends OnePort;
	parameter Real c(unit="F") "Capacitance";
equation
	c*der(v) = i;	
end Capacitor;

model VsourceAC "sin-wave voltage source"
	extends OnePort;
	parameter Real VA = 220 "Voltage Amplitude";
	parameter Real f (unit="Hz") = 50 "Frequency";
	constant Real PI = 3.14159265;
equation
	v = VA*sin(2*PI*f*time);
end VsourceAC;

model Ground "Ground"
	Pin p;
equation  
	p.v = 0;
end Ground;

model Circuit
	//Resistor R1(r=10);
	//Resistor R2(r=10);
	TwoResistor T(R1.r=10, R2.r=20);
	Capacitor C(c=1);
	VsourceAC AC;
	Ground G;
equation
	connect(AC.p, T.p);
	//connect(AC.p, R1.p);
	//AC.p.v=R1.p.v;
	//AC.p.i+R1.p.i=0;
	//connect(R1.n, R2.p);
	//R1.n.v=R2.p.v;
	//R1.n.i+R2.p.i=0;
	//connect(R2.n, C.p);
	//R2.n.v=C.p.v;
	//R2.n.i+C.p.i=0;
	connect(T.n, C.p);
	connect(C.n, AC.n);
	connect(AC.n, G.p);
	//C.n.v=AC.n.v;
	//AC.n.v=G.p.v;
	//C.n.i+AC.n.i+G.p.i=0;
end circuit;


