from whrandom import randint
from DEVS import *


class Job( object ):
  def __init__(self, szl, szh):    
    self.szl = szl
    self.szh = szh        
    self.id = 0
    self.size = randint(szl, szh)

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nid: " + str(self.id)
    strRep = strRep + "\nsize: " + str(self.size)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>id</name>"
    strRep = strRep + "\n\t<type>Integer</type>"
    strRep = strRep + "\n\t<value>"+str(self.id)+"</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>size</name>"
    strRep = strRep + "\n\t<type>Integer</type>"
    strRep = strRep + "\n\t<value>"+str(self.size)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class JobPort( Port ):
  def __init__(self):            
    self.msg = None

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nmsg: " + str(self.msg)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>msg</name>"
    strRep = strRep + "\n\t<type>Job</type>"
    if(self.msg!=None):
      strRep = strRep + "\n\t<value>"+self.msg.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class GeneratorState( object ):
  def __init__(self):            
    self.seqState = Generator.idle

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nseqState: " + str(self.seqState)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>seqState</name>"
    strRep = strRep + "\n\t<type>Generator.SeqStates</type>"
    strRep = strRep + "\n\t<value>"+str(self.seqState)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Generator( AtomicDEVS ): #"A Generator"
  idle = 0
  generating = 1
  def __init__(self, ia, ib, szl, szh, name):
    AtomicDEVS.__init__(self, name)    
    self.ia = ia
    self.ib = ib
    self.szl = szl
    self.szh = szh
    self.name = name    
    self.inport = self.addInPort("inport")    
    self.out = self.addOutPort("out")
    
    self.state = GeneratorState()
    
  def intTransition( self ):

    if(self.state.seqState == Generator.idle):
      self.state.seqState = Generator.generating
    else:
      self.state.seqState = Generator.idle
    return self.state

  
  def timeAdvance( self ):

    timespan = randint(self.ia, self.ib)
    return timespan

  
  def outputFnc( self ):

    self.poke(self.out, Job(self.szl, self.szh))



  def __str__(self):
    strRep = ''
    strRep = strRep + "\nstate: " + str(self.state)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>state</name>"
    strRep = strRep + "\n\t<type>GeneratorState</type>"
    if(self.state!=None):
      strRep = strRep + "\n\t<value>"+self.state.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class ProcessorState( object ):
  def __init__(self, n):    
    self.n = n        
    self.queue = []
    self.currentJob = None
    self.timeElapsed = 0.0
    self.seqState = Processor.idle

  def __str__(self):
    strRep = ''
    strRep = strRep + "\nqueue: "
    for item in self.queue:
      strRep = strRep + "\n\t" + str(item)
    strRep = strRep + "\ncurrentJob: " + str(self.currentJob)
    strRep = strRep + "\ntimeElapsed: " + str(self.timeElapsed)
    strRep = strRep + "\nseqState: " + str(self.seqState)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>queue</name>"
    strRep = strRep + "\n\t<type>list</type>"
    strRep = strRep + "\n\t<value>"
    for item in self.queue:
      strRep = strRep + "\n\t" + item.toXML()
    strRep = strRep + "\n\t</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>currentJob</name>"
    strRep = strRep + "\n\t<type>Job</type>"
    if(self.currentJob!=None):
      strRep = strRep + "\n\t<value>"+self.currentJob.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>timeElapsed</name>"
    strRep = strRep + "\n\t<type>Real</type>"
    strRep = strRep + "\n\t<value>"+str(self.timeElapsed)+"</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"P\">"
    strRep = strRep + "\n\t<name>seqState</name>"
    strRep = strRep + "\n\t<type>Processor.SeqStates</type>"
    strRep = strRep + "\n\t<value>"+str(self.seqState)+"</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Processor( AtomicDEVS ): #"A Processor Definition"
  idle = 0
  busy = 1
  discarding = 2
  def __init__(self, qn, name):
    AtomicDEVS.__init__(self, name)    
    self.qn = qn
    self.name = name    
    self.inport = self.addInPort("inport")    
    self.out = self.addOutPort("out")
    self.discard = self.addOutPort("discard")
    
    self.state = ProcessorState(n = qn)
    
  def extTransition( self ):

    msg = None
    msg = self.peek(self.inport)

    if(msg != None):
      self.state.queue.append(msg)
      msg = None

    if(self.state.seqState == Processor.idle):
      self.state.currentJob = self.state.queue[0]
      self.state.queue.remove(self.state.queue[0])
      self.state.seqState = Processor.busy
    else:
      self.state.timeElapsed = self.state.timeElapsed + self.elapsed
      if(len(self.state.queue) > self.qn):
        self.state.seqState = Processor.discarding
      else:
        self.state.seqState = Processor.busy
    return self.state

  
  def intTransition( self ):

    if(self.state.seqState == Processor.busy):
      self.state.timeElapsed = 0.0
      if(len(self.state.queue) < 1):
        self.state.currentJob = None
        self.state.seqState = Processor.idle
      else:
        self.state.currentJob = self.state.queue[0]
        self.state.queue.remove(self.state.queue[0])
        self.state.seqState = Processor.busy

    if(self.state.seqState == Processor.discarding):
      self.state.seqState = Processor.busy
    return self.state

  
  def timeAdvance( self ):

    if(self.state.seqState == Processor.idle):
      timespan = INFINITY
    elif(self.state.seqState == Processor.busy):
      timespan = self.state.currentJob.size - self.state.timeElapsed
    else:
      timespan = 0
    return timespan

  
  def outputFnc( self ):

    msg = None
    if(self.state.seqState == Processor.busy):
      self.poke(self.out, self.state.currentJob)

    if(self.state.seqState == Processor.discarding):
      msg = self.state.queue.pop()
      self.poke(self.discard, msg)



  def __str__(self):
    strRep = ''
    strRep = strRep + "\nstate: " + str(self.state)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>state</name>"
    strRep = strRep + "\n\t<type>ProcessorState</type>"
    if(self.state!=None):
      strRep = strRep + "\n\t<value>"+self.state.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep

class Root( CoupledDEVS ):
  def __init__(self, pqs, pia, pib, psa, psb, name):
    CoupledDEVS.__init__(self, name)    
    self.pqs = pqs
    self.pia = pia
    self.pib = pib
    self.psa = psa
    self.psb = psb
    self.name = name        
    self.out = self.addOutPort("out")
    self.discard = self.addOutPort("discard")
    self.generator = Generator(ia = pia, ib = pib, szl = psa, szh = psb, name='generator')
    self.addSubModel( self.generator )
    self.p1 = Processor(qn = pqs, name='p1')
    self.addSubModel( self.p1 )
    self.p2 = Processor(qn = pqs, name='p2')
    self.addSubModel( self.p2 )
    self.p3 = Processor(qn = pqs, name='p3')
    self.addSubModel( self.p3 )  
    self.connectPorts( self.generator.out, self.p1.inport)  
    self.connectPorts( self.p1.out, self.out)  
    self.connectPorts( self.p1.discard, self.p2.inport)  
    self.connectPorts( self.p2.out, self.out)  
    self.connectPorts( self.p2.discard, self.p3.inport)  
    self.connectPorts( self.p3.out, self.out)  
    self.connectPorts( self.p3.discard, self.discard)

  def __str__(self):
    strRep = ''
    strRep = strRep + "\ngenerator: " + str(self.generator)
    strRep = strRep + "\np1: " + str(self.p1)
    strRep = strRep + "\np2: " + str(self.p2)
    strRep = strRep + "\np3: " + str(self.p3)
    return strRep

  def toXML(self):
    strRep = ''
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>generator</name>"
    strRep = strRep + "\n\t<type>Generator</type>"
    if(self.generator!=None):
      strRep = strRep + "\n\t<value>"+self.generator.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p1</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p1!=None):
      strRep = strRep + "\n\t<value>"+self.p1.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p2</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p2!=None):
      strRep = strRep + "\n\t<value>"+self.p2.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    strRep = strRep + "\n<attribute category=\"C\">"
    strRep = strRep + "\n\t<name>p3</name>"
    strRep = strRep + "\n\t<type>Processor</type>"
    if(self.p3!=None):
      strRep = strRep + "\n\t<value>"+self.p3.toXML()+"</value>"
    else:
      strRep = strRep + "\n\t<value>None</value>"
    strRep = strRep + "\n</attribute>"
    return strRep



